@@ -1,13 +1,14 @@
+use crate::storage::sqlite3_ondisk::MAX_PAGE_SIZE;
+use crate::turso_assert;
 use crate::{
     storage::{
         self,
         pager::{PageRef, Pager},
         sqlite3_ondisk::DATABASE_HEADER_PAGE_ID,
     },
-    types::CursorResult,
+    types::IOResult,
     LimboError, Result,
 };
-use std::sync::atomic::Ordering;
 
 // const HEADER_OFFSET_MAGIC: usize = 0;
 const HEADER_OFFSET_PAGE_SIZE: usize = 16;
@@ -34,22 +35,22 @@ const HEADER_OFFSET_VERSION_VALID_FOR: usize = 92;
 const HEADER_OFFSET_VERSION_NUMBER: usize = 96;
 
 // Helper to get a read-only reference to the header page.
-fn get_header_page(pager: &Pager) -> Result<CursorResult<PageRef>> {
-    if pager.is_empty.load(Ordering::SeqCst) < 2 {
+fn get_header_page(pager: &Pager) -> Result<IOResult<PageRef>> {
+    if !pager.db_state.is_initialized() {
         return Err(LimboError::InternalError(
             "Database is empty, header does not exist - page 1 should've been allocated before this".to_string(),
         ));
     }
     let page = pager.read_page(DATABASE_HEADER_PAGE_ID)?;
     if page.is_locked() {
-        return Ok(CursorResult::IO);
+        return Ok(IOResult::IO);
     }
-    Ok(CursorResult::Ok(page))
+    Ok(IOResult::Done(page))
 }
 
 // Helper to get a writable reference to the header page and mark it dirty.
-fn get_header_page_for_write(pager: &Pager) -> Result<CursorResult<PageRef>> {
-    if pager.is_empty.load(Ordering::SeqCst) < 2 {
+fn get_header_page_for_write(pager: &Pager) -> Result<IOResult<PageRef>> {
+    if !pager.db_state.is_initialized() {
         // This should not be called on an empty DB for writing, as page 1 is allocated on first transaction.
         return Err(LimboError::InternalError(
             "Cannot write to header of an empty database - page 1 should've been allocated before this".to_string(),
@@ -57,22 +58,25 @@ fn get_header_page_for_write(pager: &Pager) -> Result<CursorResult<PageRef>> {
     }
     let page = pager.read_page(DATABASE_HEADER_PAGE_ID)?;
     if page.is_locked() {
-        return Ok(CursorResult::IO);
+        return Ok(IOResult::IO);
     }
-    page.set_dirty();
-    pager.add_dirty(DATABASE_HEADER_PAGE_ID);
-    Ok(CursorResult::Ok(page))
+    turso_assert!(
+        page.get().id == DATABASE_HEADER_PAGE_ID,
+        "page must have number 1"
+    );
+    pager.add_dirty(&page);
+    Ok(IOResult::Done(page))
 }
 
 /// Helper function to run async header accessors until completion
 fn run_header_accessor_until_done<T, F>(pager: &Pager, mut accessor: F) -> Result<T>
 where
-    F: FnMut() -> Result<CursorResult<T>>,
+    F: FnMut() -> Result<IOResult<T>>,
 {
     loop {
         match accessor()? {
-            CursorResult::Ok(value) => return Ok(value),
-            CursorResult::IO => {
+            IOResult::Done(value) => return Ok(value),
+            IOResult::IO => {
                 pager.io.run_once()?;
             }
         }
@@ -102,13 +106,13 @@ macro_rules! impl_header_field_accessor {
         paste::paste! {
             // Async version
             #[allow(dead_code)]
-            pub fn [<get_ $field_name _async>](pager: &Pager) -> Result<CursorResult<$type>> {
-                if pager.is_empty.load(Ordering::SeqCst) < 2 {
+            pub fn [<get_ $field_name _async>](pager: &Pager) -> Result<IOResult<$type>> {
+                if !pager.db_state.is_initialized() {
                     return Err(LimboError::InternalError(format!("Database is empty, header does not exist - page 1 should've been allocated before this")));
                 }
                 let page = match get_header_page(pager)? {
-                    CursorResult::Ok(page) => page,
-                    CursorResult::IO => return Ok(CursorResult::IO),
+                    IOResult::Done(page) => page,
+                    IOResult::IO => return Ok(IOResult::IO),
                 };
                 let page_inner = page.get();
                 let page_content = page_inner.contents.as_ref().unwrap();
@@ -119,10 +123,10 @@ macro_rules! impl_header_field_accessor {
                 let value = <$type>::from_be_bytes(bytes);
                 $(
                     if value == 0 {
-                        return Ok(CursorResult::Ok($ifzero));
+                        return Ok(IOResult::Done($ifzero));
                     }
                 )?
-                Ok(CursorResult::Ok(value))
+                Ok(IOResult::Done(value))
             }
 
             // Sync version
@@ -133,19 +137,19 @@ macro_rules! impl_header_field_accessor {
 
             // Async setter
             #[allow(dead_code)]
-            pub fn [<set_ $field_name _async>](pager: &Pager, value: $type) -> Result<CursorResult<()>> {
+            pub fn [<set_ $field_name _async>](pager: &Pager, value: $type) -> Result<IOResult<()>> {
                 let page = match get_header_page_for_write(pager)? {
-                    CursorResult::Ok(page) => page,
-                    CursorResult::IO => return Ok(CursorResult::IO),
+                    IOResult::Done(page) => page,
+                    IOResult::IO => return Ok(IOResult::IO),
                 };
                 let page_inner = page.get();
                 let page_content = page_inner.contents.as_ref().unwrap();
                 let mut buf = page_content.buffer.borrow_mut();
                 let buf_slice = buf.as_mut_slice();
                 buf_slice[$offset..$offset + std::mem::size_of::<$type>()].copy_from_slice(&value.to_be_bytes());
-                page.set_dirty();
-                pager.add_dirty(1);
-                Ok(CursorResult::Ok(()))
+                turso_assert!(page.get().id == 1, "page must have number 1");
+                pager.add_dirty(&page);
+                Ok(IOResult::Done(()))
             }
 
             // Sync setter
@@ -158,7 +162,7 @@ macro_rules! impl_header_field_accessor {
 }
 
 // impl_header_field_accessor!(magic, [u8; 16], HEADER_OFFSET_MAGIC);
-impl_header_field_accessor!(page_size, u16, HEADER_OFFSET_PAGE_SIZE);
+impl_header_field_accessor!(page_size_u16, u16, HEADER_OFFSET_PAGE_SIZE);
 impl_header_field_accessor!(write_version, u8, HEADER_OFFSET_WRITE_VERSION);
 impl_header_field_accessor!(read_version, u8, HEADER_OFFSET_READ_VERSION);
 impl_header_field_accessor!(reserved_space, u8, HEADER_OFFSET_RESERVED_SPACE);
@@ -189,7 +193,38 @@ impl_header_field_accessor!(
     u32,
     HEADER_OFFSET_INCREMENTAL_VACUUM_ENABLED
 );
-impl_header_field_accessor!(application_id, u32, HEADER_OFFSET_APPLICATION_ID);
+impl_header_field_accessor!(application_id, i32, HEADER_OFFSET_APPLICATION_ID);
 //impl_header_field_accessor!(reserved_for_expansion, [u8; 20], HEADER_OFFSET_RESERVED_FOR_EXPANSION);
 impl_header_field_accessor!(version_valid_for, u32, HEADER_OFFSET_VERSION_VALID_FOR);
 impl_header_field_accessor!(version_number, u32, HEADER_OFFSET_VERSION_NUMBER);
+
+pub fn get_page_size(pager: &Pager) -> Result<u32> {
+    let size = get_page_size_u16(pager)?;
+    if size == 1 {
+        return Ok(MAX_PAGE_SIZE);
+    }
+    Ok(size as u32)
+}
+
+#[allow(dead_code)]
+pub fn set_page_size(pager: &Pager, value: u32) -> Result<()> {
+    let page_size = if value == MAX_PAGE_SIZE {
+        1
+    } else {
+        value as u16
+    };
+    set_page_size_u16(pager, page_size)
+}
+
+#[allow(dead_code)]
+pub fn get_page_size_async(pager: &Pager) -> Result<IOResult<u32>> {
+    match get_page_size_u16_async(pager)? {
+        IOResult::Done(size) => {
+            if size == 1 {
+                return Ok(IOResult::Done(MAX_PAGE_SIZE));
+            }
+            Ok(IOResult::Done(size as u32))
+        }
+        IOResult::IO => Ok(IOResult::IO),
+    }
+}
