@@ -49,9 +49,7 @@ use super::pager::PageRef;
 use super::wal::LimboRwLock;
 use crate::error::LimboError;
 use crate::fast_lock::SpinLock;
-use crate::io::{
-    Buffer, Complete, Completion, CompletionType, ReadCompletion, SyncCompletion, WriteCompletion,
-};
+use crate::io::{Buffer, Complete, Completion};
 use crate::storage::btree::offset::{
     BTREE_CELL_CONTENT_AREA, BTREE_CELL_COUNT, BTREE_FIRST_FREEBLOCK, BTREE_FRAGMENTED_BYTES_COUNT,
     BTREE_PAGE_TYPE, BTREE_RIGHTMOST_PTR,
@@ -83,10 +81,10 @@ pub const MIN_PAGE_CACHE_SIZE: usize = 10;
 pub const MIN_PAGE_SIZE: u32 = 512;
 
 /// The maximum page size in bytes.
-const MAX_PAGE_SIZE: u32 = 65536;
+pub const MAX_PAGE_SIZE: u32 = 65536;
 
 /// The default page size in bytes.
-pub const DEFAULT_PAGE_SIZE: u16 = 4096;
+pub const DEFAULT_PAGE_SIZE: u32 = 4096;
 
 pub const DATABASE_HEADER_PAGE_ID: usize = 1;
 
@@ -98,6 +96,37 @@ pub const INTERIOR_PAGE_HEADER_SIZE_BYTES: usize = 12;
 pub const LEAF_PAGE_HEADER_SIZE_BYTES: usize = 8;
 pub const LEFT_CHILD_PTR_SIZE_BYTES: usize = 4;
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+#[repr(u32)]
+pub enum DatabaseEncoding {
+    Utf8 = 1,
+    Utf16Le = 2,
+    Utf16Be = 3,
+}
+
+impl TryFrom<u32> for DatabaseEncoding {
+    type Error = LimboError;
+
+    fn try_from(value: u32) -> Result<Self> {
+        match value {
+            1 => Ok(Self::Utf8),
+            2 => Ok(Self::Utf16Le),
+            3 => Ok(Self::Utf16Be),
+            _ => Err(LimboError::Corrupt(format!("Invalid encoding: {value}"))),
+        }
+    }
+}
+
+impl From<DatabaseEncoding> for &'static str {
+    fn from(encoding: DatabaseEncoding) -> Self {
+        match encoding {
+            DatabaseEncoding::Utf8 => "UTF-8",
+            DatabaseEncoding::Utf16Le => "UTF-16le",
+            DatabaseEncoding::Utf16Be => "UTF-16be",
+        }
+    }
+}
+
 /// The database header.
 /// The first 100 bytes of the database file comprise the database file header.
 /// The database file header is divided into fields as shown by the table below.
@@ -228,30 +257,36 @@ pub struct WalHeader {
 #[derive(Debug, Default, Copy, Clone)]
 pub struct WalFrameHeader {
     /// Page number
-    page_number: u32,
+    pub(crate) page_number: u32,
 
     /// For commit records, the size of the database file in pages after the commit.
     /// For all other records, zero.
-    db_size: u32,
+    pub(crate) db_size: u32,
 
     /// Salt-1 copied from the WAL header
-    salt_1: u32,
+    pub(crate) salt_1: u32,
 
     /// Salt-2 copied from the WAL header
-    salt_2: u32,
+    pub(crate) salt_2: u32,
 
     /// Checksum-1: Cumulative checksum up through and including this page
-    checksum_1: u32,
+    pub(crate) checksum_1: u32,
 
     /// Checksum-2: Second half of the cumulative checksum
-    checksum_2: u32,
+    pub(crate) checksum_2: u32,
+}
+
+impl WalFrameHeader {
+    pub fn is_commit_frame(&self) -> bool {
+        self.db_size > 0
+    }
 }
 
 impl Default for DatabaseHeader {
     fn default() -> Self {
         Self {
             magic: *b"SQLite format 3\0",
-            page_size: DEFAULT_PAGE_SIZE,
+            page_size: DEFAULT_PAGE_SIZE as u16,
             write_version: 2,
             read_version: 2,
             reserved_space: 0,
@@ -279,7 +314,7 @@ impl Default for DatabaseHeader {
 
 impl DatabaseHeader {
     pub fn update_page_size(&mut self, size: u32) {
-        if !(MIN_PAGE_SIZE..=MAX_PAGE_SIZE).contains(&size) || (size & (size - 1) != 0) {
+        if !is_valid_page_size(size) {
             return;
         }
 
@@ -299,6 +334,10 @@ impl DatabaseHeader {
     }
 }
 
+pub fn is_valid_page_size(size: u32) -> bool {
+    (MIN_PAGE_SIZE..=MAX_PAGE_SIZE).contains(&size) && (size & (size - 1)) == 0
+}
+
 pub fn write_header_to_buf(buf: &mut [u8], header: &DatabaseHeader) {
     buf[0..16].copy_from_slice(&header.magic);
     buf[16..18].copy_from_slice(&header.page_size.to_be_bytes());
@@ -620,9 +659,9 @@ impl PageContent {
         let cell_pointer = self.read_u16_no_offset(cell_pointer) as usize;
         let start = cell_pointer;
         let payload_overflow_threshold_max =
-            payload_overflow_threshold_max(self.page_type(), usable_size as u16);
+            payload_overflow_threshold_max(self.page_type(), usable_size);
         let payload_overflow_threshold_min =
-            payload_overflow_threshold_min(self.page_type(), usable_size as u16);
+            payload_overflow_threshold_min(self.page_type(), usable_size);
         let len = match self.page_type() {
             PageType::IndexInterior => {
                 let (len_payload, n_payload) = read_varint(&buf[cell_pointer + 4..]).unwrap();
@@ -717,7 +756,7 @@ impl PageContent {
     }
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 pub fn begin_read_page(
     db_file: Arc<dyn DatabaseStorage>,
     buffer_pool: Arc<BufferPool>,
@@ -743,17 +782,18 @@ pub fn begin_read_page(
             page.set_error();
         }
     });
-    let c = Completion::new(CompletionType::Read(ReadCompletion::new(buf, complete)));
+    let c = Completion::new_read(buf, complete);
     db_file.read_page(page_idx, c)?;
     Ok(())
 }
 
+#[instrument(skip_all, level = Level::INFO)]
 pub fn finish_read_page(
     page_idx: usize,
     buffer_ref: Arc<RefCell<Buffer>>,
     page: PageRef,
 ) -> Result<()> {
-    tracing::trace!("finish_read_btree_page(page_idx = {})", page_idx);
+    tracing::trace!(page_idx);
     let pos = if page_idx == DATABASE_HEADER_PAGE_ID {
         DATABASE_HEADER_SIZE
     } else {
@@ -769,7 +809,7 @@ pub fn finish_read_page(
     Ok(())
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 pub fn begin_write_btree_page(
     pager: &Pager,
     page: &PageRef,
@@ -804,7 +844,7 @@ pub fn begin_write_btree_page(
             );
         })
     };
-    let c = Completion::new(CompletionType::Write(WriteCompletion::new(write_complete)));
+    let c = Completion::new_write(write_complete);
     let res = page_source.write_page(page_id, buffer.clone(), c);
     if res.is_err() {
         // Avoid infinite loop if write page fails
@@ -813,15 +853,13 @@ pub fn begin_write_btree_page(
     res
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 pub fn begin_sync(db_file: Arc<dyn DatabaseStorage>, syncing: Rc<RefCell<bool>>) -> Result<()> {
     assert!(!*syncing.borrow());
     *syncing.borrow_mut() = true;
-    let completion = Completion::new(CompletionType::Sync(SyncCompletion {
-        complete: Box::new(move |_| {
-            *syncing.borrow_mut() = false;
-        }),
-    }));
+    let completion = Completion::new_sync(move |_| {
+        *syncing.borrow_mut() = false;
+    });
     #[allow(clippy::arc_with_non_send_sync)]
     db_file.sync(completion)?;
     Ok(())
@@ -878,8 +916,8 @@ pub fn read_btree_cell(
     usable_size: usize,
 ) -> Result<BTreeCell> {
     let page_type = page_content.page_type();
-    let max_local = payload_overflow_threshold_max(page_type, usable_size as u16);
-    let min_local = payload_overflow_threshold_min(page_type, usable_size as u16);
+    let max_local = payload_overflow_threshold_max(page_type, usable_size);
+    let min_local = payload_overflow_threshold_min(page_type, usable_size);
     match page_type {
         PageType::IndexInterior => {
             let mut pos = pos;
@@ -1256,6 +1294,26 @@ pub fn read_varint(buf: &[u8]) -> Result<(u64, usize)> {
     Ok((v, 9))
 }
 
+pub fn varint_len(value: u64) -> usize {
+    if value <= 0x7f {
+        return 1;
+    }
+    if value <= 0x3fff {
+        return 2;
+    }
+    if (value & ((0xff000000_u64) << 32)) > 0 {
+        return 9;
+    }
+
+    let mut bytes = value;
+    let mut n = 0;
+    while bytes != 0 {
+        bytes >>= 7;
+        n += 1;
+    }
+    n
+}
+
 pub fn write_varint(buf: &mut [u8], value: u64) -> usize {
     if value <= 0x7f {
         buf[0] = (value & 0x7f) as u8;
@@ -1374,15 +1432,22 @@ pub fn read_entire_wal_dumb(file: &Arc<dyn File>) -> Result<Arc<UnsafeCell<WalFi
             use_native_endian_checksum,
         );
 
-        if calculated_header_checksum != (header_locked.checksum_1, header_locked.checksum_2) {
-            panic!(
-                "WAL header checksum mismatch. Expected ({}, {}), Got ({}, {})",
+        let checksum_header_failed = if calculated_header_checksum
+            != (header_locked.checksum_1, header_locked.checksum_2)
+        {
+            tracing::error!(
+                "WAL header checksum mismatch. Expected ({}, {}), Got ({}, {}). Ignoring frames starting from frame {}",
                 header_locked.checksum_1,
                 header_locked.checksum_2,
                 calculated_header_checksum.0,
-                calculated_header_checksum.1
+                calculated_header_checksum.1,
+                0
+
             );
-        }
+            true
+        } else {
+            false
+        };
 
         let mut cumulative_checksum = (header_locked.checksum_1, header_locked.checksum_2);
         let page_size_u32 = header_locked.page_size;
@@ -1399,98 +1464,131 @@ pub fn read_entire_wal_dumb(file: &Arc<dyn File>) -> Result<Arc<UnsafeCell<WalFi
 
         let wfs_data = unsafe { &mut *wal_file_shared_for_completion.get() };
 
-        while current_offset + WAL_FRAME_HEADER_SIZE + page_size <= buf_slice.len() {
-            let frame_header_slice =
-                &buf_slice[current_offset..current_offset + WAL_FRAME_HEADER_SIZE];
-            let page_data_slice = &buf_slice[current_offset + WAL_FRAME_HEADER_SIZE
-                ..current_offset + WAL_FRAME_HEADER_SIZE + page_size];
+        if !checksum_header_failed {
+            while current_offset + WAL_FRAME_HEADER_SIZE + page_size <= buf_slice.len() {
+                let frame_header_slice =
+                    &buf_slice[current_offset..current_offset + WAL_FRAME_HEADER_SIZE];
+                let page_data_slice = &buf_slice[current_offset + WAL_FRAME_HEADER_SIZE
+                    ..current_offset + WAL_FRAME_HEADER_SIZE + page_size];
 
-            let frame_h_page_number =
-                u32::from_be_bytes(frame_header_slice[0..4].try_into().unwrap());
-            let frame_h_db_size = u32::from_be_bytes(frame_header_slice[4..8].try_into().unwrap());
-            let frame_h_salt_1 = u32::from_be_bytes(frame_header_slice[8..12].try_into().unwrap());
-            let frame_h_salt_2 = u32::from_be_bytes(frame_header_slice[12..16].try_into().unwrap());
-            let frame_h_checksum_1 =
-                u32::from_be_bytes(frame_header_slice[16..20].try_into().unwrap());
-            let frame_h_checksum_2 =
-                u32::from_be_bytes(frame_header_slice[20..24].try_into().unwrap());
+                let frame_h_page_number =
+                    u32::from_be_bytes(frame_header_slice[0..4].try_into().unwrap());
+                let frame_h_db_size =
+                    u32::from_be_bytes(frame_header_slice[4..8].try_into().unwrap());
+                let frame_h_salt_1 =
+                    u32::from_be_bytes(frame_header_slice[8..12].try_into().unwrap());
+                let frame_h_salt_2 =
+                    u32::from_be_bytes(frame_header_slice[12..16].try_into().unwrap());
+                let frame_h_checksum_1 =
+                    u32::from_be_bytes(frame_header_slice[16..20].try_into().unwrap());
+                let frame_h_checksum_2 =
+                    u32::from_be_bytes(frame_header_slice[20..24].try_into().unwrap());
 
-            // It contains more frames with mismatched SALT values, which means they're leftovers from previous checkpoints
-            if frame_h_salt_1 != header_locked.salt_1 || frame_h_salt_2 != header_locked.salt_2 {
-                tracing::trace!(
-                    "WAL frame salt mismatch: expected ({}, {}), got ({}, {}), ignoring frame",
-                    header_locked.salt_1,
-                    header_locked.salt_2,
-                    frame_h_salt_1,
-                    frame_h_salt_2
+                if frame_h_page_number == 0 {
+                    tracing::trace!(
+                        "WAL frame with page number 0. Ignoring frames starting from frame {}",
+                        frame_idx
+                    );
+                    break;
+                }
+                // It contains more frames with mismatched SALT values, which means they're leftovers from previous checkpoints
+                if frame_h_salt_1 != header_locked.salt_1 || frame_h_salt_2 != header_locked.salt_2
+                {
+                    tracing::trace!(
+                        "WAL frame salt mismatch: expected ({}, {}), got ({}, {}). Ignoring frames starting from frame {}",
+                        header_locked.salt_1,
+                        header_locked.salt_2,
+                        frame_h_salt_1,
+                        frame_h_salt_2,
+                        frame_idx
+                    );
+                    break;
+                }
+
+                let checksum_after_fh_meta = checksum_wal(
+                    &frame_header_slice[0..8],
+                    &header_locked,
+                    cumulative_checksum,
+                    use_native_endian_checksum,
                 );
-                break;
-            }
-
-            let checksum_after_fh_meta = checksum_wal(
-                &frame_header_slice[0..8],
-                &header_locked,
-                cumulative_checksum,
-                use_native_endian_checksum,
-            );
-            let calculated_frame_checksum = checksum_wal(
-                page_data_slice,
-                &header_locked,
-                checksum_after_fh_meta,
-                use_native_endian_checksum,
-            );
-
-            tracing::debug!(
-                "read_entire_wal_dumb(frame_h_checksum=({}, {}), calculated_frame_checksum=({}, {}))",
-                frame_h_checksum_1,
-                frame_h_checksum_2,
-                calculated_frame_checksum.0,
-                calculated_frame_checksum.1
-            );
-            if calculated_frame_checksum != (frame_h_checksum_1, frame_h_checksum_2) {
-                panic!(
-                    "WAL frame checksum mismatch. Expected ({}, {}), Got ({}, {})",
+                let calculated_frame_checksum = checksum_wal(
+                    page_data_slice,
+                    &header_locked,
+                    checksum_after_fh_meta,
+                    use_native_endian_checksum,
+                );
+                tracing::debug!(
+                    "read_entire_wal_dumb(frame_h_checksum=({}, {}), calculated_frame_checksum=({}, {}))",
                     frame_h_checksum_1,
                     frame_h_checksum_2,
                     calculated_frame_checksum.0,
                     calculated_frame_checksum.1
                 );
+
+                if calculated_frame_checksum != (frame_h_checksum_1, frame_h_checksum_2) {
+                    tracing::error!(
+                        "WAL frame checksum mismatch. Expected ({}, {}), Got ({}, {}). Ignoring frames starting from frame {}",
+                        frame_h_checksum_1,
+                        frame_h_checksum_2,
+                        calculated_frame_checksum.0,
+                        calculated_frame_checksum.1,
+                        frame_idx
+                    );
+                    break;
+                }
+
+                cumulative_checksum = calculated_frame_checksum;
+
+                wfs_data
+                    .frame_cache
+                    .lock()
+                    .entry(frame_h_page_number as u64)
+                    .or_default()
+                    .push(frame_idx);
+                wfs_data
+                    .pages_in_frames
+                    .lock()
+                    .push(frame_h_page_number as u64);
+
+                let is_commit_record = frame_h_db_size > 0;
+                if is_commit_record {
+                    wfs_data.max_frame.store(frame_idx, Ordering::SeqCst);
+                    wfs_data.last_checksum = cumulative_checksum;
+                }
+
+                frame_idx += 1;
+                current_offset += WAL_FRAME_HEADER_SIZE + page_size;
             }
-
-            cumulative_checksum = calculated_frame_checksum;
-
-            wfs_data
-                .frame_cache
-                .lock()
-                .entry(frame_h_page_number as u64)
-                .or_default()
-                .push(frame_idx);
-            wfs_data
-                .pages_in_frames
-                .lock()
-                .push(frame_h_page_number as u64);
-
-            let is_commit_record = frame_h_db_size > 0;
-            if is_commit_record {
-                wfs_data.max_frame.store(frame_idx, Ordering::SeqCst);
-            }
-
-            frame_idx += 1;
-            current_offset += WAL_FRAME_HEADER_SIZE + page_size;
         }
 
         wfs_data.last_checksum = cumulative_checksum;
+        wfs_data.nbackfills.store(0, Ordering::SeqCst);
         wfs_data.loaded.store(true, Ordering::SeqCst);
     });
-    let c = Completion::new(CompletionType::Read(ReadCompletion::new(
-        buf_for_pread,
-        complete,
-    )));
-    file.pread(0, c)?;
+    let c = Completion::new_read(buf_for_pread, complete);
+    file.pread(0, c.into())?;
 
     Ok(wal_file_shared_ret)
 }
 
+pub fn begin_read_wal_frame_raw(
+    io: &Arc<dyn File>,
+    offset: usize,
+    page_size: u32,
+    complete: Box<dyn Fn(Arc<RefCell<Buffer>>, i32)>,
+) -> Result<Arc<Completion>> {
+    tracing::trace!("begin_read_wal_frame_raw(offset={})", offset);
+    let drop_fn = Rc::new(|_buf| {});
+    let buf = Arc::new(RefCell::new(Buffer::allocate(
+        page_size as usize + WAL_FRAME_HEADER_SIZE,
+        drop_fn,
+    )));
+    #[allow(clippy::arc_with_non_send_sync)]
+    let c = Completion::new_read(buf, complete);
+    let c = io.pread(offset, c.into())?;
+    Ok(c)
+}
+
 pub fn begin_read_wal_frame(
     io: &Arc<dyn File>,
     offset: usize,
@@ -1505,109 +1603,63 @@ pub fn begin_read_wal_frame(
     });
     let buf = Arc::new(RefCell::new(Buffer::new(buf, drop_fn)));
     #[allow(clippy::arc_with_non_send_sync)]
-    let c = Completion::new(CompletionType::Read(ReadCompletion::new(buf, complete)));
-    let c = io.pread(offset, c)?;
+    let c = Completion::new_read(buf, complete);
+    let c = io.pread(offset, c.into())?;
     Ok(c)
 }
 
-#[instrument(err,skip(io, page, write_counter, wal_header, checksums), level = Level::INFO)]
-#[allow(clippy::too_many_arguments)]
-pub fn begin_write_wal_frame(
-    io: &Arc<dyn File>,
-    offset: usize,
-    page: &PageRef,
-    page_size: u16,
-    db_size: u32,
-    write_counter: Rc<RefCell<usize>>,
-    wal_header: &WalHeader,
-    checksums: (u32, u32),
-) -> Result<(u32, u32)> {
-    let page_finish = page.clone();
-    let page_id = page.get().id;
-    tracing::trace!(page_id);
-
-    let mut header = WalFrameHeader {
-        page_number: page_id as u32,
+pub fn parse_wal_frame_header(frame: &[u8]) -> (WalFrameHeader, &[u8]) {
+    let page_number = u32::from_be_bytes(frame[0..4].try_into().unwrap());
+    let db_size = u32::from_be_bytes(frame[4..8].try_into().unwrap());
+    let salt_1 = u32::from_be_bytes(frame[8..12].try_into().unwrap());
+    let salt_2 = u32::from_be_bytes(frame[12..16].try_into().unwrap());
+    let checksum_1 = u32::from_be_bytes(frame[16..20].try_into().unwrap());
+    let checksum_2 = u32::from_be_bytes(frame[20..24].try_into().unwrap());
+    let header = WalFrameHeader {
+        page_number,
         db_size,
-        salt_1: wal_header.salt_1,
-        salt_2: wal_header.salt_2,
-        checksum_1: 0,
-        checksum_2: 0,
+        salt_1,
+        salt_2,
+        checksum_1,
+        checksum_2,
     };
-    let (buffer, checksums) = {
-        let page = page.get();
-        let contents = page.contents.as_ref().unwrap();
-        let drop_fn = Rc::new(|_buf| {});
+    let page = &frame[WAL_FRAME_HEADER_SIZE..];
+    (header, page)
+}
 
-        let mut buffer = Buffer::allocate(
-            contents.buffer.borrow().len() + WAL_FRAME_HEADER_SIZE,
-            drop_fn,
-        );
-        let buf = buffer.as_mut_slice();
-        buf[0..4].copy_from_slice(&header.page_number.to_be_bytes());
-        buf[4..8].copy_from_slice(&header.db_size.to_be_bytes());
-        buf[8..12].copy_from_slice(&header.salt_1.to_be_bytes());
-        buf[12..16].copy_from_slice(&header.salt_2.to_be_bytes());
+pub fn prepare_wal_frame(
+    wal_header: &WalHeader,
+    prev_checksums: (u32, u32),
+    page_size: u32,
+    page_number: u32,
+    db_size: u32,
+    page: &[u8],
+) -> ((u32, u32), Arc<RefCell<Buffer>>) {
+    tracing::trace!(page_number);
 
-        let contents_buf = contents.as_ptr();
-        let content_len = contents_buf.len();
-        buf[WAL_FRAME_HEADER_SIZE..WAL_FRAME_HEADER_SIZE + content_len]
-            .copy_from_slice(contents_buf);
-        if content_len < page_size as usize {
-            buf[WAL_FRAME_HEADER_SIZE + content_len..WAL_FRAME_HEADER_SIZE + page_size as usize]
-                .fill(0);
-        }
+    let drop_fn = Rc::new(|_buf| {});
+    let mut buffer = Buffer::allocate(page_size as usize + WAL_FRAME_HEADER_SIZE, drop_fn);
+    let frame = buffer.as_mut_slice();
+    frame[WAL_FRAME_HEADER_SIZE..].copy_from_slice(page);
 
-        let expects_be = wal_header.magic & 1;
-        let use_native_endian = cfg!(target_endian = "big") as u32 == expects_be;
-        let header_checksum = checksum_wal(&buf[0..8], wal_header, checksums, use_native_endian);
-        let final_checksum = checksum_wal(
-            &buf[WAL_FRAME_HEADER_SIZE..WAL_FRAME_HEADER_SIZE + page_size as usize],
-            wal_header,
-            header_checksum,
-            use_native_endian,
-        );
-        header.checksum_1 = final_checksum.0;
-        header.checksum_2 = final_checksum.1;
-        tracing::trace!(
-            "begin_write_wal_frame(checksum=({}, {}))",
-            header.checksum_1,
-            header.checksum_2
-        );
+    frame[0..4].copy_from_slice(&page_number.to_be_bytes());
+    frame[4..8].copy_from_slice(&db_size.to_be_bytes());
+    frame[8..12].copy_from_slice(&wal_header.salt_1.to_be_bytes());
+    frame[12..16].copy_from_slice(&wal_header.salt_2.to_be_bytes());
 
-        buf[16..20].copy_from_slice(&header.checksum_1.to_be_bytes());
-        buf[20..24].copy_from_slice(&header.checksum_2.to_be_bytes());
+    let expects_be = wal_header.magic & 1;
+    let use_native_endian = cfg!(target_endian = "big") as u32 == expects_be;
+    let header_checksum = checksum_wal(&frame[0..8], wal_header, prev_checksums, use_native_endian);
+    let final_checksum = checksum_wal(
+        &frame[WAL_FRAME_HEADER_SIZE..WAL_FRAME_HEADER_SIZE + page_size as usize],
+        wal_header,
+        header_checksum,
+        use_native_endian,
+    );
+    frame[16..20].copy_from_slice(&final_checksum.0.to_be_bytes());
+    frame[20..24].copy_from_slice(&final_checksum.1.to_be_bytes());
 
-        #[allow(clippy::arc_with_non_send_sync)]
-        (Arc::new(RefCell::new(buffer)), final_checksum)
-    };
-
-    let clone_counter = write_counter.clone();
-    *write_counter.borrow_mut() += 1;
-    let write_complete = {
-        let buf_copy = buffer.clone();
-        Box::new(move |bytes_written: i32| {
-            let buf_copy = buf_copy.clone();
-            let buf_len = buf_copy.borrow().len();
-            *clone_counter.borrow_mut() -= 1;
-
-            page_finish.clear_dirty();
-            turso_assert!(
-                bytes_written == buf_len as i32,
-                "wrote({bytes_written}) != expected({buf_len})"
-            );
-        })
-    };
-    #[allow(clippy::arc_with_non_send_sync)]
-    let c = Completion::new(CompletionType::Write(WriteCompletion::new(write_complete)));
-    let res = io.pwrite(offset, buffer.clone(), c);
-    if res.is_err() {
-        // If we do not reduce the counter here on error, we incur an infinite loop when cacheflushing
-        *write_counter.borrow_mut() -= 1;
-    }
-    res?;
-    tracing::trace!("Frame written and synced");
-    Ok(checksums)
+    (final_checksum, Arc::new(RefCell::new(buffer)))
 }
 
 pub fn begin_write_wal_header(io: &Arc<dyn File>, header: &WalHeader) -> Result<()> {
@@ -1631,17 +1683,18 @@ pub fn begin_write_wal_header(io: &Arc<dyn File>, header: &WalHeader) -> Result<
         Arc::new(RefCell::new(buffer))
     };
 
-    let write_complete = {
-        Box::new(move |bytes_written: i32| {
-            turso_assert!(
-                bytes_written == WAL_HEADER_SIZE as i32,
-                "wal header wrote({bytes_written}) != expected({WAL_HEADER_SIZE})"
-            );
-        })
+    let cloned = buffer.clone();
+    let write_complete = move |bytes_written: i32| {
+        // make sure to reference buffer so it's alive for async IO
+        let _buf = cloned.borrow();
+        turso_assert!(
+            bytes_written == WAL_HEADER_SIZE as i32,
+            "wal header wrote({bytes_written}) != expected({WAL_HEADER_SIZE})"
+        );
     };
     #[allow(clippy::arc_with_non_send_sync)]
-    let c = Completion::new(CompletionType::Write(WriteCompletion::new(write_complete)));
-    io.pwrite(0, buffer.clone(), c)?;
+    let c = Completion::new_write(write_complete);
+    io.pwrite(0, buffer.clone(), c.into())?;
     Ok(())
 }
 
