@@ -17,12 +17,13 @@ use std::{
 };
 
 use crate::fast_lock::SpinLock;
-use crate::io::{CompletionType, File, SyncCompletion, IO};
+use crate::io::{File, IO};
 use crate::result::LimboResult;
 use crate::storage::sqlite3_ondisk::{
-    begin_read_wal_frame, begin_write_wal_frame, finish_read_page, WAL_FRAME_HEADER_SIZE,
-    WAL_HEADER_SIZE,
+    begin_read_wal_frame, begin_read_wal_frame_raw, finish_read_page, prepare_wal_frame,
+    WAL_FRAME_HEADER_SIZE, WAL_HEADER_SIZE,
 };
+use crate::types::IOResult;
 use crate::{turso_assert, Buffer, LimboError, Result};
 use crate::{Completion, Page};
 
@@ -38,7 +39,7 @@ pub const NO_LOCK: u32 = 0;
 pub const SHARED_LOCK: u32 = 1;
 pub const WRITE_LOCK: u32 = 2;
 
-#[derive(Debug, Copy, Clone)]
+#[derive(Debug, Copy, Clone, Default)]
 pub struct CheckpointResult {
     /// number of frames in WAL
     pub num_wal_frames: u64,
@@ -46,17 +47,11 @@ pub struct CheckpointResult {
     pub num_checkpointed_frames: u64,
 }
 
-impl Default for CheckpointResult {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
 impl CheckpointResult {
-    pub fn new() -> Self {
+    pub fn new(n_frames: u64, n_ckpt: u64) -> Self {
         Self {
-            num_wal_frames: 0,
-            num_checkpointed_frames: 0,
+            num_wal_frames: n_frames,
+            num_checkpointed_frames: n_ckpt,
         }
     }
 }
@@ -196,16 +191,16 @@ impl LimboRwLock {
 /// Write-ahead log (WAL).
 pub trait Wal {
     /// Begin a read transaction.
-    fn begin_read_tx(&mut self) -> Result<LimboResult>;
+    fn begin_read_tx(&mut self) -> Result<(LimboResult, bool)>;
 
     /// Begin a write transaction.
     fn begin_write_tx(&mut self) -> Result<LimboResult>;
 
     /// End a read transaction.
-    fn end_read_tx(&self) -> Result<LimboResult>;
+    fn end_read_tx(&self);
 
     /// End a write transaction.
-    fn end_write_tx(&self) -> Result<LimboResult>;
+    fn end_write_tx(&self);
 
     /// Find the latest frame containing a page.
     fn find_frame(&self, page_id: u64) -> Result<Option<u64>>;
@@ -213,14 +208,19 @@ pub trait Wal {
     /// Read a frame from the WAL.
     fn read_frame(&self, frame_id: u64, page: PageRef, buffer_pool: Arc<BufferPool>) -> Result<()>;
 
-    /// Read a frame from the WAL.
-    fn read_frame_raw(
-        &self,
-        frame_id: u64,
+    /// Read a raw frame (header included) from the WAL.
+    fn read_frame_raw(&self, frame_id: u64, frame: &mut [u8]) -> Result<Arc<Completion>>;
+
+    /// Write a raw frame (header included) from the WAL.
+    /// Note, that turso-db will use page_no and size_after fields from the header, but will overwrite checksum with proper value
+    fn write_frame_raw(
+        &mut self,
         buffer_pool: Arc<BufferPool>,
-        frame: *mut u8,
-        frame_len: u32,
-    ) -> Result<Arc<Completion>>;
+        frame_id: u64,
+        page_id: u64,
+        db_size: u64,
+        page: &[u8],
+    ) -> Result<()>;
 
     /// Write a frame to the WAL.
     /// db_size is the database size in pages after the transaction finishes.
@@ -244,8 +244,8 @@ pub trait Wal {
         pager: &Pager,
         write_counter: Rc<RefCell<usize>>,
         mode: CheckpointMode,
-    ) -> Result<CheckpointStatus>;
-    fn sync(&mut self) -> Result<WalFsyncStatus>;
+    ) -> Result<IOResult<CheckpointResult>>;
+    fn sync(&mut self) -> Result<IOResult<()>>;
     fn get_max_frame_in_wal(&self) -> u64;
     fn get_max_frame(&self) -> u64;
     fn get_min_frame(&self) -> u64;
@@ -259,21 +259,17 @@ pub trait Wal {
 pub struct DummyWAL;
 
 impl Wal for DummyWAL {
-    fn begin_read_tx(&mut self) -> Result<LimboResult> {
-        Ok(LimboResult::Ok)
+    fn begin_read_tx(&mut self) -> Result<(LimboResult, bool)> {
+        Ok((LimboResult::Ok, false))
     }
 
-    fn end_read_tx(&self) -> Result<LimboResult> {
-        Ok(LimboResult::Ok)
-    }
+    fn end_read_tx(&self) {}
 
     fn begin_write_tx(&mut self) -> Result<LimboResult> {
         Ok(LimboResult::Ok)
     }
 
-    fn end_write_tx(&self) -> Result<LimboResult> {
-        Ok(LimboResult::Ok)
-    }
+    fn end_write_tx(&self) {}
 
     fn find_frame(&self, _page_id: u64) -> Result<Option<u64>> {
         Ok(None)
@@ -288,13 +284,18 @@ impl Wal for DummyWAL {
         Ok(())
     }
 
-    fn read_frame_raw(
-        &self,
-        _frame_id: u64,
+    fn read_frame_raw(&self, _frame_id: u64, _frame: &mut [u8]) -> Result<Arc<Completion>> {
+        todo!();
+    }
+
+    fn write_frame_raw(
+        &mut self,
         _buffer_pool: Arc<BufferPool>,
-        _frame: *mut u8,
-        _frame_len: u32,
-    ) -> Result<Arc<Completion>> {
+        _frame_id: u64,
+        _page_id: u64,
+        _db_size: u64,
+        _page: &[u8],
+    ) -> Result<()> {
         todo!();
     }
 
@@ -316,14 +317,12 @@ impl Wal for DummyWAL {
         _pager: &Pager,
         _write_counter: Rc<RefCell<usize>>,
         _mode: crate::CheckpointMode,
-    ) -> Result<crate::CheckpointStatus> {
-        Ok(crate::CheckpointStatus::Done(
-            crate::CheckpointResult::default(),
-        ))
+    ) -> Result<IOResult<CheckpointResult>> {
+        Ok(IOResult::Done(CheckpointResult::default()))
     }
 
-    fn sync(&mut self) -> Result<crate::storage::wal::WalFsyncStatus> {
-        Ok(crate::storage::wal::WalFsyncStatus::Done)
+    fn sync(&mut self) -> Result<IOResult<()>> {
+        Ok(IOResult::Done(()))
     }
 
     fn get_max_frame_in_wal(&self) -> u64 {
@@ -366,18 +365,6 @@ pub enum CheckpointState {
     Done,
 }
 
-#[derive(Debug, Copy, Clone, PartialEq)]
-pub enum WalFsyncStatus {
-    Done,
-    IO,
-}
-
-#[derive(Debug, Copy, Clone)]
-pub enum CheckpointStatus {
-    Done(CheckpointResult),
-    IO,
-}
-
 // Checkpointing is a state machine that has multiple steps. Since there are multiple steps we save
 // in flight information of the checkpoint in OngoingCheckpoint. page is just a helper Page to do
 // page operations like reading a frame to a page, and writing a page to disk. This page should not
@@ -430,6 +417,9 @@ pub struct WalFile {
     /// Hack for now in case of rollback, will not be needed once we remove this bullshit frame cache.
     start_pages_in_frames: usize,
 
+    /// Count of possible pages to checkpoint, and number of backfilled
+    prev_checkpoint: CheckpointResult,
+
     /// Private copy of WalHeader
     pub header: WalHeader,
 }
@@ -500,10 +490,12 @@ impl fmt::Debug for WalFileShared {
 
 impl Wal for WalFile {
     /// Begin a read transaction.
-    #[instrument(skip_all, level = Level::INFO)]
-    fn begin_read_tx(&mut self) -> Result<LimboResult> {
+    #[instrument(skip_all, level = Level::DEBUG)]
+    fn begin_read_tx(&mut self) -> Result<(LimboResult, bool)> {
         let max_frame_in_wal = self.get_shared().max_frame.load(Ordering::SeqCst);
 
+        let db_has_changed = max_frame_in_wal > self.max_frame;
+
         let mut max_read_mark = 0;
         let mut max_read_mark_index = -1;
         // Find the largest mark we can find, ignore frames that are impossible to be in range and
@@ -532,7 +524,7 @@ impl Wal for WalFile {
         }
 
         if max_read_mark_index == -1 {
-            return Ok(LimboResult::Busy);
+            return Ok((LimboResult::Busy, db_has_changed));
         }
 
         let (min_frame, last_checksum, start_pages_in_frames) = {
@@ -541,7 +533,7 @@ impl Wal for WalFile {
             tracing::trace!("begin_read_tx_read_lock(lock={})", max_read_mark_index);
             let busy = !lock.read();
             if busy {
-                return Ok(LimboResult::Busy);
+                return Ok((LimboResult::Busy, db_has_changed));
             }
             (
                 shared.nbackfills.load(Ordering::SeqCst) + 1,
@@ -561,40 +553,47 @@ impl Wal for WalFile {
             self.max_frame_read_lock_index,
             max_frame_in_wal
         );
-        Ok(LimboResult::Ok)
+        Ok((LimboResult::Ok, db_has_changed))
     }
 
     /// End a read transaction.
     #[inline(always)]
-    #[instrument(skip_all, level = Level::INFO)]
-    fn end_read_tx(&self) -> Result<LimboResult> {
+    #[instrument(skip_all, level = Level::DEBUG)]
+    fn end_read_tx(&self) {
         tracing::debug!("end_read_tx(lock={})", self.max_frame_read_lock_index);
         let read_lock = &mut self.get_shared().read_locks[self.max_frame_read_lock_index];
         read_lock.unlock();
-        Ok(LimboResult::Ok)
     }
 
     /// Begin a write transaction
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn begin_write_tx(&mut self) -> Result<LimboResult> {
         let busy = !self.get_shared().write_lock.write();
         tracing::debug!("begin_write_transaction(busy={})", busy);
         if busy {
             return Ok(LimboResult::Busy);
         }
+        // If the max frame is not the same as the one in the shared state, it means another
+        // transaction wrote to the WAL after we started our read transaction. This means our
+        // snapshot is not consistent with the one in the shared state and we need to start another
+        // one.
+        let shared = self.get_shared();
+        if self.max_frame != shared.max_frame.load(Ordering::SeqCst) {
+            shared.write_lock.unlock();
+            return Ok(LimboResult::Busy);
+        }
         Ok(LimboResult::Ok)
     }
 
     /// End a write transaction
-    #[instrument(skip_all, level = Level::INFO)]
-    fn end_write_tx(&self) -> Result<LimboResult> {
+    #[instrument(skip_all, level = Level::DEBUG)]
+    fn end_write_tx(&self) {
         tracing::debug!("end_write_txn");
         self.get_shared().write_lock.unlock();
-        Ok(LimboResult::Ok)
     }
 
     /// Find the latest frame containing a page.
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn find_frame(&self, page_id: u64) -> Result<Option<u64>> {
         let shared = self.get_shared();
         let frames = shared.frame_cache.lock();
@@ -612,7 +611,7 @@ impl Wal for WalFile {
     }
 
     /// Read a frame from the WAL.
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn read_frame(&self, frame_id: u64, page: PageRef, buffer_pool: Arc<BufferPool>) -> Result<()> {
         tracing::debug!("read_frame({})", frame_id);
         let offset = self.frame_offset(frame_id);
@@ -636,16 +635,11 @@ impl Wal for WalFile {
         Ok(())
     }
 
-    #[instrument(skip_all, level = Level::INFO)]
-    fn read_frame_raw(
-        &self,
-        frame_id: u64,
-        buffer_pool: Arc<BufferPool>,
-        frame: *mut u8,
-        frame_len: u32,
-    ) -> Result<Arc<Completion>> {
+    #[instrument(skip_all, level = Level::DEBUG)]
+    fn read_frame_raw(&self, frame_id: u64, frame: &mut [u8]) -> Result<Arc<Completion>> {
         tracing::debug!("read_frame({})", frame_id);
         let offset = self.frame_offset(frame_id);
+        let (frame_ptr, frame_len) = (frame.as_mut_ptr(), frame.len());
         let complete = Box::new(move |buf: Arc<RefCell<Buffer>>, bytes_read: i32| {
             let buf = buf.borrow();
             let buf_len = buf.len();
@@ -655,20 +649,100 @@ impl Wal for WalFile {
             );
             let buf_ptr = buf.as_ptr();
             unsafe {
-                std::ptr::copy_nonoverlapping(buf_ptr, frame, frame_len as usize);
+                std::ptr::copy_nonoverlapping(buf_ptr, frame_ptr, frame_len);
             }
         });
-        let c = begin_read_wal_frame(
-            &self.get_shared().file,
-            offset + WAL_FRAME_HEADER_SIZE,
-            buffer_pool,
-            complete,
-        )?;
+        let c =
+            begin_read_wal_frame_raw(&self.get_shared().file, offset, self.page_size(), complete)?;
         Ok(c)
     }
 
+    #[instrument(skip_all, level = Level::DEBUG)]
+    fn write_frame_raw(
+        &mut self,
+        buffer_pool: Arc<BufferPool>,
+        frame_id: u64,
+        page_id: u64,
+        db_size: u64,
+        page: &[u8],
+    ) -> Result<()> {
+        tracing::debug!("write_raw_frame({})", frame_id);
+        if page.len() != self.page_size() as usize {
+            return Err(LimboError::InvalidArgument(format!(
+                "unexpected page size in frame: got={}, expected={}",
+                page.len(),
+                self.page_size(),
+            )));
+        }
+        if frame_id > self.max_frame + 1 {
+            // attempt to write frame out of sequential order - error out
+            return Err(LimboError::InvalidArgument(format!(
+                "frame_id is beyond next frame in the WAL: frame_id={}, max_frame={}",
+                frame_id, self.max_frame
+            )));
+        }
+        if frame_id <= self.max_frame {
+            // just validate if page content from the frame matches frame in the WAL
+            let offset = self.frame_offset(frame_id);
+            let conflict = Arc::new(Cell::new(false));
+            let (page_ptr, page_len) = (page.as_ptr(), page.len());
+            let complete = Box::new({
+                let conflict = conflict.clone();
+                move |buf: Arc<RefCell<Buffer>>, bytes_read: i32| {
+                    let buf = buf.borrow();
+                    let buf_len = buf.len();
+                    turso_assert!(
+                        bytes_read == buf_len as i32,
+                        "read({bytes_read}) != expected({buf_len})"
+                    );
+                    let page = unsafe { std::slice::from_raw_parts(page_ptr, page_len) };
+                    if buf.as_slice() != page {
+                        conflict.set(true);
+                    }
+                }
+            });
+            let c = begin_read_wal_frame(
+                &self.get_shared().file,
+                offset + WAL_FRAME_HEADER_SIZE,
+                buffer_pool,
+                complete,
+            )?;
+            self.io.wait_for_completion(c)?;
+            return if conflict.get() {
+                Err(LimboError::Conflict(format!(
+                    "frame content differs from the WAL: frame_id={frame_id}"
+                )))
+            } else {
+                Ok(())
+            };
+        }
+
+        // perform actual write
+        let offset = self.frame_offset(frame_id);
+        let shared = self.get_shared();
+        let header = shared.wal_header.clone();
+        let header = header.lock();
+        let checksums = self.last_checksum;
+        let (checksums, frame_bytes) = prepare_wal_frame(
+            &header,
+            checksums,
+            header.page_size,
+            page_id as u32,
+            db_size as u32,
+            page,
+        );
+        let c = Arc::new(Completion::new_write(|_| {}));
+        let c = shared.file.pwrite(offset, frame_bytes, c)?;
+        self.io.wait_for_completion(c)?;
+        self.complete_append_frame(page_id, frame_id, checksums);
+        if db_size > 0 {
+            self.finish_append_frames_commit()?;
+        }
+        Ok(())
+    }
+
     /// Write a frame to the WAL.
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn append_frame(
         &mut self,
         page: PageRef,
@@ -676,8 +750,7 @@ impl Wal for WalFile {
         write_counter: Rc<RefCell<usize>>,
     ) -> Result<()> {
         let page_id = page.get().id;
-        let max_frame = self.max_frame;
-        let frame_id = if max_frame == 0 { 1 } else { max_frame + 1 };
+        let frame_id = self.max_frame + 1;
         let offset = self.frame_offset(frame_id);
         tracing::debug!(frame_id, offset, page_id);
         let checksums = {
@@ -685,48 +758,58 @@ impl Wal for WalFile {
             let header = shared.wal_header.clone();
             let header = header.lock();
             let checksums = self.last_checksum;
-            begin_write_wal_frame(
-                &shared.file,
-                offset,
-                &page,
-                header.page_size as u16,
-                db_size,
-                write_counter,
+            let page_content = page.get_contents();
+            let page_buf = page_content.as_ptr();
+            let (frame_checksums, frame_bytes) = prepare_wal_frame(
                 &header,
                 checksums,
-            )?
-        };
-        self.last_checksum = checksums;
-        self.max_frame = frame_id;
-        let shared = self.get_shared();
-        {
-            let mut frame_cache = shared.frame_cache.lock();
-            let frames = frame_cache.get_mut(&(page_id as u64));
-            match frames {
-                Some(frames) => frames.push(frame_id),
-                None => {
-                    frame_cache.insert(page_id as u64, vec![frame_id]);
-                    shared.pages_in_frames.lock().push(page_id as u64);
+                header.page_size,
+                page_id as u32,
+                db_size,
+                page_buf,
+            );
+
+            *write_counter.borrow_mut() += 1;
+            let c = Completion::new_write({
+                let frame_bytes = frame_bytes.clone();
+                let write_counter = write_counter.clone();
+                move |bytes_written| {
+                    let frame_len = frame_bytes.borrow().len();
+                    turso_assert!(
+                        bytes_written == frame_len as i32,
+                        "wrote({bytes_written}) != expected({frame_len})"
+                    );
+
+                    page.clear_dirty();
+                    *write_counter.borrow_mut() -= 1;
                 }
+            });
+            let result = shared.file.pwrite(offset, frame_bytes.clone(), c.into());
+            if let Err(err) = result {
+                *write_counter.borrow_mut() -= 1;
+                return Err(err);
             }
-        }
+            frame_checksums
+        };
+        self.complete_append_frame(page_id as u64, frame_id, checksums);
         Ok(())
     }
 
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn should_checkpoint(&self) -> bool {
         let shared = self.get_shared();
         let frame_id = shared.max_frame.load(Ordering::SeqCst) as usize;
-        frame_id >= self.checkpoint_threshold
+        let nbackfills = shared.nbackfills.load(Ordering::SeqCst) as usize;
+        frame_id > self.checkpoint_threshold + nbackfills
     }
 
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn checkpoint(
         &mut self,
         pager: &Pager,
         write_counter: Rc<RefCell<usize>>,
         mode: CheckpointMode,
-    ) -> Result<CheckpointStatus> {
+    ) -> Result<IOResult<CheckpointResult>> {
         assert!(
             matches!(mode, CheckpointMode::Passive),
             "only passive mode supported for now"
@@ -737,7 +820,19 @@ impl Wal for WalFile {
             match state {
                 CheckpointState::Start => {
                     // TODO(pere): check what frames are safe to checkpoint between many readers!
-                    self.ongoing_checkpoint.min_frame = self.min_frame;
+                    let (shared_max, nbackfills) = {
+                        let shared = self.get_shared();
+                        (
+                            shared.max_frame.load(Ordering::SeqCst),
+                            shared.nbackfills.load(Ordering::SeqCst),
+                        )
+                    };
+                    if shared_max <= nbackfills {
+                        // if there's nothing to do and we are fully back-filled, to match sqlite
+                        // we return the previous number of backfilled pages from last checkpoint.
+                        return Ok(IOResult::Done(self.prev_checkpoint));
+                    }
+                    self.ongoing_checkpoint.min_frame = nbackfills + 1;
                     let shared = self.get_shared();
                     let busy = !shared.checkpoint_lock.write();
                     if busy {
@@ -812,7 +907,7 @@ impl Wal for WalFile {
                 }
                 CheckpointState::WaitReadFrame => {
                     if self.ongoing_checkpoint.page.is_locked() {
-                        return Ok(CheckpointStatus::IO);
+                        return Ok(IOResult::IO);
                     } else {
                         self.ongoing_checkpoint.state = CheckpointState::WritePage;
                     }
@@ -828,7 +923,7 @@ impl Wal for WalFile {
                 }
                 CheckpointState::WaitWritePage => {
                     if *write_counter.borrow() > 0 {
-                        return Ok(CheckpointStatus::IO);
+                        return Ok(IOResult::IO);
                     }
                     // If page was in cache clear it.
                     if let Some(page) = pager.cache_get(self.ongoing_checkpoint.page.get().id) {
@@ -847,19 +942,27 @@ impl Wal for WalFile {
                 }
                 CheckpointState::Done => {
                     if *write_counter.borrow() > 0 {
-                        return Ok(CheckpointStatus::IO);
+                        return Ok(IOResult::IO);
                     }
                     let shared = self.get_shared();
                     shared.checkpoint_lock.unlock();
+                    let max_frame = shared.max_frame.load(Ordering::SeqCst);
+                    let nbackfills = shared.nbackfills.load(Ordering::SeqCst);
 
                     // Record two num pages fields to return as checkpoint result to caller.
                     // Ref: pnLog, pnCkpt on https://www.sqlite.org/c3ref/wal_checkpoint_v2.html
-                    let checkpoint_result = CheckpointResult {
-                        num_wal_frames: shared.max_frame.load(Ordering::SeqCst),
-                        num_checkpointed_frames: self.ongoing_checkpoint.max_frame,
-                    };
+                    let frames_in_wal = max_frame.saturating_sub(nbackfills);
+                    let frames_checkpointed = self
+                        .ongoing_checkpoint
+                        .max_frame
+                        .saturating_sub(self.ongoing_checkpoint.min_frame - 1);
+                    let checkpoint_result =
+                        CheckpointResult::new(frames_in_wal, frames_checkpointed);
                     let everything_backfilled = shared.max_frame.load(Ordering::SeqCst)
                         == self.ongoing_checkpoint.max_frame;
+                    shared
+                        .nbackfills
+                        .store(self.ongoing_checkpoint.max_frame, Ordering::SeqCst);
                     if everything_backfilled {
                         // TODO: Even in Passive mode, if everything was backfilled we should
                         // truncate and fsync the *db file*
@@ -877,43 +980,38 @@ impl Wal for WalFile {
                             // TODO: if all frames were backfilled into the db file, calls fsync
                             // TODO(pere): truncate wal file here.
                         }
-                    } else {
-                        shared
-                            .nbackfills
-                            .store(self.ongoing_checkpoint.max_frame, Ordering::SeqCst);
                     }
+                    self.prev_checkpoint = checkpoint_result;
                     self.ongoing_checkpoint.state = CheckpointState::Start;
-                    return Ok(CheckpointStatus::Done(checkpoint_result));
+                    return Ok(IOResult::Done(checkpoint_result));
                 }
             }
         }
     }
 
-    #[instrument(err, skip_all, level = Level::INFO)]
-    fn sync(&mut self) -> Result<WalFsyncStatus> {
+    #[instrument(err, skip_all, level = Level::DEBUG)]
+    fn sync(&mut self) -> Result<IOResult<()>> {
         match self.sync_state.get() {
             SyncState::NotSyncing => {
                 tracing::debug!("wal_sync");
                 let syncing = self.syncing.clone();
                 self.syncing.set(true);
-                let completion = Completion::new(CompletionType::Sync(SyncCompletion {
-                    complete: Box::new(move |_| {
-                        tracing::debug!("wal_sync finish");
-                        syncing.set(false);
-                    }),
-                }));
+                let completion = Completion::new_sync(move |_| {
+                    tracing::debug!("wal_sync finish");
+                    syncing.set(false);
+                });
                 let shared = self.get_shared();
-                shared.file.sync(completion)?;
+                shared.file.sync(completion.into())?;
                 self.sync_state.set(SyncState::Syncing);
-                Ok(WalFsyncStatus::IO)
+                Ok(IOResult::IO)
             }
             SyncState::Syncing => {
                 if self.syncing.get() {
                     tracing::debug!("wal_sync is already syncing");
-                    Ok(WalFsyncStatus::IO)
+                    Ok(IOResult::IO)
                 } else {
                     self.sync_state.set(SyncState::NotSyncing);
-                    Ok(WalFsyncStatus::Done)
+                    Ok(IOResult::Done(()))
                 }
             }
         }
@@ -931,7 +1029,7 @@ impl Wal for WalFile {
         self.min_frame
     }
 
-    #[instrument(err, skip_all, level = Level::INFO)]
+    #[instrument(err, skip_all, level = Level::DEBUG)]
     fn rollback(&mut self) -> Result<()> {
         // TODO(pere): have to remove things from frame_cache because they are no longer valid.
         // TODO(pere): clear page cache in pager.
@@ -940,24 +1038,28 @@ impl Wal for WalFile {
             let shared = self.get_shared();
             let max_frame = shared.max_frame.load(Ordering::SeqCst);
             tracing::debug!(to_max_frame = max_frame);
-            let mut frame_cache = shared.frame_cache.lock();
-            for (_, frames) in frame_cache.iter_mut() {
-                let mut last_valid_frame = frames.len();
-                for frame in frames.iter().rev() {
-                    if *frame <= max_frame {
-                        break;
+            {
+                let mut frame_cache = shared.frame_cache.lock();
+                for (_, frames) in frame_cache.iter_mut() {
+                    let mut last_valid_frame = frames.len();
+                    for frame in frames.iter().rev() {
+                        if *frame <= max_frame {
+                            break;
+                        }
+                        last_valid_frame -= 1;
                     }
-                    last_valid_frame -= 1;
+                    frames.truncate(last_valid_frame);
                 }
-                frames.truncate(last_valid_frame);
+                let mut pages_in_frames = shared.pages_in_frames.lock();
+                pages_in_frames.truncate(self.start_pages_in_frames);
             }
-            let mut pages_in_frames = shared.pages_in_frames.lock();
-            pages_in_frames.truncate(self.start_pages_in_frames);
+            self.last_checksum = shared.last_checksum;
         }
+        self.reset_internal_states();
         Ok(())
     }
 
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     fn finish_append_frames_commit(&mut self) -> Result<()> {
         let shared = self.get_shared();
         shared.max_frame.store(self.max_frame, Ordering::SeqCst);
@@ -1007,6 +1109,7 @@ impl WalFile {
             min_frame: 0,
             max_frame_read_lock_index: 0,
             last_checksum,
+            prev_checkpoint: CheckpointResult::default(),
             start_pages_in_frames: 0,
             header: *header,
         }
@@ -1027,6 +1130,32 @@ impl WalFile {
     fn get_shared(&self) -> &mut WalFileShared {
         unsafe { self.shared.get().as_mut().unwrap() }
     }
+
+    fn complete_append_frame(&mut self, page_id: u64, frame_id: u64, checksums: (u32, u32)) {
+        self.last_checksum = checksums;
+        self.max_frame = frame_id;
+        let shared = self.get_shared();
+        {
+            let mut frame_cache = shared.frame_cache.lock();
+            let frames = frame_cache.get_mut(&page_id);
+            match frames {
+                Some(frames) => frames.push(frame_id),
+                None => {
+                    frame_cache.insert(page_id, vec![frame_id]);
+                    shared.pages_in_frames.lock().push(page_id);
+                }
+            }
+        }
+    }
+
+    fn reset_internal_states(&mut self) {
+        self.ongoing_checkpoint.state = CheckpointState::Start;
+        self.ongoing_checkpoint.min_frame = 0;
+        self.ongoing_checkpoint.max_frame = 0;
+        self.ongoing_checkpoint.current_page = 0;
+        self.sync_state.set(SyncState::NotSyncing);
+        self.syncing.set(false);
+    }
 }
 
 impl WalFileShared {
