@@ -1,7 +1,7 @@
 // This module contains code for emitting bytecode instructions for SQL query execution.
 // It handles translating high-level SQL operations into low-level bytecode that can be executed by the virtual machine.
 
-use std::rc::Rc;
+use std::sync::Arc;
 
 use tracing::{instrument, Level};
 use turso_sqlite3_parser::ast::{self, Expr};
@@ -24,6 +24,7 @@ use crate::error::SQLITE_CONSTRAINT_PRIMARYKEY;
 use crate::function::Func;
 use crate::schema::{Schema, Table};
 use crate::translate::compound_select::emit_program_for_compound_select;
+use crate::translate::expr::{emit_returning_results, ReturningValueRegisters};
 use crate::translate::plan::{DeletePlan, Plan, QueryDestination, Search};
 use crate::translate::values::emit_values;
 use crate::util::exprs_are_equivalent;
@@ -201,7 +202,7 @@ pub enum TransactionMode {
 
 /// Main entry point for emitting bytecode for a SQL query
 /// Takes a query plan and generates the corresponding bytecode program
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 pub fn emit_program(
     program: &mut ProgramBuilder,
     plan: Plan,
@@ -219,7 +220,7 @@ pub fn emit_program(
     }
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 fn emit_program_for_select(
     program: &mut ProgramBuilder,
     mut plan: SelectPlan,
@@ -258,7 +259,7 @@ fn emit_program_for_select(
     Ok(())
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 pub fn emit_query<'a>(
     program: &mut ProgramBuilder,
     plan: &'a mut SelectPlan,
@@ -398,7 +399,7 @@ pub fn emit_query<'a>(
     Ok(t_ctx.reg_result_cols_start.unwrap())
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 fn emit_program_for_delete(
     program: &mut ProgramBuilder,
     plan: DeletePlan,
@@ -453,7 +454,12 @@ fn emit_program_for_delete(
         None,
     )?;
 
-    emit_delete_insns(program, &mut t_ctx, &plan.table_references)?;
+    emit_delete_insns(
+        program,
+        &mut t_ctx,
+        &plan.table_references,
+        &plan.result_columns,
+    )?;
 
     // Clean up and close the main execution loop
     close_loop(
@@ -476,8 +482,16 @@ fn emit_delete_insns(
     program: &mut ProgramBuilder,
     t_ctx: &mut TranslateCtx,
     table_references: &TableReferences,
+    result_columns: &[super::plan::ResultSetColumn],
 ) -> Result<()> {
     let table_reference = table_references.joined_tables().first().unwrap();
+    if table_reference
+        .virtual_table()
+        .is_some_and(|t| t.readonly())
+    {
+        return Err(crate::LimboError::ReadOnly);
+    }
+
     let cursor_id = match &table_reference.op {
         Operation::Scan { .. } => {
             program.resolve_cursor_id(&CursorKey::table(table_reference.internal_id))
@@ -600,6 +614,33 @@ fn emit_delete_insns(
             )?;
         }
 
+        // Emit RETURNING results if specified (must be before DELETE)
+        if !result_columns.is_empty() {
+            // Get rowid for RETURNING
+            let rowid_reg = program.alloc_register();
+            program.emit_insn(Insn::RowId {
+                cursor_id: main_table_cursor_id,
+                dest: rowid_reg,
+            });
+
+            // Allocate registers for column values
+            let columns_start_reg = program.alloc_registers(table_reference.columns().len());
+
+            // Read all column values from the row to be deleted
+            for (i, _column) in table_reference.columns().iter().enumerate() {
+                program.emit_column(main_table_cursor_id, i, columns_start_reg + i);
+            }
+
+            // Emit RETURNING results using the values we just read
+            let value_registers = ReturningValueRegisters {
+                rowid_register: rowid_reg,
+                columns_start_register: columns_start_reg,
+                num_columns: table_reference.columns().len(),
+            };
+
+            emit_returning_results(program, result_columns, &value_registers)?;
+        }
+
         program.emit_insn(Insn::Delete {
             cursor_id: main_table_cursor_id,
         });
@@ -614,7 +655,7 @@ fn emit_delete_insns(
     Ok(())
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 fn emit_program_for_update(
     program: &mut ProgramBuilder,
     mut plan: UpdatePlan,
@@ -692,7 +733,7 @@ fn emit_program_for_update(
             program.emit_insn(Insn::OpenWrite {
                 cursor_id: cursor,
                 root_page: RegisterOrLiteral::Literal(index.root_page),
-                name: index.name.clone(),
+                db: 0,
             });
             cursor
         };
@@ -733,7 +774,7 @@ fn emit_program_for_update(
     Ok(())
 }
 
-#[instrument(skip_all, level = Level::INFO)]
+#[instrument(skip_all, level = Level::DEBUG)]
 fn emit_update_insns(
     plan: &UpdatePlan,
     t_ctx: &TranslateCtx,
@@ -824,7 +865,7 @@ fn emit_update_insns(
         program.emit_insn(Insn::Copy {
             src_reg: beg,
             dst_reg: beg + 1,
-            amount: 0,
+            extra_amount: 0,
         })
     }
 
@@ -941,14 +982,14 @@ fn emit_update_insns(
             program.emit_insn(Insn::Copy {
                 src_reg: idx_cols_start_reg + col.pos_in_table,
                 dst_reg: idx_start_reg + i,
-                amount: 0,
+                extra_amount: 0,
             });
         }
         // last register is the rowid
         program.emit_insn(Insn::Copy {
             src_reg: rowid_reg,
             dst_reg: idx_start_reg + num_cols,
-            amount: 0,
+            extra_amount: 0,
         });
 
         // this record will be inserted into the index later
@@ -1014,7 +1055,7 @@ fn emit_update_insns(
                 start_reg: start,
                 count: table_ref.columns().len(),
                 check_generated: true,
-                table_reference: Rc::clone(&btree_table),
+                table_reference: Arc::clone(&btree_table),
             });
         }
 
@@ -1153,10 +1194,29 @@ fn emit_update_insns(
             cursor: cursor_id,
             key_reg: rowid_set_clause_reg.unwrap_or(beg),
             record_reg,
-            flag: InsertFlags::new().update(true),
+            flag: if has_user_provided_rowid {
+                // The previous Insn::NotExists and Insn::Delete seek to the old rowid,
+                // so to insert a new user-provided rowid, we need to seek to the correct place.
+                InsertFlags::new().require_seek()
+            } else {
+                InsertFlags::new()
+            },
             table_name: table_ref.identifier.clone(),
         });
 
+        // Emit RETURNING results if specified
+        if let Some(returning_columns) = &plan.returning {
+            if !returning_columns.is_empty() {
+                let value_registers = ReturningValueRegisters {
+                    rowid_register: rowid_set_clause_reg.unwrap_or(beg),
+                    columns_start_register: start,
+                    num_columns: table_ref.columns().len(),
+                };
+
+                emit_returning_results(program, returning_columns, &value_registers)?;
+            }
+        }
+
         // create full CDC record after update if necessary
         let cdc_after_reg = if program.capture_data_changes_mode().has_after() {
             Some(emit_cdc_patch_record(
@@ -1247,7 +1307,7 @@ pub fn emit_cdc_patch_record(
         program.emit_insn(Insn::Copy {
             src_reg: rowid_reg,
             dst_reg: columns_reg + rowid_alias_position,
-            amount: 0,
+            extra_amount: 0,
         });
         program.emit_insn(Insn::MakeRecord {
             start_reg: columns_reg,
@@ -1274,7 +1334,7 @@ pub fn emit_cdc_full_record(
             program.emit_insn(Insn::Copy {
                 src_reg: rowid_reg,
                 dst_reg: columns_reg + 1 + i,
-                amount: 0,
+                extra_amount: 0,
             });
         } else {
             program.emit_column(table_cursor_id, i, columns_reg + 1 + i);
@@ -1300,7 +1360,7 @@ pub fn emit_cdc_insns(
     after_record_reg: Option<usize>,
     table_name: &str,
 ) -> Result<()> {
-    // (operation_id INTEGER PRIMARY KEY AUTOINCREMENT, operation_time INTEGER, operation_type INTEGER, table_name TEXT, id, before BLOB, after BLOB)
+    // (change_id INTEGER PRIMARY KEY AUTOINCREMENT, change_time INTEGER, change_type INTEGER, table_name TEXT, id, before BLOB, after BLOB)
     let turso_cdc_registers = program.alloc_registers(7);
     program.emit_insn(Insn::Null {
         dest: turso_cdc_registers,
@@ -1323,12 +1383,12 @@ pub fn emit_cdc_insns(
         func: unixepoch_fn_ctx,
     });
 
-    let operation_type = match operation_mode {
+    let change_type = match operation_mode {
         OperationMode::INSERT => 1,
         OperationMode::UPDATE | OperationMode::SELECT => 0,
         OperationMode::DELETE => -1,
     };
-    program.emit_int(operation_type, turso_cdc_registers + 2);
+    program.emit_int(change_type, turso_cdc_registers + 2);
     program.mark_last_insn_constant();
 
     program.emit_string8(table_name.to_string(), turso_cdc_registers + 3);
@@ -1337,14 +1397,14 @@ pub fn emit_cdc_insns(
     program.emit_insn(Insn::Copy {
         src_reg: rowid_reg,
         dst_reg: turso_cdc_registers + 4,
-        amount: 0,
+        extra_amount: 0,
     });
 
     if let Some(before_record_reg) = before_record_reg {
         program.emit_insn(Insn::Copy {
             src_reg: before_record_reg,
             dst_reg: turso_cdc_registers + 5,
-            amount: 0,
+            extra_amount: 0,
         });
     } else {
         program.emit_null(turso_cdc_registers + 5, None);
@@ -1355,7 +1415,7 @@ pub fn emit_cdc_insns(
         program.emit_insn(Insn::Copy {
             src_reg: after_record_reg,
             dst_reg: turso_cdc_registers + 6,
-            amount: 0,
+            extra_amount: 0,
         });
     } else {
         program.emit_null(turso_cdc_registers + 6, None);
