@@ -1,12 +1,17 @@
-use std::rc::Rc;
+use std::sync::Arc;
 
 use turso_sqlite3_parser::ast::{
     DistinctNames, Expr, InsertBody, OneSelect, QualifiedName, ResolveType, ResultColumn, With,
 };
 
 use crate::error::{SQLITE_CONSTRAINT_NOTNULL, SQLITE_CONSTRAINT_PRIMARYKEY};
-use crate::schema::{IndexColumn, Table};
+use crate::schema::{self, IndexColumn, Table};
 use crate::translate::emitter::{emit_cdc_insns, emit_cdc_patch_record, OperationMode};
+use crate::translate::expr::{
+    emit_returning_results, process_returning_clause, ReturningValueRegisters,
+};
+use crate::translate::plan::TableReferences;
+use crate::translate::planner::ROWID;
 use crate::util::normalize_ident;
 use crate::vdbe::builder::ProgramBuilderOpts;
 use crate::vdbe::insn::{IdxInsertFlags, InsertFlags, RegisterOrLiteral};
@@ -40,9 +45,10 @@ pub fn translate_insert(
     tbl_name: QualifiedName,
     columns: Option<DistinctNames>,
     mut body: InsertBody,
-    _returning: Option<Vec<ResultColumn>>,
+    mut returning: Option<Vec<ResultColumn>>,
     syms: &SymbolTable,
     mut program: ProgramBuilder,
+    connection: &Arc<crate::Connection>,
 ) -> Result<ProgramBuilder> {
     let opts = ProgramBuilderOpts {
         num_cursors: 1,
@@ -65,7 +71,7 @@ pub fn translate_insert(
         );
     }
     let table_name = &tbl_name.name;
-    let table = match schema.get_table(table_name.0.as_str()) {
+    let table = match schema.get_table(table_name.as_str()) {
         Some(table) => table,
         None => crate::bail_parse_error!("no such table: {}", table_name),
     };
@@ -137,6 +143,24 @@ pub fn translate_insert(
         None
     };
 
+    // Process RETURNING clause using shared module
+    let (result_columns, _) = if let Some(returning) = &mut returning {
+        process_returning_clause(
+            returning,
+            &table,
+            table_name.as_str(),
+            &mut program,
+            connection,
+        )?
+    } else {
+        (vec![], TableReferences::new(vec![], vec![]))
+    };
+
+    // Set up the program to return result columns if RETURNING is specified
+    if !result_columns.is_empty() {
+        program.result_columns = result_columns.clone();
+    }
+
     let mut yield_reg_opt = None;
     let mut temp_table_ctx = None;
     let (num_values, cursor_id) = match body {
@@ -167,7 +191,14 @@ pub fn translate_insert(
                     coroutine_implementation_start: halt_label,
                 };
                 program.incr_nesting();
-                let result = translate_select(schema, *select, syms, program, query_destination)?;
+                let result = translate_select(
+                    schema,
+                    *select,
+                    syms,
+                    program,
+                    query_destination,
+                    connection,
+                )?;
                 program = result.program;
                 program.decr_nesting();
 
@@ -231,7 +262,8 @@ pub fn translate_insert(
                         cursor: temp_cursor_id,
                         key_reg: rowid_reg,
                         record_reg,
-                        flag: InsertFlags::new(),
+                        // since we are not doing an Insn::NewRowid or an Insn::NotExists here, we need to seek to ensure the insertion happens in the correct place.
+                        flag: InsertFlags::new().require_seek(),
                         table_name: "".to_string(),
                     });
 
@@ -245,13 +277,13 @@ pub fn translate_insert(
                     program.emit_insn(Insn::OpenWrite {
                         cursor_id,
                         root_page: RegisterOrLiteral::Literal(root_page),
-                        name: table_name.0.clone(),
+                        db: 0,
                     });
                 } else {
                     program.emit_insn(Insn::OpenWrite {
                         cursor_id,
                         root_page: RegisterOrLiteral::Literal(root_page),
-                        name: table_name.0.clone(),
+                        db: 0,
                     });
 
                     // Main loop
@@ -277,7 +309,7 @@ pub fn translate_insert(
     // allocate cursor id's for each btree index cursor we'll need to populate the indexes
     // (idx name, root_page, idx cursor id)
     let idx_cursors = schema
-        .get_indices(&table_name.0)
+        .get_indices(table_name.as_str())
         .iter()
         .map(|idx| {
             (
@@ -289,29 +321,15 @@ pub fn translate_insert(
         .collect::<Vec<(&String, usize, usize)>>();
 
     let column_mappings = resolve_columns_for_insert(&table, &columns, num_values)?;
-    // Check if rowid was provided (through INTEGER PRIMARY KEY as a rowid alias)
-    let rowid_alias_index = btree_table.columns.iter().position(|c| c.is_rowid_alias);
-    let has_user_provided_rowid = {
-        assert_eq!(column_mappings.len(), btree_table.columns.len());
-        if let Some(index) = rowid_alias_index {
-            column_mappings[index].value_index.is_some()
-        } else {
-            false
-        }
-    };
+    let has_user_provided_rowid = column_mappings
+        .iter()
+        .any(|x| x.value_index.is_some() && x.column.is_rowid_alias);
 
     // allocate a register for each column in the table. if not provided by user, they will simply be set as null.
     // allocate an extra register for rowid regardless of whether user provided a rowid alias column.
     let num_cols = btree_table.columns.len();
-    let rowid_reg = program.alloc_registers(num_cols + 1);
-    let column_registers_start = rowid_reg + 1;
-    let rowid_alias_reg = {
-        if has_user_provided_rowid {
-            Some(column_registers_start + rowid_alias_index.unwrap())
-        } else {
-            None
-        }
-    };
+    let rowid_and_columns_start_register = program.alloc_registers(num_cols + 1);
+    let columns_start_register = rowid_and_columns_start_register + 1;
 
     let record_register = program.alloc_register();
 
@@ -327,7 +345,7 @@ pub fn translate_insert(
         populate_columns_multiple_rows(
             &mut program,
             &column_mappings,
-            column_registers_start,
+            rowid_and_columns_start_register,
             yield_reg_opt.unwrap() + 1,
             &resolver,
             &temp_table_ctx,
@@ -337,15 +355,14 @@ pub fn translate_insert(
         program.emit_insn(Insn::OpenWrite {
             cursor_id,
             root_page: RegisterOrLiteral::Literal(root_page),
-            name: table_name.0.clone(),
+            db: 0,
         });
 
         populate_column_registers(
             &mut program,
             &values.unwrap(),
             &column_mappings,
-            column_registers_start,
-            rowid_reg,
+            rowid_and_columns_start_register,
             &resolver,
         )?;
     }
@@ -354,7 +371,7 @@ pub fn translate_insert(
         program.emit_insn(Insn::OpenWrite {
             cursor_id: *cdc_cursor_id,
             root_page: cdc_btree.root_page.into(),
-            name: cdc_btree.name.clone(),
+            db: 0,
         });
     }
 
@@ -363,30 +380,18 @@ pub fn translate_insert(
         program.emit_insn(Insn::OpenWrite {
             cursor_id: idx_cursor.2,
             root_page: idx_cursor.1.into(),
-            name: idx_cursor.0.clone(),
+            db: 0,
         });
     }
     // Common record insertion logic for both single and multiple rows
-    let check_rowid_is_integer_label = rowid_alias_reg.and(Some(program.allocate_label()));
-    if let Some(reg) = rowid_alias_reg {
-        // for the row record, the rowid alias column (INTEGER PRIMARY KEY) is always set to NULL
-        // and its value is copied to the rowid register. in the case where a single row is inserted,
-        // the value is written directly to the rowid register (see populate_column_registers()).
-        // again, not sure why this only happens in the single row case, but let's mimic sqlite.
-        // in the single row case we save a Copy instruction, but in the multiple rows case we do
-        // it here in the loop.
-        if inserting_multiple_rows {
-            program.emit_insn(Insn::Copy {
-                src_reg: reg,
-                dst_reg: rowid_reg,
-                amount: 0, // TODO: rename 'amount' to something else; amount==0 means 1
-            });
-            // for the row record, the rowid alias column is always set to NULL
-            program.emit_insn(Insn::SoftNull { reg });
-        }
-        // the user provided rowid value might itself be NULL. If it is, we create a new rowid on the next instruction.
+    let check_rowid_is_integer_label = if has_user_provided_rowid {
+        Some(program.allocate_label())
+    } else {
+        None
+    };
+    if has_user_provided_rowid {
         program.emit_insn(Insn::NotNull {
-            reg: rowid_reg,
+            reg: rowid_and_columns_start_register,
             target_pc: check_rowid_is_integer_label.unwrap(),
         });
     }
@@ -394,14 +399,16 @@ pub fn translate_insert(
     // Create new rowid if a) not provided by user or b) provided by user but is NULL
     program.emit_insn(Insn::NewRowid {
         cursor: cursor_id,
-        rowid_reg,
+        rowid_reg: rowid_and_columns_start_register,
         prev_largest_reg: 0,
     });
 
     if let Some(must_be_int_label) = check_rowid_is_integer_label {
         program.resolve_label(must_be_int_label, program.offset());
         // If the user provided a rowid, it must be an integer.
-        program.emit_insn(Insn::MustBeInt { reg: rowid_reg });
+        program.emit_insn(Insn::MustBeInt {
+            reg: rowid_and_columns_start_register,
+        });
     }
 
     // Check uniqueness constraint for rowid if it was provided by user.
@@ -410,24 +417,17 @@ pub fn translate_insert(
         let make_record_label = program.allocate_label();
         program.emit_insn(Insn::NotExists {
             cursor: cursor_id,
-            rowid_reg,
+            rowid_reg: rowid_and_columns_start_register,
             target_pc: make_record_label,
         });
-        let rowid_column_name = if let Some(index) = rowid_alias_index {
-            btree_table
-                .columns
-                .get(index)
-                .unwrap()
-                .name
-                .as_ref()
-                .expect("column name is None")
-        } else {
-            "rowid"
-        };
-
+        let rowid_column = column_mappings
+            .iter()
+            .find(|x| x.value_index.is_some() && x.column.is_rowid_alias)
+            .expect("rowid alias column must be provided");
+        let rowid_column_name = rowid_column.column.name.as_deref().unwrap_or(ROWID);
         program.emit_insn(Insn::Halt {
             err_code: SQLITE_CONSTRAINT_PRIMARYKEY,
-            description: format!("{}.{}", table_name.0, rowid_column_name),
+            description: format!("{}.{}", table_name.as_str(), rowid_column_name),
         });
         program.preassign_label_to_next_insn(make_record_label);
     }
@@ -435,10 +435,10 @@ pub fn translate_insert(
     match table.btree() {
         Some(t) if t.is_strict => {
             program.emit_insn(Insn::TypeCheck {
-                start_reg: column_registers_start,
+                start_reg: columns_start_register,
                 count: num_cols,
                 check_generated: true,
-                table_reference: Rc::clone(&t),
+                table_reference: Arc::clone(&t),
             });
         }
         _ => (),
@@ -462,20 +462,20 @@ pub fn translate_insert(
             // copy from the table's column register over to the index's scratch register
 
             program.emit_insn(Insn::Copy {
-                src_reg: column_registers_start + col.0,
+                src_reg: rowid_and_columns_start_register + col.0,
                 dst_reg: idx_start_reg + i,
-                amount: 0,
+                extra_amount: 0,
             });
         }
         // last register is the rowid
         program.emit_insn(Insn::Copy {
-            src_reg: rowid_reg,
+            src_reg: rowid_and_columns_start_register,
             dst_reg: idx_start_reg + num_cols,
-            amount: 0,
+            extra_amount: 0,
         });
 
         let index = schema
-            .get_index(&table_name.0, &index_col_mapping.idx_name)
+            .get_index(table_name.as_str(), &index_col_mapping.idx_name)
             .expect("index should be present");
 
         let record_reg = program.alloc_register();
@@ -504,10 +504,10 @@ pub fn translate_insert(
                     accum.push_str(&btree_table.name);
                     accum.push('.');
 
-                    let name = btree_table
-                        .columns
+                    let name = column_mappings
                         .get(*index)
                         .unwrap()
+                        .column
                         .name
                         .as_ref()
                         .expect("column name is None");
@@ -545,7 +545,7 @@ pub fn translate_insert(
         if col.column.is_rowid_alias {
             continue;
         }
-        let target_reg = i + column_registers_start;
+        let target_reg = i + rowid_and_columns_start_register;
         program.emit_insn(Insn::HaltIfNull {
             target_reg,
             err_code: SQLITE_CONSTRAINT_NOTNULL,
@@ -561,14 +561,14 @@ pub fn translate_insert(
     }
     // Create and insert the record
     program.emit_insn(Insn::MakeRecord {
-        start_reg: column_registers_start,
+        start_reg: columns_start_register,
         count: num_cols,
         dest_reg: record_register,
         index_name: None,
     });
     program.emit_insn(Insn::Insert {
         cursor: cursor_id,
-        key_reg: rowid_reg,
+        key_reg: rowid_and_columns_start_register,
         record_reg: record_register,
         flag: InsertFlags::new(),
         table_name: table_name.to_string(),
@@ -581,9 +581,9 @@ pub fn translate_insert(
             Some(emit_cdc_patch_record(
                 &mut program,
                 &table,
-                column_registers_start,
+                columns_start_register,
                 record_register,
-                rowid_reg,
+                rowid_and_columns_start_register,
             ))
         } else {
             None
@@ -593,13 +593,24 @@ pub fn translate_insert(
             &resolver,
             OperationMode::INSERT,
             *cdc_cursor_id,
-            rowid_reg,
+            rowid_and_columns_start_register,
             None,
             after_record_reg,
-            &table_name.0,
+            table_name.as_str(),
         )?;
     }
 
+    // Emit RETURNING results if specified
+    if !result_columns.is_empty() {
+        let value_registers = ReturningValueRegisters {
+            rowid_register: rowid_and_columns_start_register,
+            columns_start_register,
+            num_columns: table.columns().len(),
+        };
+
+        emit_returning_results(&mut program, &result_columns, &value_registers)?;
+    }
+
     if inserting_multiple_rows {
         if let Some(temp_table_ctx) = temp_table_ctx {
             program.emit_insn(Insn::Next {
@@ -638,6 +649,19 @@ struct ColumnMapping<'a> {
     default_value: Option<&'a Expr>,
 }
 
+pub const ROWID_COLUMN: &Column = &Column {
+    name: None,
+    ty: schema::Type::Integer,
+    ty_str: String::new(),
+    primary_key: true,
+    is_rowid_alias: true,
+    notnull: true,
+    default: None,
+    unique: false,
+    collation: None,
+    hidden: false,
+};
+
 /// Resolves how each column in a table should be populated during an INSERT.
 /// Returns a Vec of ColumnMapping, one for each column in the table's schema.
 ///
@@ -660,60 +684,69 @@ fn resolve_columns_for_insert<'a>(
     num_values: usize,
 ) -> Result<Vec<ColumnMapping<'a>>> {
     let table_columns = table.columns();
-    // Case 1: No columns specified - map values to columns in order
+
+    // +1 for rowid implicit ROWID column
+    let mut column_mappings = Vec::with_capacity(table_columns.len() + 1);
+    column_mappings.push(ColumnMapping {
+        column: ROWID_COLUMN,
+        value_index: None,
+        default_value: None,
+    });
+    for column in table_columns {
+        column_mappings.push(ColumnMapping {
+            column,
+            value_index: None,
+            default_value: column.default.as_ref(),
+        });
+    }
+
     if columns.is_none() {
-        if num_values != table_columns.len() {
+        // Case 1: No columns specified - map values to columns in order
+        let mut value_idx = 0;
+        for (i, col) in table_columns.iter().enumerate() {
+            column_mappings[i + 1].value_index = if col.hidden { None } else { Some(value_idx) };
+            if !col.hidden {
+                value_idx += 1;
+            }
+        }
+
+        if num_values != value_idx {
             crate::bail_parse_error!(
                 "table {} has {} columns but {} values were supplied",
                 &table.get_name(),
-                table_columns.len(),
+                value_idx,
                 num_values
             );
         }
+    } else {
+        // Case 2: Columns specified - map named columns to their values
+        // Map each named column to its value index
+        for (value_index, column_name) in columns.as_ref().unwrap().iter().enumerate() {
+            let column_name = normalize_ident(column_name.as_str());
+            let table_index = if column_name == ROWID {
+                Some(0)
+            } else {
+                column_mappings.iter().position(|c| {
+                    c.column
+                        .name
+                        .as_ref()
+                        .is_some_and(|name| name.eq_ignore_ascii_case(&column_name))
+                })
+            };
 
-        // Map each column to either its corresponding value index or None
-        return Ok(table_columns
-            .iter()
-            .enumerate()
-            .map(|(i, col)| ColumnMapping {
-                column: col,
-                value_index: if i < num_values { Some(i) } else { None },
-                default_value: col.default.as_ref(),
-            })
-            .collect());
+            let Some(table_index) = table_index else {
+                crate::bail_parse_error!(
+                    "table {} has no column named {}",
+                    &table.get_name(),
+                    column_name
+                );
+            };
+
+            column_mappings[table_index].value_index = Some(value_index);
+        }
     }
 
-    // Case 2: Columns specified - map named columns to their values
-    let mut mappings: Vec<_> = table_columns
-        .iter()
-        .map(|col| ColumnMapping {
-            column: col,
-            value_index: None,
-            default_value: col.default.as_ref(),
-        })
-        .collect();
-
-    // Map each named column to its value index
-    for (value_index, column_name) in columns.as_ref().unwrap().iter().enumerate() {
-        let column_name = normalize_ident(column_name.0.as_str());
-        let table_index = table_columns.iter().position(|c| {
-            c.name
-                .as_ref()
-                .is_some_and(|name| name.eq_ignore_ascii_case(&column_name))
-        });
-
-        let Some(table_index) = table_index else {
-            crate::bail_parse_error!(
-                "table {} has no column named {}",
-                &table.get_name(),
-                column_name
-            );
-        };
-
-        mappings[table_index].value_index = Some(value_index);
-    }
-
-    Ok(mappings)
+    Ok(column_mappings)
 }
 
 /// Represents how a column in an index should be populated during an INSERT.
@@ -781,47 +814,57 @@ fn resolve_indicies_for_insert(
 
 fn populate_columns_multiple_rows(
     program: &mut ProgramBuilder,
-    column_mappings: &[ColumnMapping],
+    column_mappings: &[ColumnMapping], // columns in order of table definition
     column_registers_start: usize,
     yield_reg: usize,
     resolver: &Resolver,
     temp_table_ctx: &Option<TempTableCtx>,
 ) -> Result<()> {
-    let mut value_index_seen = 0;
-    let mut other_values_seen = 0;
+    // In case when both rowid and rowid-alias column provided in the query - turso-db overwrite rowid with **latest** value from the list
+    // As we iterate by column in natural order of their definition in scheme,
+    // we need to track last value_index we wrote to the rowid and overwrite rowid register only if new value_index is greater
+    let mut last_rowid_explicit_value = None;
     for (i, mapping) in column_mappings.iter().enumerate() {
-        let target_reg = column_registers_start + i;
+        let column_register = column_registers_start + i;
 
-        other_values_seen += 1;
         if let Some(value_index) = mapping.value_index {
-            // Decrement as we have now seen a value index instead
-            other_values_seen -= 1;
+            let write_directly_to_rowid_reg = mapping.column.is_rowid_alias;
+            let write_reg = if write_directly_to_rowid_reg {
+                if last_rowid_explicit_value.is_some_and(|x| x > value_index) {
+                    continue;
+                }
+                last_rowid_explicit_value = Some(value_index);
+                column_registers_start // rowid always the first register in the array for insertion record
+            } else {
+                column_register
+            };
             if let Some(temp_table_ctx) = temp_table_ctx {
-                program.emit_column(
-                    temp_table_ctx.cursor_id,
-                    value_index_seen,
-                    column_registers_start + i,
-                );
+                program.emit_column(temp_table_ctx.cursor_id, value_index, write_reg);
             } else {
                 program.emit_insn(Insn::Copy {
-                    src_reg: yield_reg + value_index_seen,
-                    dst_reg: column_registers_start + value_index + other_values_seen,
-                    amount: 0,
+                    src_reg: yield_reg + value_index,
+                    dst_reg: write_reg,
+                    extra_amount: 0,
                 });
             }
-
-            value_index_seen += 1;
+            // write_reg can be euqal to column_register if column list explicitly mention "rowid"
+            if write_reg != column_register {
+                program.emit_null(column_register, None);
+                program.mark_last_insn_constant();
+            }
         } else if mapping.column.is_rowid_alias {
-            program.emit_insn(Insn::SoftNull { reg: target_reg });
+            program.emit_insn(Insn::SoftNull {
+                reg: column_register,
+            });
         } else if let Some(default_expr) = mapping.default_value {
-            translate_expr(program, None, default_expr, target_reg, resolver)?;
+            translate_expr(program, None, default_expr, column_register, resolver)?;
         } else {
             // Column was not specified as has no DEFAULT - use NULL if it is nullable, otherwise error
             // Rowid alias columns can be NULL because we will autogenerate a rowid in that case.
             let is_nullable = !mapping.column.primary_key || mapping.column.is_rowid_alias;
             if is_nullable {
                 program.emit_insn(Insn::Null {
-                    dest: target_reg,
+                    dest: column_register,
                     dest_end: None,
                 });
             } else {
@@ -842,39 +885,55 @@ fn populate_column_registers(
     value: &[Expr],
     column_mappings: &[ColumnMapping],
     column_registers_start: usize,
-    rowid_reg: usize,
     resolver: &Resolver,
 ) -> Result<()> {
+    // In case when both rowid and rowid-alias column provided in the query - turso-db overwrite rowid with **latest** value from the list
+    // As we iterate by column in natural order of their definition in scheme,
+    // we need to track last value_index we wrote to the rowid and overwrite rowid register only if new value_index is greater
+    let mut last_rowid_explicit_value = None;
     for (i, mapping) in column_mappings.iter().enumerate() {
-        let target_reg = column_registers_start + i;
+        let column_register = column_registers_start + i;
 
         // Column has a value in the VALUES tuple
         if let Some(value_index) = mapping.value_index {
             // When inserting a single row, SQLite writes the value provided for the rowid alias column (INTEGER PRIMARY KEY)
             // directly into the rowid register and writes a NULL into the rowid alias column.
             let write_directly_to_rowid_reg = mapping.column.is_rowid_alias;
-            let reg = if write_directly_to_rowid_reg {
-                rowid_reg
+            let write_reg = if write_directly_to_rowid_reg {
+                if last_rowid_explicit_value.is_some_and(|x| x > value_index) {
+                    continue;
+                }
+                last_rowid_explicit_value = Some(value_index);
+                column_registers_start // rowid always the first register in the array for insertion record
             } else {
-                target_reg
+                column_register
             };
+
             translate_expr_no_constant_opt(
                 program,
                 None,
                 value.get(value_index).expect("value index out of bounds"),
-                reg,
+                write_reg,
                 resolver,
                 NoConstantOptReason::RegisterReuse,
             )?;
-            if write_directly_to_rowid_reg {
-                program.emit_insn(Insn::SoftNull { reg: target_reg });
+            // write_reg can be euqal to column_register if column list explicitly mention "rowid"
+            if write_reg != column_register {
+                program.emit_null(column_register, None);
+                program.mark_last_insn_constant();
             }
+        } else if mapping.column.hidden {
+            program.emit_insn(Insn::Null {
+                dest: column_register,
+                dest_end: None,
+            });
+            program.mark_last_insn_constant();
         } else if let Some(default_expr) = mapping.default_value {
             translate_expr_no_constant_opt(
                 program,
                 None,
                 default_expr,
-                target_reg,
+                column_register,
                 resolver,
                 NoConstantOptReason::RegisterReuse,
             )?;
@@ -884,7 +943,7 @@ fn populate_column_registers(
             let is_nullable = !mapping.column.primary_key || mapping.column.is_rowid_alias;
             if is_nullable {
                 program.emit_insn(Insn::Null {
-                    dest: target_reg,
+                    dest: column_register,
                     dest_end: None,
                 });
                 program.mark_last_insn_constant();
@@ -902,12 +961,15 @@ fn populate_column_registers(
 // TODO: comeback here later to apply the same improvements on select
 fn translate_virtual_table_insert(
     mut program: ProgramBuilder,
-    virtual_table: Rc<VirtualTable>,
+    virtual_table: Arc<VirtualTable>,
     columns: Option<DistinctNames>,
     mut body: InsertBody,
     on_conflict: Option<ResolveType>,
     resolver: &Resolver,
 ) -> Result<ProgramBuilder> {
+    if virtual_table.readonly() {
+        crate::bail_constraint_error!("Table is read-only: {}", virtual_table.name);
+    }
     let (num_values, value) = match &mut body {
         InsertBody::Select(select, None) => match select.body.select.as_mut() {
             OneSelect::Values(values) => (values[0].len(), values.pop().unwrap()),
@@ -918,27 +980,25 @@ fn translate_virtual_table_insert(
     };
     let table = Table::Virtual(virtual_table.clone());
     let column_mappings = resolve_columns_for_insert(&table, &columns, num_values)?;
-    let registers_start = program.alloc_registers(2);
+    let registers_start = program.alloc_registers(column_mappings.len() + 1);
 
     /* *
      * Inserts for virtual tables are done in a single step.
      * argv[0] = (NULL for insert)
-     * argv[1] = (NULL for insert)
+     * argv[1] = (rowid for insert - NULL in most cases)
      * argv[2..] = column values
      * */
 
     program.emit_insn(Insn::Null {
         dest: registers_start,
-        dest_end: Some(registers_start + 1),
+        dest_end: None,
     });
 
-    let values_reg = program.alloc_registers(column_mappings.len());
     populate_column_registers(
         &mut program,
         &value,
         &column_mappings,
-        values_reg,
-        registers_start,
+        registers_start + 1,
         resolver,
     )?;
     let conflict_action = on_conflict.as_ref().map(|c| c.bit_value()).unwrap_or(0) as u16;
@@ -947,7 +1007,7 @@ fn translate_virtual_table_insert(
 
     program.emit_insn(Insn::VUpdate {
         cursor_id,
-        arg_count: column_mappings.len() + 2,
+        arg_count: column_mappings.len() + 1,
         start_reg: registers_start,
         conflict_action,
     });
