@@ -490,7 +490,7 @@ fn generate_join_bitmasks(table_number_max_exclusive: usize, how_many: usize) ->
 
 #[cfg(test)]
 mod tests {
-    use std::{cell::Cell, rc::Rc, sync::Arc};
+    use std::{cell::Cell, sync::Arc};
 
     use turso_sqlite3_parser::ast::{self, Expr, Operator, SortOrder, TableInternalId};
 
@@ -1310,6 +1310,7 @@ mod tests {
             identifier: "t1".to_string(),
             join_info: None,
             col_used_mask: ColumnUsedMask::default(),
+            database_id: 0,
         });
 
         // Create where clause that only references second column
@@ -1401,6 +1402,7 @@ mod tests {
             identifier: "t1".to_string(),
             join_info: None,
             col_used_mask: ColumnUsedMask::default(),
+            database_id: 0,
         });
 
         // Create where clause that references first and third columns
@@ -1517,6 +1519,7 @@ mod tests {
             identifier: "t1".to_string(),
             join_info: None,
             col_used_mask: ColumnUsedMask::default(),
+            database_id: 0,
         });
 
         // Create where clause: c1 = 5 AND c2 > 10 AND c3 = 7
@@ -1610,6 +1613,7 @@ mod tests {
             default: None,
             unique: false,
             collation: None,
+            hidden: false,
         }
     }
     fn _create_column_of_type(name: &str, ty: Type) -> Column {
@@ -1636,8 +1640,8 @@ mod tests {
     }
 
     /// Creates a BTreeTable with the given name and columns
-    fn _create_btree_table(name: &str, columns: Vec<Column>) -> Rc<BTreeTable> {
-        Rc::new(BTreeTable {
+    fn _create_btree_table(name: &str, columns: Vec<Column>) -> Arc<BTreeTable> {
+        Arc::new(BTreeTable {
             root_page: 1, // Page number doesn't matter for tests
             name: name.to_string(),
             primary_key_columns: vec![],
@@ -1650,7 +1654,7 @@ mod tests {
 
     /// Creates a TableReference for a BTreeTable
     fn _create_table_reference(
-        table: Rc<BTreeTable>,
+        table: Arc<BTreeTable>,
         join_info: Option<JoinInfo>,
         internal_id: TableInternalId,
     ) -> JoinedTable {
@@ -1665,6 +1669,7 @@ mod tests {
             internal_id,
             join_info,
             col_used_mask: ColumnUsedMask::default(),
+            database_id: 0,
         }
     }
 
