@@ -7,15 +7,12 @@ use cost::Cost;
 use join::{compute_best_join_order, BestJoinOrderResult};
 use lift_common_subexpressions::lift_common_subexpressions_from_binary_or_terms;
 use order::{compute_order_target, plan_satisfies_order_target, EliminatesSortBy};
-use turso_sqlite3_parser::{
-    ast::{self, Expr, SortOrder},
-    to_sql_string::ToSqlString as _,
-};
+use turso_sqlite3_parser::ast::{self, fmt::ToTokens as _, Expr, SortOrder};
 
 use crate::{
     parameters::PARAM_PREFIX,
     schema::{Index, IndexColumn, Schema, Table},
-    translate::{expr::walk_expr_mut, plan::TerminationKey},
+    translate::{expr::is_double_quoted_identifier, expr::walk_expr_mut, plan::TerminationKey},
     types::SeekOp,
     Result,
 };
@@ -51,7 +48,11 @@ pub fn optimize_plan(plan: &mut Plan, schema: &Schema) -> Result<()> {
         }
     }
     // When debug tracing is enabled, print the optimized plan as a SQL string for debugging
-    tracing::debug!(plan_sql = plan.to_sql_string(&crate::translate::display::PlanContext(&[])));
+    tracing::debug!(
+        plan_sql = plan
+            .format_with_context(&crate::translate::display::PlanContext(&[]))
+            .unwrap()
+    );
     Ok(())
 }
 
@@ -243,11 +244,11 @@ fn optimize_table_access(
             let try_to_build_ephemeral_index = if schema.indexes_enabled() {
                 let is_leftmost_table = i == 0;
                 let uses_index = access_method.index.is_some();
-                let source_table_is_from_clause_subquery = matches!(
+                let source_table_does_not_support_search = matches!(
                     &joined_tables[table_idx].table,
-                    Table::FromClauseSubquery(_)
+                    Table::FromClauseSubquery(_) | Table::Virtual(_)
                 );
-                !is_leftmost_table && !uses_index && !source_table_is_from_clause_subquery
+                !is_leftmost_table && !uses_index && !source_table_does_not_support_search
             } else {
                 false
             };
@@ -592,8 +593,8 @@ impl Optimizable for ast::Expr {
             }
             Expr::Exists(_) => false,
             Expr::FunctionCall { args, name, .. } => {
-                let Some(func) =
-                    resolver.resolve_function(&name.0, args.as_ref().map_or(0, |args| args.len()))
+                let Some(func) = resolver
+                    .resolve_function(name.as_str(), args.as_ref().map_or(0, |args| args.len()))
                 else {
                     return false;
                 };
@@ -603,7 +604,11 @@ impl Optimizable for ast::Expr {
                         .is_none_or(|args| args.iter().all(|arg| arg.is_constant(resolver)))
             }
             Expr::FunctionCallStar { .. } => false,
-            Expr::Id(_) => panic!("Id should have been rewritten as Column"),
+            Expr::Id(id) => {
+                // If we got here with an id, this has to be double-quotes identifier
+                assert!(is_double_quoted_identifier(id.as_str()));
+                true
+            }
             Expr::Column { .. } => false,
             Expr::RowId { .. } => false,
             Expr::InList { lhs, rhs, .. } => {
@@ -1302,11 +1307,11 @@ pub fn rewrite_expr(top_level_expr: &mut ast::Expr, param_idx: &mut usize) -> Re
         match expr {
             ast::Expr::Id(id) => {
                 // Convert "true" and "false" to 1 and 0
-                if id.0.eq_ignore_ascii_case("true") {
+                if id.as_str().eq_ignore_ascii_case("true") {
                     *expr = ast::Expr::Literal(ast::Literal::Numeric(1.to_string()));
                     return Ok(());
                 }
-                if id.0.eq_ignore_ascii_case("false") {
+                if id.as_str().eq_ignore_ascii_case("false") {
                     *expr = ast::Expr::Literal(ast::Literal::Numeric(0.to_string()));
                 }
             }
