@@ -1,4 +1,5 @@
 use std::cell::Cell;
+use std::sync::Arc;
 
 use super::{
     expr::walk_expr,
@@ -19,8 +20,10 @@ use crate::{
     vdbe::{builder::TableRefIdCounter, BranchOffset},
     Result,
 };
+use turso_sqlite3_parser::ast::Literal::Null;
 use turso_sqlite3_parser::ast::{
-    self, Expr, FromClause, JoinType, Limit, Materialized, TableInternalId, UnaryOperator, With,
+    self, As, Expr, FromClause, JoinType, Limit, Literal, Materialized, QualifiedName,
+    TableInternalId, UnaryOperator, With,
 };
 
 pub const ROWID: &str = "rowid";
@@ -51,7 +54,7 @@ pub fn resolve_aggregates(
                 } else {
                     0
                 };
-                match Func::resolve_function(&name.0, args_count) {
+                match Func::resolve_function(name.as_str(), args_count) {
                     Ok(Func::Agg(f)) => {
                         let distinctness = Distinctness::from_ast(distinctness.as_ref());
                         if !schema.indexes_enabled() && distinctness.is_distinct() {
@@ -83,7 +86,7 @@ pub fn resolve_aggregates(
                 }
             }
             Expr::FunctionCallStar { name, .. } => {
-                if let Ok(Func::Agg(f)) = Func::resolve_function(&name.0, 0) {
+                if let Ok(Func::Agg(f)) = Func::resolve_function(name.as_str(), 0) {
                     aggs.push(Aggregate {
                         func: f,
                         args: vec![],
@@ -106,16 +109,19 @@ pub fn bind_column_references(
     top_level_expr: &mut Expr,
     referenced_tables: &mut TableReferences,
     result_columns: Option<&[ResultSetColumn]>,
+    connection: &Arc<crate::Connection>,
 ) -> Result<()> {
     walk_expr_mut(top_level_expr, &mut |expr: &mut Expr| -> Result<()> {
         match expr {
             Expr::Id(id) => {
                 // true and false are special constants that are effectively aliases for 1 and 0
                 // and not identifiers of columns
-                if id.0.eq_ignore_ascii_case("true") || id.0.eq_ignore_ascii_case("false") {
+                if id.as_str().eq_ignore_ascii_case("true")
+                    || id.as_str().eq_ignore_ascii_case("false")
+                {
                     return Ok(());
                 }
-                let normalized_id = normalize_ident(id.0.as_str());
+                let normalized_id = normalize_ident(id.as_str());
 
                 if !referenced_tables.joined_tables().is_empty() {
                     if let Some(row_id_expr) = parse_row_id(
@@ -139,7 +145,7 @@ pub fn bind_column_references(
                     });
                     if col_idx.is_some() {
                         if match_result.is_some() {
-                            crate::bail_parse_error!("Column {} is ambiguous", id.0);
+                            crate::bail_parse_error!("Column {} is ambiguous", id.as_str());
                         }
                         let col = joined_table.table.columns().get(col_idx.unwrap()).unwrap();
                         match_result = Some((
@@ -167,7 +173,7 @@ pub fn bind_column_references(
                         });
                         if col_idx.is_some() {
                             if match_result.is_some() {
-                                crate::bail_parse_error!("Column {} is ambiguous", id.0);
+                                crate::bail_parse_error!("Column {} is ambiguous", id.as_str());
                             }
                             let col = outer_ref.table.columns().get(col_idx.unwrap()).unwrap();
                             match_result =
@@ -198,17 +204,26 @@ pub fn bind_column_references(
                         }
                     }
                 }
-                crate::bail_parse_error!("Column {} not found", id.0);
+                // SQLite behavior: Only double-quoted identifiers get fallback to string literals
+                // Single quotes are handled as literals earlier, unquoted identifiers must resolve to columns
+                if crate::translate::expr::is_double_quoted_identifier(id.as_str()) {
+                    // Convert failed double-quoted identifier to string literal
+                    *expr = Expr::Literal(Literal::String(id.as_str().to_string()));
+                    Ok(())
+                } else {
+                    // Unquoted identifiers must resolve to columns - no fallback
+                    crate::bail_parse_error!("no such column: {}", id.as_str())
+                }
             }
             Expr::Qualified(tbl, id) => {
-                let normalized_table_name = normalize_ident(tbl.0.as_str());
+                let normalized_table_name = normalize_ident(tbl.as_str());
                 let matching_tbl = referenced_tables
                     .find_table_and_internal_id_by_identifier(&normalized_table_name);
                 if matching_tbl.is_none() {
                     crate::bail_parse_error!("no such table: {}", normalized_table_name);
                 }
                 let (tbl_id, tbl) = matching_tbl.unwrap();
-                let normalized_id = normalize_ident(id.0.as_str());
+                let normalized_id = normalize_ident(id.as_str());
 
                 if let Some(row_id_expr) = parse_row_id(&normalized_id, tbl_id, || false)? {
                     *expr = row_id_expr;
@@ -221,7 +236,7 @@ pub fn bind_column_references(
                         .is_some_and(|name| name.eq_ignore_ascii_case(&normalized_id))
                 });
                 let Some(col_idx) = col_idx else {
-                    crate::bail_parse_error!("Column {} not found", normalized_id);
+                    crate::bail_parse_error!("no such column: {}", normalized_id);
                 };
                 let col = tbl.columns().get(col_idx).unwrap();
                 *expr = Expr::Column {
@@ -233,92 +248,102 @@ pub fn bind_column_references(
                 referenced_tables.mark_column_used(tbl_id, col_idx);
                 Ok(())
             }
+            Expr::DoublyQualified(db_name, tbl_name, col_name) => {
+                let normalized_col_name = normalize_ident(col_name.as_str());
+
+                // Create a QualifiedName and use existing resolve_database_id method
+                let qualified_name = ast::QualifiedName {
+                    db_name: Some(db_name.clone()),
+                    name: tbl_name.clone(),
+                    alias: None,
+                };
+                let database_id = connection.resolve_database_id(&qualified_name)?;
+
+                // Get the table from the specified database
+                let table = connection
+                    .with_schema(database_id, |schema| schema.get_table(tbl_name.as_str()))
+                    .ok_or_else(|| {
+                        crate::LimboError::ParseError(format!(
+                            "no such table: {}.{}",
+                            db_name.as_str(),
+                            tbl_name.as_str()
+                        ))
+                    })?;
+
+                // Find the column in the table
+                let col_idx = table
+                    .columns()
+                    .iter()
+                    .position(|c| {
+                        c.name
+                            .as_ref()
+                            .is_some_and(|name| name.eq_ignore_ascii_case(&normalized_col_name))
+                    })
+                    .ok_or_else(|| {
+                        crate::LimboError::ParseError(format!(
+                            "Column: {}.{}.{} not found",
+                            db_name.as_str(),
+                            tbl_name.as_str(),
+                            col_name.as_str()
+                        ))
+                    })?;
+
+                let col = table.columns().get(col_idx).unwrap();
+
+                // Check if this is a rowid alias
+                let is_rowid_alias = col.is_rowid_alias;
+
+                // Convert to Column expression - since this is a cross-database reference,
+                // we need to create a synthetic table reference for it
+                // For now, we'll error if the table isn't already in the referenced tables
+                let normalized_tbl_name = normalize_ident(tbl_name.as_str());
+                let matching_tbl = referenced_tables
+                    .find_table_and_internal_id_by_identifier(&normalized_tbl_name);
+
+                if let Some((tbl_id, _)) = matching_tbl {
+                    // Table is already in referenced tables, use existing internal ID
+                    *expr = Expr::Column {
+                        database: Some(database_id),
+                        table: tbl_id,
+                        column: col_idx,
+                        is_rowid_alias,
+                    };
+                    referenced_tables.mark_column_used(tbl_id, col_idx);
+                } else {
+                    return Err(crate::LimboError::ParseError(format!(
+                        "table {normalized_tbl_name} is not in FROM clause - cross-database column references require the table to be explicitly joined"
+                    )));
+                }
+
+                Ok(())
+            }
             _ => Ok(()),
         }
     })
 }
 
+#[allow(clippy::too_many_arguments)]
 fn parse_from_clause_table(
     schema: &Schema,
     table: ast::SelectTable,
     table_references: &mut TableReferences,
+    out_where_clause: &mut Vec<WhereTerm>,
     ctes: &mut Vec<JoinedTable>,
     syms: &SymbolTable,
     table_ref_counter: &mut TableRefIdCounter,
+    connection: &Arc<crate::Connection>,
 ) -> Result<()> {
     match table {
-        ast::SelectTable::Table(qualified_name, maybe_alias, _) => {
-            let normalized_qualified_name = normalize_ident(qualified_name.name.0.as_str());
-            // Check if the FROM clause table is referring to a CTE in the current scope.
-            if let Some(cte_idx) = ctes
-                .iter()
-                .position(|cte| cte.identifier == normalized_qualified_name)
-            {
-                // TODO: what if the CTE is referenced multiple times?
-                let cte_table = ctes.remove(cte_idx);
-                table_references.add_joined_table(cte_table);
-                return Ok(());
-            };
-
-            // Check if our top level schema has this table.
-            if let Some(table) = schema.get_table(&normalized_qualified_name) {
-                let alias = maybe_alias
-                    .map(|a| match a {
-                        ast::As::As(id) => id,
-                        ast::As::Elided(id) => id,
-                    })
-                    .map(|a| a.0);
-                let tbl_ref = if let Table::Virtual(tbl) = table.as_ref() {
-                    Table::Virtual(tbl.clone())
-                } else if let Table::BTree(table) = table.as_ref() {
-                    Table::BTree(table.clone())
-                } else {
-                    return Err(crate::LimboError::InvalidArgument(
-                        "Table type not supported".to_string(),
-                    ));
-                };
-                table_references.add_joined_table(JoinedTable {
-                    op: Operation::Scan {
-                        iter_dir: IterationDirection::Forwards,
-                        index: None,
-                    },
-                    table: tbl_ref,
-                    identifier: alias.unwrap_or(normalized_qualified_name),
-                    internal_id: table_ref_counter.next(),
-                    join_info: None,
-                    col_used_mask: ColumnUsedMask::default(),
-                });
-                return Ok(());
-            };
-
-            // CTEs are transformed into FROM clause subqueries.
-            // If we find a CTE with this name in our outer query references,
-            // we can use it as a joined table, but we must clone it since it's not MATERIALIZED.
-            //
-            // For other types of tables in the outer query references, we do not add them as joined tables,
-            // because the query can simply _reference_ them in e.g. the SELECT columns or the WHERE clause,
-            // but it's not part of the join order.
-            if let Some(outer_ref) =
-                table_references.find_outer_query_ref_by_identifier(&normalized_qualified_name)
-            {
-                if matches!(outer_ref.table, Table::FromClauseSubquery(_)) {
-                    table_references.add_joined_table(JoinedTable {
-                        op: Operation::Scan {
-                            iter_dir: IterationDirection::Forwards,
-                            index: None,
-                        },
-                        table: outer_ref.table.clone(),
-                        identifier: outer_ref.identifier.clone(),
-                        internal_id: table_ref_counter.next(),
-                        join_info: None,
-                        col_used_mask: ColumnUsedMask::default(),
-                    });
-                    return Ok(());
-                }
-            }
-
-            crate::bail_parse_error!("no such table: {}", normalized_qualified_name);
-        }
+        ast::SelectTable::Table(qualified_name, maybe_alias, _) => parse_table(
+            table_references,
+            ctes,
+            table_ref_counter,
+            out_where_clause,
+            qualified_name,
+            maybe_alias,
+            None,
+            connection,
+        ),
         ast::SelectTable::Select(subselect, maybe_alias) => {
             let Plan::Select(subplan) = prepare_select_plan(
                 schema,
@@ -330,6 +355,7 @@ fn parse_from_clause_table(
                     yield_reg: usize::MAX, // will be set later in bytecode emission
                     coroutine_implementation_start: BranchOffset::Placeholder, // will be set later in bytecode emission
                 },
+                connection,
             )?
             else {
                 crate::bail_parse_error!("Only non-compound SELECT queries are currently supported in FROM clause subqueries");
@@ -337,8 +363,8 @@ fn parse_from_clause_table(
             let cur_table_index = table_references.joined_tables().len();
             let identifier = maybe_alias
                 .map(|a| match a {
-                    ast::As::As(id) => id.0.clone(),
-                    ast::As::Elided(id) => id.0.clone(),
+                    ast::As::As(id) => id.as_str().to_string(),
+                    ast::As::Elided(id) => id.as_str().to_string(),
                 })
                 .unwrap_or(format!("subquery_{cur_table_index}"));
             table_references.add_joined_table(JoinedTable::new_subquery(
@@ -349,35 +375,186 @@ fn parse_from_clause_table(
             ));
             Ok(())
         }
-        ast::SelectTable::TableCall(qualified_name, maybe_args, maybe_alias) => {
-            let normalized_name = &normalize_ident(qualified_name.name.0.as_str());
-            let vtab = crate::VirtualTable::function(normalized_name, maybe_args, syms)?;
-            let alias = maybe_alias
-                .as_ref()
-                .map(|a| match a {
-                    ast::As::As(id) => id.0.clone(),
-                    ast::As::Elided(id) => id.0.clone(),
-                })
-                .unwrap_or(normalized_name.to_string());
+        ast::SelectTable::TableCall(qualified_name, maybe_args, maybe_alias) => parse_table(
+            table_references,
+            ctes,
+            table_ref_counter,
+            out_where_clause,
+            qualified_name,
+            maybe_alias,
+            maybe_args,
+            connection,
+        ),
+        _ => todo!(),
+    }
+}
 
+#[allow(clippy::too_many_arguments)]
+fn parse_table(
+    table_references: &mut TableReferences,
+    ctes: &mut Vec<JoinedTable>,
+    table_ref_counter: &mut TableRefIdCounter,
+    out_where_clause: &mut Vec<WhereTerm>,
+    qualified_name: QualifiedName,
+    maybe_alias: Option<As>,
+    maybe_args: Option<Vec<Expr>>,
+    connection: &Arc<crate::Connection>,
+) -> Result<()> {
+    let normalized_qualified_name = normalize_ident(qualified_name.name.as_str());
+    let database_id = connection.resolve_database_id(&qualified_name)?;
+    let table_name = qualified_name.name;
+
+    // Check if the FROM clause table is referring to a CTE in the current scope.
+    if let Some(cte_idx) = ctes
+        .iter()
+        .position(|cte| cte.identifier == normalized_qualified_name)
+    {
+        // TODO: what if the CTE is referenced multiple times?
+        let cte_table = ctes.remove(cte_idx);
+        table_references.add_joined_table(cte_table);
+        return Ok(());
+    };
+
+    // Resolve table using connection's with_schema method
+    let table = connection.with_schema(database_id, |schema| schema.get_table(table_name.as_str()));
+
+    if let Some(table) = table {
+        let alias = maybe_alias
+            .map(|a| match a {
+                ast::As::As(id) => id,
+                ast::As::Elided(id) => id,
+            })
+            .map(|a| a.as_str().to_string());
+        let internal_id = table_ref_counter.next();
+        let tbl_ref = if let Table::Virtual(tbl) = table.as_ref() {
+            if let Some(args) = maybe_args {
+                transform_args_into_where_terms(
+                    args,
+                    internal_id,
+                    out_where_clause,
+                    table.as_ref(),
+                )?;
+            }
+            Table::Virtual(tbl.clone())
+        } else if let Table::BTree(table) = table.as_ref() {
+            Table::BTree(table.clone())
+        } else {
+            return Err(crate::LimboError::InvalidArgument(
+                "Table type not supported".to_string(),
+            ));
+        };
+        table_references.add_joined_table(JoinedTable {
+            op: Operation::Scan {
+                iter_dir: IterationDirection::Forwards,
+                index: None,
+            },
+            table: tbl_ref,
+            identifier: alias.unwrap_or(normalized_qualified_name),
+            internal_id,
+            join_info: None,
+            col_used_mask: ColumnUsedMask::default(),
+            database_id,
+        });
+        return Ok(());
+    };
+
+    // CTEs are transformed into FROM clause subqueries.
+    // If we find a CTE with this name in our outer query references,
+    // we can use it as a joined table, but we must clone it since it's not MATERIALIZED.
+    //
+    // For other types of tables in the outer query references, we do not add them as joined tables,
+    // because the query can simply _reference_ them in e.g. the SELECT columns or the WHERE clause,
+    // but it's not part of the join order.
+    if let Some(outer_ref) =
+        table_references.find_outer_query_ref_by_identifier(&normalized_qualified_name)
+    {
+        if matches!(outer_ref.table, Table::FromClauseSubquery(_)) {
             table_references.add_joined_table(JoinedTable {
                 op: Operation::Scan {
                     iter_dir: IterationDirection::Forwards,
                     index: None,
                 },
-                join_info: None,
-                table: Table::Virtual(vtab),
-                identifier: alias,
+                table: outer_ref.table.clone(),
+                identifier: outer_ref.identifier.clone(),
                 internal_id: table_ref_counter.next(),
+                join_info: None,
                 col_used_mask: ColumnUsedMask::default(),
+                database_id,
             });
-
-            Ok(())
+            return Ok(());
         }
-        _ => todo!(),
     }
+
+    crate::bail_parse_error!("no such table: {}", normalized_qualified_name);
 }
 
+fn transform_args_into_where_terms(
+    args: Vec<Expr>,
+    internal_id: TableInternalId,
+    out_where_clause: &mut Vec<WhereTerm>,
+    table: &Table,
+) -> Result<()> {
+    let mut args_iter = args.into_iter();
+    let mut hidden_count = 0;
+    for (i, col) in table.columns().iter().enumerate() {
+        if !col.hidden {
+            continue;
+        }
+        hidden_count += 1;
+
+        if let Some(arg_expr) = args_iter.next() {
+            if contains_column_reference(&arg_expr)? {
+                crate::bail_parse_error!(
+                    "Column references are not supported as table-valued function arguments yet"
+                );
+            }
+            let column_expr = Expr::Column {
+                database: None,
+                table: internal_id,
+                column: i,
+                is_rowid_alias: col.is_rowid_alias,
+            };
+            let expr = match arg_expr {
+                Expr::Literal(Null) => Expr::IsNull(Box::new(column_expr)),
+                other => Expr::Binary(
+                    Box::new(column_expr),
+                    ast::Operator::Equals,
+                    Box::new(other),
+                ),
+            };
+            out_where_clause.push(WhereTerm {
+                expr,
+                from_outer_join: None,
+                consumed: Cell::new(false),
+            });
+        }
+    }
+
+    if args_iter.next().is_some() {
+        return Err(crate::LimboError::ParseError(format!(
+            "Too many arguments for {}: expected at most {}, got {}",
+            table.get_name(),
+            hidden_count,
+            hidden_count + 1 + args_iter.count()
+        )));
+    }
+
+    Ok(())
+}
+
+fn contains_column_reference(top_level_expr: &Expr) -> Result<bool> {
+    let mut contains = false;
+    walk_expr(top_level_expr, &mut |expr: &Expr| -> Result<WalkControl> {
+        match expr {
+            Expr::Id(_) | Expr::Qualified(_, _) | Expr::Column { .. } => contains = true,
+            _ => {}
+        };
+        Ok(WalkControl::Continue)
+    })?;
+    Ok(contains)
+}
+
+#[allow(clippy::too_many_arguments)]
 pub fn parse_from(
     schema: &Schema,
     mut from: Option<FromClause>,
@@ -386,6 +563,7 @@ pub fn parse_from(
     out_where_clause: &mut Vec<WhereTerm>,
     table_references: &mut TableReferences,
     table_ref_counter: &mut TableRefIdCounter,
+    connection: &Arc<crate::Connection>,
 ) -> Result<()> {
     if from.as_ref().and_then(|f| f.select.as_ref()).is_none() {
         return Ok(());
@@ -408,11 +586,11 @@ pub fn parse_from(
             // Check if normalized name conflicts with catalog tables or other CTEs
             // TODO: sqlite actually allows overriding a catalog table with a CTE.
             // We should carry over the 'Scope' struct to all of our identifier resolution.
-            let cte_name_normalized = normalize_ident(&cte.tbl_name.0);
+            let cte_name_normalized = normalize_ident(cte.tbl_name.as_str());
             if schema.get_table(&cte_name_normalized).is_some() {
                 crate::bail_parse_error!(
                     "CTE name {} conflicts with catalog table name",
-                    cte.tbl_name.0
+                    cte.tbl_name.as_str()
                 );
             }
             if table_references
@@ -422,7 +600,7 @@ pub fn parse_from(
             {
                 crate::bail_parse_error!(
                     "CTE name {} conflicts with WITH table name {}",
-                    cte.tbl_name.0,
+                    cte.tbl_name.as_str(),
                     cte_name_normalized
                 );
             }
@@ -448,6 +626,7 @@ pub fn parse_from(
                     yield_reg: usize::MAX, // will be set later in bytecode emission
                     coroutine_implementation_start: BranchOffset::Placeholder, // will be set later in bytecode emission
                 },
+                connection,
             )?;
             let Plan::Select(cte_plan) = cte_plan else {
                 crate::bail_parse_error!("Only SELECT queries are currently supported in CTEs");
@@ -468,9 +647,11 @@ pub fn parse_from(
         schema,
         select_owned,
         table_references,
+        out_where_clause,
         &mut ctes_as_subqueries,
         syms,
         table_ref_counter,
+        connection,
     )?;
 
     for join in joins_owned.into_iter() {
@@ -482,6 +663,7 @@ pub fn parse_from(
             out_where_clause,
             table_references,
             table_ref_counter,
+            connection,
         )?;
     }
 
@@ -493,12 +675,13 @@ pub fn parse_where(
     table_references: &mut TableReferences,
     result_columns: Option<&[ResultSetColumn]>,
     out_where_clause: &mut Vec<WhereTerm>,
+    connection: &Arc<crate::Connection>,
 ) -> Result<()> {
     if let Some(where_expr) = where_clause {
         let mut predicates = vec![];
         break_predicate_at_and_boundaries(where_expr, &mut predicates);
         for expr in predicates.iter_mut() {
-            bind_column_references(expr, table_references, result_columns)?;
+            bind_column_references(expr, table_references, result_columns, connection)?;
         }
         for expr in predicates {
             out_where_clause.push(WhereTerm {
@@ -687,6 +870,7 @@ pub fn determine_where_to_eval_expr(
     Ok(eval_at)
 }
 
+#[allow(clippy::too_many_arguments)]
 fn parse_join(
     schema: &Schema,
     join: ast::JoinedSelectTable,
@@ -695,6 +879,7 @@ fn parse_join(
     out_where_clause: &mut Vec<WhereTerm>,
     table_references: &mut TableReferences,
     table_ref_counter: &mut TableRefIdCounter,
+    connection: &Arc<crate::Connection>,
 ) -> Result<()> {
     let ast::JoinedSelectTable {
         operator: join_operator,
@@ -706,9 +891,11 @@ fn parse_join(
         schema,
         table,
         table_references,
+        out_where_clause,
         ctes,
         syms,
         table_ref_counter,
+        connection,
     )?;
 
     let (outer, natural) = match join_operator {
@@ -730,27 +917,27 @@ fn parse_join(
         assert!(table_references.joined_tables().len() >= 2);
         let rightmost_table = table_references.joined_tables().last().unwrap();
         // NATURAL JOIN is first transformed into a USING join with the common columns
-        let right_cols = rightmost_table.columns();
         let mut distinct_names: Option<ast::DistinctNames> = None;
         // TODO: O(n^2) maybe not great for large tables or big multiway joins
-        for right_col in right_cols.iter() {
+        // SQLite doesn't use HIDDEN columns for NATURAL joins: https://www3.sqlite.org/src/info/ab09ef427181130b
+        for right_col in rightmost_table.columns().iter().filter(|col| !col.hidden) {
             let mut found_match = false;
             for left_table in table_references
                 .joined_tables()
                 .iter()
                 .take(table_references.joined_tables().len() - 1)
             {
-                for left_col in left_table.columns().iter() {
+                for left_col in left_table.columns().iter().filter(|col| !col.hidden) {
                     if left_col.name == right_col.name {
                         if let Some(distinct_names) = distinct_names.as_mut() {
                             distinct_names
-                                .insert(ast::Name(
-                                    left_col.name.clone().expect("column name is None"),
+                                .insert(ast::Name::from_str(
+                                    &left_col.name.clone().expect("column name is None"),
                                 ))
                                 .unwrap();
                         } else {
-                            distinct_names = Some(ast::DistinctNames::new(ast::Name(
-                                left_col.name.clone().expect("column name is None"),
+                            distinct_names = Some(ast::DistinctNames::new(ast::Name::from_str(
+                                &left_col.name.clone().expect("column name is None"),
                             )));
                         }
                         found_match = true;
@@ -777,7 +964,7 @@ fn parse_join(
                 let mut preds = vec![];
                 break_predicate_at_and_boundaries(expr, &mut preds);
                 for predicate in preds.iter_mut() {
-                    bind_column_references(predicate, table_references, None)?;
+                    bind_column_references(predicate, table_references, None, connection)?;
                 }
                 for pred in preds {
                     out_where_clause.push(WhereTerm {
@@ -794,7 +981,7 @@ fn parse_join(
             ast::JoinConstraint::Using(distinct_names) => {
                 // USING join is replaced with a list of equality predicates
                 for distinct_name in distinct_names.iter() {
-                    let name_normalized = normalize_ident(distinct_name.0.as_str());
+                    let name_normalized = normalize_ident(distinct_name.as_str());
                     let cur_table_idx = table_references.joined_tables().len() - 1;
                     let left_tables = &table_references.joined_tables()[..cur_table_idx];
                     assert!(!left_tables.is_empty());
@@ -805,6 +992,7 @@ fn parse_join(
                             .columns()
                             .iter()
                             .enumerate()
+                            .filter(|(_, col)| !natural || !col.hidden)
                             .find(|(_, col)| {
                                 col.name
                                     .as_ref()
@@ -818,7 +1006,7 @@ fn parse_join(
                     if left_col.is_none() {
                         crate::bail_parse_error!(
                             "cannot join using column {} - column not present in all tables",
-                            distinct_name.0
+                            distinct_name.as_str()
                         );
                     }
                     let right_col = right_table.columns().iter().enumerate().find(|(_, col)| {
@@ -829,7 +1017,7 @@ fn parse_join(
                     if right_col.is_none() {
                         crate::bail_parse_error!(
                             "cannot join using column {} - column not present in all tables",
-                            distinct_name.0
+                            distinct_name.as_str()
                         );
                     }
                     let (left_table_idx, left_table_id, left_col_idx, left_col) = left_col.unwrap();
@@ -913,9 +1101,9 @@ pub fn parse_limit(limit: &Limit) -> Result<(Option<isize>, Option<isize>)> {
             crate::bail_parse_error!("Invalid LIMIT clause");
         }
     } else if let Expr::Id(id) = &limit.expr {
-        if id.0.eq_ignore_ascii_case("true") {
+        if id.as_str().eq_ignore_ascii_case("true") {
             Ok((Some(1), offset_val))
-        } else if id.0.eq_ignore_ascii_case("false") {
+        } else if id.as_str().eq_ignore_ascii_case("false") {
             Ok((Some(0), offset_val))
         } else {
             crate::bail_parse_error!("Invalid LIMIT clause");
