@@ -1,6 +1,7 @@
 //! VDBE bytecode generation for pragma statements.
 //! More info: https://www.sqlite.org/pragma.html.
 
+use chrono::Datelike;
 use std::rc::Rc;
 use std::sync::Arc;
 use turso_sqlite3_parser::ast::{self, ColumnDefinition, Expr};
@@ -9,7 +10,7 @@ use turso_sqlite3_parser::ast::{PragmaName, QualifiedName};
 use crate::pragma::pragma_for;
 use crate::schema::Schema;
 use crate::storage::pager::AutoVacuumMode;
-use crate::storage::sqlite3_ondisk::MIN_PAGE_CACHE_SIZE;
+use crate::storage::sqlite3_ondisk::{DatabaseEncoding, MIN_PAGE_CACHE_SIZE};
 use crate::storage::wal::CheckpointMode;
 use crate::translate::schema::translate_create_table;
 use crate::util::{normalize_ident, parse_signed_number, parse_string};
@@ -22,6 +23,7 @@ use strum::IntoEnumIterator;
 use super::integrity_check::translate_integrity_check;
 use crate::storage::header_accessor;
 use crate::storage::pager::Pager;
+use crate::translate::emitter::TransactionMode;
 
 fn list_pragmas(program: &mut ProgramBuilder) {
     for x in PragmaName::iter() {
@@ -29,7 +31,7 @@ fn list_pragmas(program: &mut ProgramBuilder) {
         program.emit_result_row(register, 1);
     }
     program.add_pragma_result_column("pragma_list".into());
-    program.epilogue(crate::translate::emitter::TransactionMode::None);
+    program.epilogue(TransactionMode::None);
 }
 
 #[allow(clippy::too_many_arguments)]
@@ -47,34 +49,27 @@ pub fn translate_pragma(
         approx_num_labels: 0,
     };
     program.extend(&opts);
-    let mut write = false;
 
-    if name.name.0.eq_ignore_ascii_case("pragma_list") {
+    if name.name.as_str().eq_ignore_ascii_case("pragma_list") {
         list_pragmas(&mut program);
         return Ok(program);
     }
 
-    let pragma = match PragmaName::from_str(&name.name.0) {
+    let pragma = match PragmaName::from_str(name.name.as_str()) {
         Ok(pragma) => pragma,
         Err(_) => bail_parse_error!("Not a valid pragma name"),
     };
 
-    let mut program = match body {
+    let (mut program, mode) = match body {
         None => query_pragma(pragma, schema, None, pager, connection, program)?,
         Some(ast::PragmaBody::Equals(value) | ast::PragmaBody::Call(value)) => match pragma {
             PragmaName::TableInfo => {
                 query_pragma(pragma, schema, Some(value), pager, connection, program)?
             }
-            _ => {
-                write = true;
-                update_pragma(pragma, schema, value, pager, connection, program)?
-            }
+            _ => update_pragma(pragma, schema, value, pager, connection, program)?,
         },
     };
-    program.epilogue(match write {
-        false => super::emitter::TransactionMode::Read,
-        true => super::emitter::TransactionMode::Write,
-    });
+    program.epilogue(mode);
 
     Ok(program)
 }
@@ -86,8 +81,24 @@ fn update_pragma(
     pager: Rc<Pager>,
     connection: Arc<crate::Connection>,
     mut program: ProgramBuilder,
-) -> crate::Result<ProgramBuilder> {
+) -> crate::Result<(ProgramBuilder, TransactionMode)> {
     match pragma {
+        PragmaName::ApplicationId => {
+            let data = parse_signed_number(&value)?;
+            let app_id_value = match data {
+                Value::Integer(i) => i as i32,
+                Value::Float(f) => f as i32,
+                _ => unreachable!(),
+            };
+
+            program.emit_insn(Insn::SetCookie {
+                db: 0,
+                cookie: Cookie::ApplicationId,
+                value: app_id_value,
+                p5: 1,
+            });
+            Ok((program, TransactionMode::Write))
+        }
         PragmaName::CacheSize => {
             let cache_size = match parse_signed_number(&value)? {
                 Value::Integer(size) => size,
@@ -95,7 +106,11 @@ fn update_pragma(
                 _ => bail_parse_error!("Invalid value for cache size pragma"),
             };
             update_cache_size(cache_size, pager, connection)?;
-            Ok(program)
+            Ok((program, TransactionMode::None))
+        }
+        PragmaName::Encoding => {
+            let year = chrono::Local::now().year();
+            bail_parse_error!("It's {year}. UTF-8 won.");
         }
         PragmaName::JournalMode => query_pragma(
             PragmaName::JournalMode,
@@ -105,7 +120,7 @@ fn update_pragma(
             connection,
             program,
         ),
-        PragmaName::LegacyFileFormat => Ok(program),
+        PragmaName::LegacyFileFormat => Ok((program, TransactionMode::None)),
         PragmaName::WalCheckpoint => query_pragma(
             PragmaName::WalCheckpoint,
             schema,
@@ -136,11 +151,13 @@ fn update_pragma(
                 value: version_value,
                 p5: 1,
             });
-            Ok(program)
+            Ok((program, TransactionMode::Write))
         }
         PragmaName::SchemaVersion => {
-            // TODO: Implement updating schema_version
-            todo!("updating schema_version not yet implemented")
+            // SQLite allowing this to be set is an incredibly stupid idea in my view.
+            // In "defensive mode", this is a silent nop. So let's emulate that always.
+            program.emit_insn(Insn::Noop {});
+            Ok((program, TransactionMode::None))
         }
         PragmaName::TableInfo => {
             // because we need control over the write parameter for the transaction,
@@ -149,12 +166,18 @@ fn update_pragma(
             unreachable!();
         }
         PragmaName::PageSize => {
-            bail_parse_error!("Updating database page size is not supported.");
+            let page_size = match parse_signed_number(&value)? {
+                Value::Integer(size) => size,
+                Value::Float(size) => size as i64,
+                _ => bail_parse_error!("Invalid value for page size pragma"),
+            };
+            update_page_size(connection, page_size as u32)?;
+            Ok((program, TransactionMode::None))
         }
         PragmaName::AutoVacuum => {
             let auto_vacuum_mode = match value {
                 Expr::Name(name) => {
-                    let name = name.0.to_lowercase();
+                    let name = name.as_str().to_lowercase();
                     match name.as_str() {
                         "none" => 0,
                         "full" => 1,
@@ -205,7 +228,7 @@ fn update_pragma(
                 value: auto_vacuum_mode - 1,
                 p5: 0,
             });
-            Ok(program)
+            Ok((program, TransactionMode::None))
         }
         PragmaName::IntegrityCheck => unreachable!("integrity_check cannot be set"),
         PragmaName::UnstableCaptureDataChangesConn => {
@@ -216,7 +239,7 @@ fn update_pragma(
             if let Some(table) = &opts.table() {
                 // make sure that we have table created
                 program = translate_create_table(
-                    QualifiedName::single(ast::Name(table.to_string())),
+                    QualifiedName::single(ast::Name::from_str(table)),
                     false,
                     ast::CreateTableBody::columns_and_constraints_from_definition(
                         turso_cdc_table_columns(),
@@ -230,8 +253,9 @@ fn update_pragma(
                 )?;
             }
             connection.set_capture_data_changes(opts);
-            Ok(program)
+            Ok((program, TransactionMode::Write))
         }
+        PragmaName::DatabaseList => unreachable!("database_list cannot be set"),
     }
 }
 
@@ -242,26 +266,77 @@ fn query_pragma(
     pager: Rc<Pager>,
     connection: Arc<crate::Connection>,
     mut program: ProgramBuilder,
-) -> crate::Result<ProgramBuilder> {
+) -> crate::Result<(ProgramBuilder, TransactionMode)> {
     let register = program.alloc_register();
     match pragma {
+        PragmaName::ApplicationId => {
+            program.emit_insn(Insn::ReadCookie {
+                db: 0,
+                dest: register,
+                cookie: Cookie::ApplicationId,
+            });
+            program.add_pragma_result_column(pragma.to_string());
+            program.emit_result_row(register, 1);
+            Ok((program, TransactionMode::Read))
+        }
         PragmaName::CacheSize => {
             program.emit_int(connection.get_cache_size() as i64, register);
             program.emit_result_row(register, 1);
             program.add_pragma_result_column(pragma.to_string());
+            Ok((program, TransactionMode::None))
+        }
+        PragmaName::DatabaseList => {
+            let base_reg = register;
+            program.alloc_registers(2);
+
+            // Get all databases (main + attached) and emit a row for each
+            let all_databases = connection.list_all_databases();
+            for (seq_number, name, file_path) in all_databases {
+                // seq (sequence number)
+                program.emit_int(seq_number as i64, base_reg);
+
+                // name (alias)
+                program.emit_string8(name, base_reg + 1);
+
+                // file path
+                program.emit_string8(file_path, base_reg + 2);
+
+                program.emit_result_row(base_reg, 3);
+            }
+
+            let pragma = pragma_for(&pragma);
+            for col_name in pragma.columns.iter() {
+                program.add_pragma_result_column(col_name.to_string());
+            }
+            Ok((program, TransactionMode::None))
+        }
+        PragmaName::Encoding => {
+            let encoding: &str = if !pager.db_state.is_initialized() {
+                DatabaseEncoding::Utf8
+            } else {
+                let encoding: DatabaseEncoding =
+                    header_accessor::get_text_encoding(&pager)?.try_into()?;
+                encoding
+            }
+            .into();
+            program.emit_string8(encoding.into(), register);
+            program.emit_result_row(register, 1);
+            program.add_pragma_result_column(pragma.to_string());
+            Ok((program, TransactionMode::None))
         }
         PragmaName::JournalMode => {
             program.emit_string8("wal".into(), register);
             program.emit_result_row(register, 1);
             program.add_pragma_result_column(pragma.to_string());
+            Ok((program, TransactionMode::None))
         }
-        PragmaName::LegacyFileFormat => {}
+        PragmaName::LegacyFileFormat => Ok((program, TransactionMode::None)),
         PragmaName::WalCheckpoint => {
             // Checkpoint uses 3 registers: P1, P2, P3. Ref Insn::Checkpoint for more info.
             // Allocate two more here as one was allocated at the top.
             let mode = match value {
                 Some(ast::Expr::Name(name)) => {
-                    let mode_name = normalize_ident(&name.0);
+                    let mode_name = normalize_ident(name.as_str());
                     CheckpointMode::from_str(&mode_name).map_err(|e| {
                         LimboError::ParseError(format!("Unknown Checkpoint Mode: {e}"))
                     })?
@@ -282,6 +357,7 @@ fn query_pragma(
                 dest: register,
             });
             program.emit_result_row(register, 3);
+            Ok((program, TransactionMode::None))
         }
         PragmaName::PageCount => {
             program.emit_insn(Insn::PageCount {
@@ -290,11 +366,12 @@ fn query_pragma(
             });
             program.emit_result_row(register, 1);
             program.add_pragma_result_column(pragma.to_string());
+            Ok((program, TransactionMode::Read))
         }
         PragmaName::TableInfo => {
             let table = match value {
                 Some(ast::Expr::Name(name)) => {
-                    let tbl = normalize_ident(&name.0);
+                    let tbl = normalize_ident(name.as_str());
                     schema.get_table(&tbl)
                 }
                 _ => None,
@@ -303,7 +380,10 @@ fn query_pragma(
             let base_reg = register;
             program.alloc_registers(5);
             if let Some(table) = table {
-                for (i, column) in table.columns().iter().enumerate() {
+                // According to the SQLite documentation: "The 'cid' column should not be taken to
+                // mean more than 'rank within the current result set'."
+                // Therefore, we enumerate only after filtering out hidden columns.
+                for (i, column) in table.columns().iter().filter(|col| !col.hidden).enumerate() {
                     // cid
                     program.emit_int(i as i64, base_reg);
                     // name
@@ -335,6 +415,7 @@ fn query_pragma(
             for name in col_names {
                 program.add_pragma_result_column(name.into());
             }
+            Ok((program, TransactionMode::None))
         }
         PragmaName::UserVersion => {
             program.emit_insn(Insn::ReadCookie {
@@ -344,6 +425,7 @@ fn query_pragma(
             });
             program.add_pragma_result_column(pragma.to_string());
             program.emit_result_row(register, 1);
+            Ok((program, TransactionMode::Read))
         }
         PragmaName::SchemaVersion => {
             program.emit_insn(Insn::ReadCookie {
@@ -353,15 +435,16 @@ fn query_pragma(
             });
             program.add_pragma_result_column(pragma.to_string());
             program.emit_result_row(register, 1);
+            Ok((program, TransactionMode::Read))
         }
         PragmaName::PageSize => {
             program.emit_int(
-                header_accessor::get_page_size(&pager)
-                    .unwrap_or(storage::sqlite3_ondisk::DEFAULT_PAGE_SIZE) as i64,
+                header_accessor::get_page_size(&pager).unwrap_or(connection.get_page_size()) as i64,
                 register,
             );
             program.emit_result_row(register, 1);
             program.add_pragma_result_column(pragma.to_string());
+            Ok((program, TransactionMode::None))
         }
         PragmaName::AutoVacuum => {
             let auto_vacuum_mode = pager.get_auto_vacuum_mode();
@@ -378,12 +461,14 @@ fn query_pragma(
                 value: auto_vacuum_mode_i64,
             });
             program.emit_result_row(register, 1);
+            Ok((program, TransactionMode::None))
         }
         PragmaName::IntegrityCheck => {
             translate_integrity_check(schema, &mut program)?;
+            Ok((program, TransactionMode::Read))
         }
         PragmaName::UnstableCaptureDataChangesConn => {
-            let pragma = pragma_for(pragma);
+            let pragma = pragma_for(&pragma);
             let second_column = program.alloc_register();
             let opts = connection.get_capture_data_changes();
             program.emit_string8(opts.mode_name().to_string(), register);
@@ -395,10 +480,9 @@ fn query_pragma(
             program.emit_result_row(register, 2);
             program.add_pragma_result_column(pragma.columns[0].to_string());
             program.add_pragma_result_column(pragma.columns[1].to_string());
+            Ok((program, TransactionMode::Read))
         }
     }
-
-    Ok(program)
 }
 
 fn update_auto_vacuum_mode(
@@ -467,7 +551,7 @@ pub const TURSO_CDC_DEFAULT_TABLE_NAME: &str = "turso_cdc";
 fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
     vec![
         ast::ColumnDefinition {
-            col_name: ast::Name("operation_id".to_string()),
+            col_name: ast::Name::from_str("change_id"),
             col_type: Some(ast::Type {
                 name: "INTEGER".to_string(),
                 size: None,
@@ -482,7 +566,7 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
             }],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("operation_time".to_string()),
+            col_name: ast::Name::from_str("change_time"),
             col_type: Some(ast::Type {
                 name: "INTEGER".to_string(),
                 size: None,
@@ -490,7 +574,7 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
             constraints: vec![],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("operation_type".to_string()),
+            col_name: ast::Name::from_str("change_type"),
             col_type: Some(ast::Type {
                 name: "INTEGER".to_string(),
                 size: None,
@@ -498,7 +582,7 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
             constraints: vec![],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("table_name".to_string()),
+            col_name: ast::Name::from_str("table_name"),
             col_type: Some(ast::Type {
                 name: "TEXT".to_string(),
                 size: None,
@@ -506,12 +590,12 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
             constraints: vec![],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("id".to_string()),
+            col_name: ast::Name::from_str("id"),
             col_type: None,
             constraints: vec![],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("before".to_string()),
+            col_name: ast::Name::from_str("before"),
             col_type: Some(ast::Type {
                 name: "BLOB".to_string(),
                 size: None,
@@ -519,7 +603,7 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
             constraints: vec![],
         },
         ast::ColumnDefinition {
-            col_name: ast::Name("after".to_string()),
+            col_name: ast::Name::from_str("after"),
             col_type: Some(ast::Type {
                 name: "BLOB".to_string(),
                 size: None,
@@ -528,3 +612,8 @@ fn turso_cdc_table_columns() -> Vec<ColumnDefinition> {
         },
     ]
 }
+
+fn update_page_size(connection: Arc<crate::Connection>, page_size: u32) -> crate::Result<()> {
+    connection.reset_page_size(page_size)?;
+    Ok(())
+}
