@@ -16,6 +16,7 @@ use crate::vdbe::builder::{ProgramBuilderOpts, TableRefIdCounter};
 use crate::vdbe::insn::Insn;
 use crate::SymbolTable;
 use crate::{schema::Schema, vdbe::builder::ProgramBuilder, Result};
+use std::sync::Arc;
 use turso_sqlite3_parser::ast::{self, CompoundSelect, SortOrder};
 use turso_sqlite3_parser::ast::{ResultColumn, SelectInner};
 
@@ -30,6 +31,7 @@ pub fn translate_select(
     syms: &SymbolTable,
     mut program: ProgramBuilder,
     query_destination: QueryDestination,
+    connection: &Arc<crate::Connection>,
 ) -> Result<TranslateSelectResult> {
     let mut select_plan = prepare_select_plan(
         schema,
@@ -38,6 +40,7 @@ pub fn translate_select(
         &[],
         &mut program.table_reference_counter,
         query_destination,
+        connection,
     )?;
     optimize_plan(&mut select_plan, schema)?;
     let num_result_cols;
@@ -92,6 +95,7 @@ pub fn prepare_select_plan(
     outer_query_refs: &[OuterQueryReference],
     table_ref_counter: &mut TableRefIdCounter,
     query_destination: QueryDestination,
+    connection: &Arc<crate::Connection>,
 ) -> Result<Plan> {
     let compounds = select.body.compounds.take();
     match compounds {
@@ -107,6 +111,7 @@ pub fn prepare_select_plan(
                 outer_query_refs,
                 table_ref_counter,
                 query_destination,
+                connection,
             )?))
         }
         Some(compounds) => {
@@ -120,6 +125,7 @@ pub fn prepare_select_plan(
                 outer_query_refs,
                 table_ref_counter,
                 query_destination.clone(),
+                connection,
             )?;
 
             let mut left = Vec::with_capacity(compounds.len());
@@ -135,6 +141,7 @@ pub fn prepare_select_plan(
                     outer_query_refs,
                     table_ref_counter,
                     query_destination.clone(),
+                    connection,
                 )?;
             }
 
@@ -181,6 +188,7 @@ fn prepare_one_select_plan(
     outer_query_refs: &[OuterQueryReference],
     table_ref_counter: &mut TableRefIdCounter,
     query_destination: QueryDestination,
+    connection: &Arc<crate::Connection>,
 ) -> Result<SelectPlan> {
     match select {
         ast::OneSelect::Select(select_inner) => {
@@ -223,6 +231,7 @@ fn prepare_one_select_plan(
                 &mut where_predicates,
                 &mut table_references,
                 table_ref_counter,
+                connection,
             )?;
 
             // Preallocate space for the result columns
@@ -234,14 +243,14 @@ fn prepare_one_select_plan(
                         ResultColumn::Star => table_references
                             .joined_tables()
                             .iter()
-                            .map(|t| t.columns().len())
+                            .map(|t| t.columns().iter().filter(|col| !col.hidden).count())
                             .sum(),
                         // Guess 5 columns if we can't find the table using the identifier (maybe it's in [brackets] or `tick_quotes`, or miXeDcAse)
                         ResultColumn::TableStar(n) => table_references
                             .joined_tables()
                             .iter()
-                            .find(|t| t.identifier == n.0)
-                            .map(|t| t.columns().len())
+                            .find(|t| t.identifier == n.as_str())
+                            .map(|t| t.columns().iter().filter(|col| !col.hidden).count())
                             .unwrap_or(5),
                         // Otherwise allocate space for 1 column
                         ResultColumn::Expr(_, _) => 1,
@@ -284,12 +293,16 @@ fn prepare_one_select_plan(
                         );
                         for table in plan.table_references.joined_tables_mut() {
                             for idx in 0..table.columns().len() {
+                                let column = &table.columns()[idx];
+                                if column.hidden {
+                                    continue;
+                                }
                                 table.mark_column_used(idx);
                             }
                         }
                     }
                     ResultColumn::TableStar(name) => {
-                        let name_normalized = normalize_ident(name.0.as_str());
+                        let name_normalized = normalize_ident(name.as_str());
                         let referenced_table = plan
                             .table_references
                             .joined_tables_mut()
@@ -297,21 +310,21 @@ fn prepare_one_select_plan(
                             .find(|t| t.identifier == name_normalized);
 
                         if referenced_table.is_none() {
-                            crate::bail_parse_error!("no such table: {}", name.0);
+                            crate::bail_parse_error!("no such table: {}", name.as_str());
                         }
                         let table = referenced_table.unwrap();
                         let num_columns = table.columns().len();
                         for idx in 0..num_columns {
-                            let is_rowid_alias = {
-                                let columns = table.columns();
-                                columns[idx].is_rowid_alias
-                            };
+                            let column = &table.columns()[idx];
+                            if column.hidden {
+                                continue;
+                            }
                             plan.result_columns.push(ResultSetColumn {
                                 expr: ast::Expr::Column {
                                     database: None, // TODO: support different databases
                                     table: table.internal_id,
                                     column: idx,
-                                    is_rowid_alias,
+                                    is_rowid_alias: column.is_rowid_alias,
                                 },
                                 alias: None,
                                 contains_aggregates: false,
@@ -324,6 +337,7 @@ fn prepare_one_select_plan(
                             expr,
                             &mut plan.table_references,
                             Some(&plan.result_columns),
+                            connection,
                         )?;
                         match expr {
                             ast::Expr::FunctionCall {
@@ -348,7 +362,7 @@ fn prepare_one_select_plan(
                                 if distinctness.is_distinct() && args_count != 1 {
                                     crate::bail_parse_error!("DISTINCT aggregate functions must have exactly one argument");
                                 }
-                                match Func::resolve_function(&name.0, args_count) {
+                                match Func::resolve_function(name.as_str(), args_count) {
                                     Ok(Func::Agg(f)) => {
                                         let agg_args = match (args, &f) {
                                             (None, crate::function::AggFunc::Count0) => {
@@ -359,7 +373,7 @@ fn prepare_one_select_plan(
                                             }
                                             (None, _) => crate::bail_parse_error!(
                                                 "Aggregate function {} requires arguments",
-                                                name.0
+                                                name.as_str()
                                             ),
                                             (Some(args), _) => args.clone(),
                                         };
@@ -373,8 +387,10 @@ fn prepare_one_select_plan(
                                         aggregate_expressions.push(agg.clone());
                                         plan.result_columns.push(ResultSetColumn {
                                             alias: maybe_alias.as_ref().map(|alias| match alias {
-                                                ast::As::Elided(alias) => alias.0.clone(),
-                                                ast::As::As(alias) => alias.0.clone(),
+                                                ast::As::Elided(alias) => {
+                                                    alias.as_str().to_string()
+                                                }
+                                                ast::As::As(alias) => alias.as_str().to_string(),
                                             }),
                                             expr: expr.clone(),
                                             contains_aggregates: true,
@@ -388,15 +404,18 @@ fn prepare_one_select_plan(
                                         )?;
                                         plan.result_columns.push(ResultSetColumn {
                                             alias: maybe_alias.as_ref().map(|alias| match alias {
-                                                ast::As::Elided(alias) => alias.0.clone(),
-                                                ast::As::As(alias) => alias.0.clone(),
+                                                ast::As::Elided(alias) => {
+                                                    alias.as_str().to_string()
+                                                }
+                                                ast::As::As(alias) => alias.as_str().to_string(),
                                             }),
                                             expr: expr.clone(),
                                             contains_aggregates,
                                         });
                                     }
                                     Err(e) => {
-                                        if let Some(f) = syms.resolve_function(&name.0, args_count)
+                                        if let Some(f) =
+                                            syms.resolve_function(name.as_str(), args_count)
                                         {
                                             if let ExtFunc::Scalar(_) = f.as_ref().func {
                                                 let contains_aggregates = resolve_aggregates(
@@ -408,9 +427,11 @@ fn prepare_one_select_plan(
                                                     alias: maybe_alias.as_ref().map(|alias| {
                                                         match alias {
                                                             ast::As::Elided(alias) => {
-                                                                alias.0.clone()
+                                                                alias.as_str().to_string()
+                                                            }
+                                                            ast::As::As(alias) => {
+                                                                alias.as_str().to_string()
                                                             }
-                                                            ast::As::As(alias) => alias.0.clone(),
                                                         }
                                                     }),
                                                     expr: expr.clone(),
@@ -428,9 +449,11 @@ fn prepare_one_select_plan(
                                                     alias: maybe_alias.as_ref().map(|alias| {
                                                         match alias {
                                                             ast::As::Elided(alias) => {
-                                                                alias.0.clone()
+                                                                alias.as_str().to_string()
+                                                            }
+                                                            ast::As::As(alias) => {
+                                                                alias.as_str().to_string()
                                                             }
-                                                            ast::As::As(alias) => alias.0.clone(),
                                                         }
                                                     }),
                                                     expr: expr.clone(),
@@ -447,7 +470,7 @@ fn prepare_one_select_plan(
                             ast::Expr::FunctionCallStar {
                                 name,
                                 filter_over: _,
-                            } => match Func::resolve_function(&name.0, 0) {
+                            } => match Func::resolve_function(name.as_str(), 0) {
                                 Ok(Func::Agg(f)) => {
                                     let agg = Aggregate {
                                         func: f,
@@ -460,8 +483,8 @@ fn prepare_one_select_plan(
                                     aggregate_expressions.push(agg.clone());
                                     plan.result_columns.push(ResultSetColumn {
                                         alias: maybe_alias.as_ref().map(|alias| match alias {
-                                            ast::As::Elided(alias) => alias.0.clone(),
-                                            ast::As::As(alias) => alias.0.clone(),
+                                            ast::As::Elided(alias) => alias.as_str().to_string(),
+                                            ast::As::As(alias) => alias.as_str().to_string(),
                                         }),
                                         expr: expr.clone(),
                                         contains_aggregates: true,
@@ -470,7 +493,7 @@ fn prepare_one_select_plan(
                                 Ok(_) => {
                                     crate::bail_parse_error!(
                                         "Invalid aggregate function: {}",
-                                        name.0
+                                        name.as_str()
                                     );
                                 }
                                 Err(e) => match e {
@@ -480,7 +503,7 @@ fn prepare_one_select_plan(
                                     _ => {
                                         crate::bail_parse_error!(
                                             "Invalid aggregate function: {}",
-                                            name.0
+                                            name.as_str()
                                         );
                                     }
                                 },
@@ -490,8 +513,8 @@ fn prepare_one_select_plan(
                                     resolve_aggregates(schema, expr, &mut aggregate_expressions)?;
                                 plan.result_columns.push(ResultSetColumn {
                                     alias: maybe_alias.as_ref().map(|alias| match alias {
-                                        ast::As::Elided(alias) => alias.0.clone(),
-                                        ast::As::As(alias) => alias.0.clone(),
+                                        ast::As::Elided(alias) => alias.as_str().to_string(),
+                                        ast::As::As(alias) => alias.as_str().to_string(),
                                     }),
                                     expr: expr.clone(),
                                     contains_aggregates,
@@ -508,6 +531,7 @@ fn prepare_one_select_plan(
                 &mut plan.table_references,
                 Some(&plan.result_columns),
                 &mut plan.where_clause,
+                connection,
             )?;
 
             if let Some(mut group_by) = group_by {
@@ -517,6 +541,7 @@ fn prepare_one_select_plan(
                         expr,
                         &mut plan.table_references,
                         Some(&plan.result_columns),
+                        connection,
                     )?;
                 }
 
@@ -531,6 +556,7 @@ fn prepare_one_select_plan(
                                 expr,
                                 &mut plan.table_references,
                                 Some(&plan.result_columns),
+                                connection,
                             )?;
                             let contains_aggregates =
                                 resolve_aggregates(schema, expr, &mut aggregate_expressions)?;
@@ -567,6 +593,7 @@ fn prepare_one_select_plan(
                         &mut o.expr,
                         &mut plan.table_references,
                         Some(&plan.result_columns),
+                        connection,
                     )?;
                     resolve_aggregates(schema, &o.expr, &mut plan.aggregates)?;
 
@@ -737,7 +764,7 @@ pub fn emit_simple_count(
     program.emit_insn(Insn::Copy {
         src_reg: target_reg,
         dst_reg: output_reg,
-        amount: 0,
+        extra_amount: 0,
     });
     program.emit_result_row(output_reg, 1);
     Ok(())
