@@ -1,15 +1,35 @@
+#![allow(unused)]
+use crate::storage::header_accessor::get_schema_cookie;
 use crate::translate::expr::WalkControl;
+use crate::types::IOResult;
 use crate::{
     schema::{self, Column, Schema, Type},
     translate::{collate::CollationSeq, expr::walk_expr, plan::JoinOrderMember},
     types::{Value, ValueType},
     LimboError, OpenFlags, Result, Statement, StepResult, SymbolTable,
 };
+use crate::{Connection, IO};
 use std::{rc::Rc, sync::Arc};
+use tracing::{instrument, Level};
 use turso_sqlite3_parser::ast::{
     self, CreateTableBody, Expr, FunctionTail, Literal, UnaryOperator,
 };
 
+pub trait IOExt {
+    fn block<T>(&self, f: impl FnMut() -> Result<IOResult<T>>) -> Result<T>;
+}
+
+impl<I: ?Sized + IO> IOExt for I {
+    fn block<T>(&self, mut f: impl FnMut() -> Result<IOResult<T>>) -> Result<T> {
+        Ok(loop {
+            match f()? {
+                IOResult::Done(v) => break v,
+                IOResult::IO => self.run_once()?,
+            }
+        })
+    }
+}
+
 pub trait RoundToPrecision {
     fn round_to_precision(self, precision: i32) -> f64;
 }
@@ -48,123 +68,121 @@ pub struct UnparsedFromSqlIndex {
     pub sql: String,
 }
 
+#[instrument(skip_all, level = Level::INFO)]
 pub fn parse_schema_rows(
-    rows: Option<Statement>,
+    mut rows: Statement,
     schema: &mut Schema,
     syms: &SymbolTable,
     mv_tx_id: Option<u64>,
 ) -> Result<()> {
-    if let Some(mut rows) = rows {
-        rows.set_mv_tx_id(mv_tx_id);
-        // TODO: if we IO, this unparsed indexes is lost. Will probably need some state between
-        // IO runs
-        let mut from_sql_indexes = Vec::with_capacity(10);
-        let mut automatic_indices: std::collections::HashMap<String, Vec<(String, usize)>> =
-            std::collections::HashMap::with_capacity(10);
-        loop {
-            match rows.step()? {
-                StepResult::Row => {
-                    let row = rows.row().unwrap();
-                    let ty = row.get::<&str>(0)?;
-                    if !["table", "index"].contains(&ty) {
-                        continue;
-                    }
-                    match ty {
-                        "table" => {
-                            let root_page: i64 = row.get::<i64>(3)?;
-                            let sql: &str = row.get::<&str>(4)?;
-                            if root_page == 0 && sql.to_lowercase().contains("create virtual") {
-                                let name: &str = row.get::<&str>(1)?;
-                                // a virtual table is found in the sqlite_schema, but it's no
-                                // longer in the in-memory schema. We need to recreate it if
-                                // the module is loaded in the symbol table.
-                                let vtab = if let Some(vtab) = syms.vtabs.get(name) {
-                                    vtab.clone()
-                                } else {
-                                    let mod_name = module_name_from_sql(sql)?;
-                                    crate::VirtualTable::table(
-                                        Some(name),
-                                        mod_name,
-                                        module_args_from_sql(sql)?,
-                                        syms,
-                                    )?
-                                };
-                                schema.add_virtual_table(vtab);
+    rows.set_mv_tx_id(mv_tx_id);
+    // TODO: if we IO, this unparsed indexes is lost. Will probably need some state between
+    // IO runs
+    let mut from_sql_indexes = Vec::with_capacity(10);
+    let mut automatic_indices = std::collections::HashMap::with_capacity(10);
+    loop {
+        match rows.step()? {
+            StepResult::Row => {
+                let row = rows.row().unwrap();
+                let ty = row.get::<&str>(0)?;
+                if !["table", "index"].contains(&ty) {
+                    continue;
+                }
+                match ty {
+                    "table" => {
+                        let root_page: i64 = row.get::<i64>(3)?;
+                        let sql: &str = row.get::<&str>(4)?;
+                        if root_page == 0 && sql.to_lowercase().contains("create virtual") {
+                            let name: &str = row.get::<&str>(1)?;
+                            // a virtual table is found in the sqlite_schema, but it's no
+                            // longer in the in-memory schema. We need to recreate it if
+                            // the module is loaded in the symbol table.
+                            let vtab = if let Some(vtab) = syms.vtabs.get(name) {
+                                vtab.clone()
                             } else {
-                                let table = schema::BTreeTable::from_sql(sql, root_page as usize)?;
-                                schema.add_btree_table(Rc::new(table));
-                            }
+                                let mod_name = module_name_from_sql(sql)?;
+                                crate::VirtualTable::table(
+                                    Some(name),
+                                    mod_name,
+                                    module_args_from_sql(sql)?,
+                                    syms,
+                                )?
+                            };
+                            schema.add_virtual_table(vtab);
+                        } else {
+                            let table = schema::BTreeTable::from_sql(sql, root_page as usize)?;
+                            schema.add_btree_table(Arc::new(table));
                         }
-                        "index" => {
-                            let root_page: i64 = row.get::<i64>(3)?;
-                            match row.get::<&str>(4) {
-                                Ok(sql) => {
-                                    from_sql_indexes.push(UnparsedFromSqlIndex {
-                                        table_name: row.get::<&str>(2)?.to_string(),
-                                        root_page: root_page as usize,
-                                        sql: sql.to_string(),
-                                    });
-                                }
-                                _ => {
-                                    // Automatic index on primary key and/or unique constraint, e.g.
-                                    // table|foo|foo|2|CREATE TABLE foo (a text PRIMARY KEY, b)
-                                    // index|sqlite_autoindex_foo_1|foo|3|
-                                    let index_name = row.get::<&str>(1)?.to_string();
-                                    let table_name = row.get::<&str>(2)?.to_string();
-                                    let root_page = row.get::<i64>(3)?;
-                                    match automatic_indices.entry(table_name) {
-                                        std::collections::hash_map::Entry::Vacant(e) => {
-                                            e.insert(vec![(index_name, root_page as usize)]);
-                                        }
-                                        std::collections::hash_map::Entry::Occupied(mut e) => {
-                                            e.get_mut().push((index_name, root_page as usize));
-                                        }
+                    }
+                    "index" => {
+                        let root_page: i64 = row.get::<i64>(3)?;
+                        match row.get::<&str>(4) {
+                            Ok(sql) => {
+                                from_sql_indexes.push(UnparsedFromSqlIndex {
+                                    table_name: row.get::<&str>(2)?.to_string(),
+                                    root_page: root_page as usize,
+                                    sql: sql.to_string(),
+                                });
+                            }
+                            _ => {
+                                // Automatic index on primary key and/or unique constraint, e.g.
+                                // table|foo|foo|2|CREATE TABLE foo (a text PRIMARY KEY, b)
+                                // index|sqlite_autoindex_foo_1|foo|3|
+                                let index_name = row.get::<&str>(1)?.to_string();
+                                let table_name = row.get::<&str>(2)?.to_string();
+                                let root_page = row.get::<i64>(3)?;
+                                match automatic_indices.entry(table_name) {
+                                    std::collections::hash_map::Entry::Vacant(e) => {
+                                        e.insert(vec![(index_name, root_page as usize)]);
+                                    }
+                                    std::collections::hash_map::Entry::Occupied(mut e) => {
+                                        e.get_mut().push((index_name, root_page as usize));
                                     }
                                 }
                             }
                         }
-                        _ => continue,
                     }
+                    _ => continue,
                 }
-                StepResult::IO => {
-                    // TODO: How do we ensure that the I/O we submitted to
-                    // read the schema is actually complete?
-                    rows.run_once()?;
-                }
-                StepResult::Interrupt => break,
-                StepResult::Done => break,
-                StepResult::Busy => break,
             }
+            StepResult::IO => {
+                // TODO: How do we ensure that the I/O we submitted to
+                // read the schema is actually complete?
+                rows.run_once()?;
+            }
+            StepResult::Interrupt => break,
+            StepResult::Done => break,
+            StepResult::Busy => break,
         }
-        for unparsed_sql_from_index in from_sql_indexes {
-            if !schema.indexes_enabled() {
-                schema.table_set_has_index(&unparsed_sql_from_index.table_name);
-            } else {
-                let table = schema
-                    .get_btree_table(&unparsed_sql_from_index.table_name)
-                    .unwrap();
-                let index = schema::Index::from_sql(
-                    &unparsed_sql_from_index.sql,
-                    unparsed_sql_from_index.root_page,
-                    table.as_ref(),
-                )?;
+    }
+    for unparsed_sql_from_index in from_sql_indexes {
+        if !schema.indexes_enabled() {
+            schema.table_set_has_index(&unparsed_sql_from_index.table_name);
+        } else {
+            let table = schema
+                .get_btree_table(&unparsed_sql_from_index.table_name)
+                .unwrap();
+            let index = schema::Index::from_sql(
+                &unparsed_sql_from_index.sql,
+                unparsed_sql_from_index.root_page,
+                table.as_ref(),
+            )?;
+            schema.add_index(Arc::new(index));
+        }
+    }
+    for automatic_index in automatic_indices {
+        if !schema.indexes_enabled() {
+            schema.table_set_has_index(&automatic_index.0);
+        } else {
+            let table = schema.get_btree_table(&automatic_index.0).unwrap();
+            let ret_index = schema::Index::automatic_from_primary_key_and_unique(
+                table.as_ref(),
+                automatic_index.1,
+            )?;
+            for index in ret_index {
                 schema.add_index(Arc::new(index));
             }
         }
-        for automatic_index in automatic_indices {
-            if !schema.indexes_enabled() {
-                schema.table_set_has_index(&automatic_index.0);
-            } else {
-                let table = schema.get_btree_table(&automatic_index.0).unwrap();
-                let ret_index = schema::Index::automatic_from_primary_key_and_unique(
-                    table.as_ref(),
-                    automatic_index.1,
-                )?;
-                for index in ret_index {
-                    schema.add_index(Arc::new(index));
-                }
-            }
-        }
     }
     Ok(())
 }
@@ -382,7 +400,7 @@ pub fn exprs_are_equivalent(expr1: &Expr, expr2: &Expr) -> bool {
                 filter_over: filter2,
             },
         ) => {
-            name1.0.eq_ignore_ascii_case(&name2.0)
+            name1.as_str().eq_ignore_ascii_case(name2.as_str())
                 && distinct1 == distinct2
                 && args1 == args2
                 && order1 == order2
@@ -398,7 +416,7 @@ pub fn exprs_are_equivalent(expr1: &Expr, expr2: &Expr) -> bool {
                 filter_over: filter2,
             },
         ) => {
-            name1.0.eq_ignore_ascii_case(&name2.0)
+            name1.as_str().eq_ignore_ascii_case(name2.as_str())
                 && match (filter1, filter2) {
                     (None, None) => true,
                     (
@@ -423,7 +441,7 @@ pub fn exprs_are_equivalent(expr1: &Expr, expr2: &Expr) -> bool {
         (Expr::NotNull(expr1), Expr::NotNull(expr2)) => exprs_are_equivalent(expr1, expr2),
         (Expr::IsNull(expr1), Expr::IsNull(expr2)) => exprs_are_equivalent(expr1, expr2),
         (Expr::Literal(lit1), Expr::Literal(lit2)) => check_literal_equivalency(lit1, lit2),
-        (Expr::Id(id1), Expr::Id(id2)) => check_ident_equivalency(&id1.0, &id2.0),
+        (Expr::Id(id1), Expr::Id(id2)) => check_ident_equivalency(id1.as_str(), id2.as_str()),
         (Expr::Unary(op1, expr1), Expr::Unary(op2, expr2)) => {
             op1 == op2 && exprs_are_equivalent(expr1, expr2)
         }
@@ -443,12 +461,13 @@ pub fn exprs_are_equivalent(expr1: &Expr, expr2: &Expr) -> bool {
             exprs1.len() == 1 && exprs_are_equivalent(&exprs1[0], exprs2)
         }
         (Expr::Qualified(tn1, cn1), Expr::Qualified(tn2, cn2)) => {
-            check_ident_equivalency(&tn1.0, &tn2.0) && check_ident_equivalency(&cn1.0, &cn2.0)
+            check_ident_equivalency(tn1.as_str(), tn2.as_str())
+                && check_ident_equivalency(cn1.as_str(), cn2.as_str())
         }
         (Expr::DoublyQualified(sn1, tn1, cn1), Expr::DoublyQualified(sn2, tn2, cn2)) => {
-            check_ident_equivalency(&sn1.0, &sn2.0)
-                && check_ident_equivalency(&tn1.0, &tn2.0)
-                && check_ident_equivalency(&cn1.0, &cn2.0)
+            check_ident_equivalency(sn1.as_str(), sn2.as_str())
+                && check_ident_equivalency(tn1.as_str(), tn2.as_str())
+                && check_ident_equivalency(cn1.as_str(), cn2.as_str())
         }
         (
             Expr::InList {
@@ -490,91 +509,90 @@ pub fn columns_from_create_table_body(body: &ast::CreateTableBody) -> crate::Res
 
     Ok(columns
         .into_iter()
-        .filter_map(|(name, column_def)| {
-            // if column_def.col_type includes HIDDEN, omit it for now
-            if let Some(data_type) = column_def.col_type.as_ref() {
-                if data_type.name.as_str().contains("HIDDEN") {
-                    return None;
-                }
-            }
-            let column =
-                Column {
-                    name: Some(normalize_ident(&name.0)),
-                    ty: match column_def.col_type {
-                        Some(ref data_type) => {
-                            // https://www.sqlite.org/datatype3.html
-                            let type_name = data_type.name.as_str().to_uppercase();
-                            if type_name.contains("INT") {
-                                Type::Integer
-                            } else if type_name.contains("CHAR")
-                                || type_name.contains("CLOB")
-                                || type_name.contains("TEXT")
-                            {
-                                Type::Text
-                            } else if type_name.contains("BLOB") || type_name.is_empty() {
-                                Type::Blob
-                            } else if type_name.contains("REAL")
-                                || type_name.contains("FLOA")
-                                || type_name.contains("DOUB")
-                            {
-                                Type::Real
-                            } else {
-                                Type::Numeric
-                            }
+        .map(|(name, column_def)| {
+            Column {
+                name: Some(normalize_ident(name.as_str())),
+                ty: match column_def.col_type {
+                    Some(ref data_type) => {
+                        // https://www.sqlite.org/datatype3.html
+                        let type_name = data_type.name.as_str().to_uppercase();
+                        if type_name.contains("INT") {
+                            Type::Integer
+                        } else if type_name.contains("CHAR")
+                            || type_name.contains("CLOB")
+                            || type_name.contains("TEXT")
+                        {
+                            Type::Text
+                        } else if type_name.contains("BLOB") || type_name.is_empty() {
+                            Type::Blob
+                        } else if type_name.contains("REAL")
+                            || type_name.contains("FLOA")
+                            || type_name.contains("DOUB")
+                        {
+                            Type::Real
+                        } else {
+                            Type::Numeric
                         }
-                        None => Type::Null,
-                    },
-                    default: column_def
-                        .constraints
-                        .iter()
-                        .find_map(|c| match &c.constraint {
-                            turso_sqlite3_parser::ast::ColumnConstraint::Default(val) => {
-                                Some(val.clone())
-                            }
-                            _ => None,
-                        }),
-                    notnull: column_def.constraints.iter().any(|c| {
-                        matches!(
-                            c.constraint,
-                            turso_sqlite3_parser::ast::ColumnConstraint::NotNull { .. }
-                        )
+                    }
+                    None => Type::Null,
+                },
+                default: column_def
+                    .constraints
+                    .iter()
+                    .find_map(|c| match &c.constraint {
+                        turso_sqlite3_parser::ast::ColumnConstraint::Default(val) => {
+                            Some(val.clone())
+                        }
+                        _ => None,
                     }),
-                    ty_str: column_def
-                        .col_type
-                        .clone()
-                        .map(|t| t.name.to_string())
-                        .unwrap_or_default(),
-                    primary_key: column_def.constraints.iter().any(|c| {
-                        matches!(
-                            c.constraint,
-                            turso_sqlite3_parser::ast::ColumnConstraint::PrimaryKey { .. }
-                        )
+                notnull: column_def.constraints.iter().any(|c| {
+                    matches!(
+                        c.constraint,
+                        turso_sqlite3_parser::ast::ColumnConstraint::NotNull { .. }
+                    )
+                }),
+                ty_str: column_def
+                    .col_type
+                    .clone()
+                    .map(|t| t.name.to_string())
+                    .unwrap_or_default(),
+                primary_key: column_def.constraints.iter().any(|c| {
+                    matches!(
+                        c.constraint,
+                        turso_sqlite3_parser::ast::ColumnConstraint::PrimaryKey { .. }
+                    )
+                }),
+                is_rowid_alias: false,
+                unique: column_def.constraints.iter().any(|c| {
+                    matches!(
+                        c.constraint,
+                        turso_sqlite3_parser::ast::ColumnConstraint::Unique(..)
+                    )
+                }),
+                collation: column_def
+                    .constraints
+                    .iter()
+                    .find_map(|c| match &c.constraint {
+                        // TODO: see if this should be the correct behavior
+                        // currently there cannot be any user defined collation sequences.
+                        // But in the future, when a user defines a collation sequence, creates a table with it,
+                        // then closes the db and opens it again. This may panic here if the collation seq is not registered
+                        // before reading the columns
+                        turso_sqlite3_parser::ast::ColumnConstraint::Collate { collation_name } => {
+                            Some(
+                                CollationSeq::new(collation_name.as_str()).expect(
+                                    "collation should have been set correctly in create table",
+                                ),
+                            )
+                        }
+                        _ => None,
                     }),
-                    is_rowid_alias: false,
-                    unique: column_def.constraints.iter().any(|c| {
-                        matches!(
-                            c.constraint,
-                            turso_sqlite3_parser::ast::ColumnConstraint::Unique(..)
-                        )
-                    }),
-                    collation: column_def
-                        .constraints
-                        .iter()
-                        .find_map(|c| match &c.constraint {
-                            // TODO: see if this should be the correct behavior
-                            // currently there cannot be any user defined collation sequences.
-                            // But in the future, when a user defines a collation sequence, creates a table with it,
-                            // then closes the db and opens it again. This may panic here if the collation seq is not registered
-                            // before reading the columns
-                            turso_sqlite3_parser::ast::ColumnConstraint::Collate {
-                                collation_name,
-                            } => Some(CollationSeq::new(collation_name.0.as_str()).expect(
-                                "collation should have been set correctly in create table",
-                            )),
-                            _ => None,
-                        }),
-                };
-            Some(column)
+                hidden: column_def
+                    .col_type
+                    .as_ref()
+                    .map(|data_type| data_type.name.as_str().contains("HIDDEN"))
+                    .unwrap_or(false),
+            }
         })
         .collect::<Vec<_>>())
 }
@@ -598,7 +616,7 @@ pub fn can_pushdown_predicate(
             }
             Expr::FunctionCall { args, name, .. } => {
                 let function = crate::function::Func::resolve_function(
-                    &name.0,
+                    name.as_str(),
                     args.as_ref().map_or(0, |a| a.len()),
                 )?;
                 // is deterministic
@@ -1043,7 +1061,9 @@ pub fn parse_signed_number(expr: &Expr) -> Result<Value> {
 
 pub fn parse_string(expr: &Expr) -> Result<String> {
     match expr {
-        Expr::Name(ast::Name(s)) if s.len() >= 2 && s.starts_with("'") && s.ends_with("'") => {
+        Expr::Name(ast::Name::Ident(s)) | Expr::Name(ast::Name::Quoted(s))
+            if s.len() >= 2 && s.starts_with("'") && s.ends_with("'") =>
+        {
             Ok(s[1..s.len() - 1].to_string())
         }
         _ => Err(LimboError::InvalidArgument(format!(
@@ -1061,7 +1081,7 @@ pub fn parse_pragma_bool(expr: &Expr) -> Result<bool> {
             return Ok(x != 0);
         }
     } else if let Expr::Name(name) = expr {
-        let ident = normalize_ident(&name.0);
+        let ident = normalize_ident(name.as_str());
         if TRUE_VALUES.contains(&ident.as_str()) {
             return Ok(true);
         }
@@ -1075,39 +1095,10 @@ pub fn parse_pragma_bool(expr: &Expr) -> Result<bool> {
     ))
 }
 
-// for TVF's we need these at planning time so we cannot emit translate_expr
-pub fn vtable_args(args: &[ast::Expr]) -> Vec<turso_ext::Value> {
-    let mut vtable_args = Vec::new();
-    for arg in args {
-        match arg {
-            Expr::Literal(lit) => match lit {
-                Literal::Numeric(i) => {
-                    if i.contains('.') {
-                        vtable_args.push(turso_ext::Value::from_float(i.parse().unwrap()));
-                    } else {
-                        vtable_args.push(turso_ext::Value::from_integer(i.parse().unwrap()));
-                    }
-                }
-                Literal::String(s) => {
-                    vtable_args.push(turso_ext::Value::from_text(s.clone()));
-                }
-                Literal::Blob(b) => {
-                    vtable_args.push(turso_ext::Value::from_blob(b.as_bytes().into()));
-                }
-                _ => {
-                    vtable_args.push(turso_ext::Value::null());
-                }
-            },
-            _ => vtable_args.push(turso_ext::Value::null()),
-        }
-    }
-    vtable_args
-}
-
 #[cfg(test)]
 pub mod tests {
     use super::*;
-    use turso_sqlite3_parser::ast::{self, Expr, Id, Literal, Name, Operator::*, Type};
+    use turso_sqlite3_parser::ast::{self, Expr, Literal, Name, Operator::*, Type};
 
     #[test]
     fn test_normalize_ident() {
@@ -1198,25 +1189,25 @@ pub mod tests {
     #[test]
     fn test_expressions_equivalent_case_insensitive_functioncalls() {
         let func1 = Expr::FunctionCall {
-            name: Id("SUM".to_string()),
+            name: Name::Ident("SUM".to_string()),
             distinctness: None,
-            args: Some(vec![Expr::Id(Id("x".to_string()))]),
+            args: Some(vec![Expr::Id(Name::Ident("x".to_string()))]),
             order_by: None,
             filter_over: None,
         };
         let func2 = Expr::FunctionCall {
-            name: Id("sum".to_string()),
+            name: Name::Ident("sum".to_string()),
             distinctness: None,
-            args: Some(vec![Expr::Id(Id("x".to_string()))]),
+            args: Some(vec![Expr::Id(Name::Ident("x".to_string()))]),
             order_by: None,
             filter_over: None,
         };
         assert!(exprs_are_equivalent(&func1, &func2));
 
         let func3 = Expr::FunctionCall {
-            name: Id("SUM".to_string()),
+            name: Name::Ident("SUM".to_string()),
             distinctness: Some(ast::Distinctness::Distinct),
-            args: Some(vec![Expr::Id(Id("x".to_string()))]),
+            args: Some(vec![Expr::Id(Name::Ident("x".to_string()))]),
             order_by: None,
             filter_over: None,
         };
@@ -1226,16 +1217,16 @@ pub mod tests {
     #[test]
     fn test_expressions_equivalent_identical_fn_with_distinct() {
         let sum = Expr::FunctionCall {
-            name: Id("SUM".to_string()),
+            name: Name::Ident("SUM".to_string()),
             distinctness: None,
-            args: Some(vec![Expr::Id(Id("x".to_string()))]),
+            args: Some(vec![Expr::Id(Name::Ident("x".to_string()))]),
             order_by: None,
             filter_over: None,
         };
         let sum_distinct = Expr::FunctionCall {
-            name: Id("SUM".to_string()),
+            name: Name::Ident("SUM".to_string()),
             distinctness: Some(ast::Distinctness::Distinct),
-            args: Some(vec![Expr::Id(Id("x".to_string()))]),
+            args: Some(vec![Expr::Id(Name::Ident("x".to_string()))]),
             order_by: None,
             filter_over: None,
         };
@@ -1289,14 +1280,14 @@ pub mod tests {
     #[test]
     fn test_like_expressions_equivalent() {
         let expr1 = Expr::Like {
-            lhs: Box::new(Expr::Id(Id("name".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("name".to_string()))),
             not: false,
             op: ast::LikeOperator::Like,
             rhs: Box::new(Expr::Literal(Literal::String("%john%".to_string()))),
             escape: Some(Box::new(Expr::Literal(Literal::String("\\".to_string())))),
         };
         let expr2 = Expr::Like {
-            lhs: Box::new(Expr::Id(Id("name".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("name".to_string()))),
             not: false,
             op: ast::LikeOperator::Like,
             rhs: Box::new(Expr::Literal(Literal::String("%john%".to_string()))),
@@ -1308,14 +1299,14 @@ pub mod tests {
     #[test]
     fn test_expressions_equivalent_like_escaped() {
         let expr1 = Expr::Like {
-            lhs: Box::new(Expr::Id(Id("name".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("name".to_string()))),
             not: false,
             op: ast::LikeOperator::Like,
             rhs: Box::new(Expr::Literal(Literal::String("%john%".to_string()))),
             escape: Some(Box::new(Expr::Literal(Literal::String("\\".to_string())))),
         };
         let expr2 = Expr::Like {
-            lhs: Box::new(Expr::Id(Id("name".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("name".to_string()))),
             not: false,
             op: ast::LikeOperator::Like,
             rhs: Box::new(Expr::Literal(Literal::String("%john%".to_string()))),
@@ -1326,13 +1317,13 @@ pub mod tests {
     #[test]
     fn test_expressions_equivalent_between() {
         let expr1 = Expr::Between {
-            lhs: Box::new(Expr::Id(Id("age".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("age".to_string()))),
             not: false,
             start: Box::new(Expr::Literal(Literal::Numeric("18".to_string()))),
             end: Box::new(Expr::Literal(Literal::Numeric("65".to_string()))),
         };
         let expr2 = Expr::Between {
-            lhs: Box::new(Expr::Id(Id("age".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("age".to_string()))),
             not: false,
             start: Box::new(Expr::Literal(Literal::Numeric("18".to_string()))),
             end: Box::new(Expr::Literal(Literal::Numeric("65".to_string()))),
@@ -1341,7 +1332,7 @@ pub mod tests {
 
         // differing BETWEEN bounds
         let expr3 = Expr::Between {
-            lhs: Box::new(Expr::Id(Id("age".to_string()))),
+            lhs: Box::new(Expr::Id(Name::Ident("age".to_string()))),
             not: false,
             start: Box::new(Expr::Literal(Literal::Numeric("20".to_string()))),
             end: Box::new(Expr::Literal(Literal::Numeric("65".to_string()))),
@@ -2066,17 +2057,17 @@ pub mod tests {
     #[test]
     fn test_parse_pragma_bool() {
         assert!(parse_pragma_bool(&Expr::Literal(Literal::Numeric("1".into()))).unwrap(),);
-        assert!(parse_pragma_bool(&Expr::Name(Name("true".into()))).unwrap(),);
-        assert!(parse_pragma_bool(&Expr::Name(Name("on".into()))).unwrap(),);
-        assert!(parse_pragma_bool(&Expr::Name(Name("yes".into()))).unwrap(),);
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("true".into()))).unwrap(),);
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("on".into()))).unwrap(),);
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("yes".into()))).unwrap(),);
 
         assert!(!parse_pragma_bool(&Expr::Literal(Literal::Numeric("0".into()))).unwrap(),);
-        assert!(!parse_pragma_bool(&Expr::Name(Name("false".into()))).unwrap(),);
-        assert!(!parse_pragma_bool(&Expr::Name(Name("off".into()))).unwrap(),);
-        assert!(!parse_pragma_bool(&Expr::Name(Name("no".into()))).unwrap(),);
+        assert!(!parse_pragma_bool(&Expr::Name(Name::Ident("false".into()))).unwrap(),);
+        assert!(!parse_pragma_bool(&Expr::Name(Name::Ident("off".into()))).unwrap(),);
+        assert!(!parse_pragma_bool(&Expr::Name(Name::Ident("no".into()))).unwrap(),);
 
-        assert!(parse_pragma_bool(&Expr::Name(Name("nono".into()))).is_err());
-        assert!(parse_pragma_bool(&Expr::Name(Name("10".into()))).is_err());
-        assert!(parse_pragma_bool(&Expr::Name(Name("-1".into()))).is_err());
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("nono".into()))).is_err());
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("10".into()))).is_err());
+        assert!(parse_pragma_bool(&Expr::Name(Name::Ident("-1".into()))).is_err());
     }
 }
