@@ -1,4 +1,4 @@
-use std::{cell::Cell, cmp::Ordering, rc::Rc, sync::Arc};
+use std::{cell::Cell, cmp::Ordering, sync::Arc};
 
 use tracing::{instrument, Level};
 use turso_sqlite3_parser::ast::{self, TableInternalId};
@@ -115,11 +115,11 @@ pub struct ProgramBuilder {
 
 #[derive(Debug, Clone)]
 pub enum CursorType {
-    BTreeTable(Rc<BTreeTable>),
+    BTreeTable(Arc<BTreeTable>),
     BTreeIndex(Arc<Index>),
     Pseudo(PseudoCursorType),
     Sorter,
-    VirtualTable(Rc<VirtualTable>),
+    VirtualTable(Arc<VirtualTable>),
 }
 
 impl CursorType {
@@ -293,7 +293,7 @@ impl ProgramBuilder {
     pub fn add_pragma_result_column(&mut self, col_name: String) {
         // TODO figure out a better type definition for ResultSetColumn
         // or invent another way to set pragma result columns
-        let expr = ast::Expr::Id(ast::Id("".to_string()));
+        let expr = ast::Expr::Id(ast::Name::Ident("".to_string()));
         self.result_columns.push(ResultSetColumn {
             expr,
             alias: Some(col_name),
@@ -301,7 +301,7 @@ impl ProgramBuilder {
         });
     }
 
-    #[instrument(skip(self), level = Level::INFO)]
+    #[instrument(skip(self), level = Level::DEBUG)]
     pub fn emit_insn(&mut self, insn: Insn) {
         let function = insn.to_function();
         // This seemingly empty trace here is needed so that a function span is emmited with it
@@ -768,8 +768,11 @@ impl ProgramBuilder {
             self.preassign_label_to_next_insn(self.init_label);
 
             match txn_mode {
-                TransactionMode::Read => self.emit_insn(Insn::Transaction { write: false }),
-                TransactionMode::Write => self.emit_insn(Insn::Transaction { write: true }),
+                TransactionMode::Read => self.emit_insn(Insn::Transaction {
+                    db: 0,
+                    write: false,
+                }),
+                TransactionMode::Write => self.emit_insn(Insn::Transaction { db: 0, write: true }),
                 TransactionMode::None => {}
             }
 
