@@ -357,11 +357,12 @@ pub fn insn_to_str(
             Insn::OpenRead {
                 cursor_id,
                 root_page,
+                db,
             } => (
                 "OpenRead",
                 *cursor_id as i32,
                 *root_page as i32,
-                0,
+                *db as i32,
                 Value::build_text(""),
                 0,
                 {
@@ -377,10 +378,11 @@ pub fn insn_to_str(
                                 }
                             });
                     format!(
-                        "{}={}, root={}",
+                        "{}={}, root={}, iDb={}",
                         cursor_type,
                         get_table_or_index_name(*cursor_id),
-                        root_page
+                        root_page,
+                        db
                     )
                 },
             ),
@@ -645,14 +647,14 @@ pub fn insn_to_str(
                 0,
                 "".to_string(),
             ),
-            Insn::Transaction { write } => (
+            Insn::Transaction { db, write } => (
                 "Transaction",
-                0,
+                *db as i32,
                 *write as i32,
                 0,
                 Value::build_text(""),
                 0,
-                format!("write={write}"),
+                format!("iDb={db} write={write}"),
             ),
             Insn::Goto { target_pc } => (
                 "Goto",
@@ -1213,7 +1215,7 @@ pub fn insn_to_str(
             Insn::OpenWrite {
                 cursor_id,
                 root_page,
-                name,
+                db,
                 ..
             } => (
                 "OpenWrite",
@@ -1222,20 +1224,20 @@ pub fn insn_to_str(
                     RegisterOrLiteral::Literal(i) => *i as _,
                     RegisterOrLiteral::Register(i) => *i as _,
                 },
-                0,
+                *db as i32,
                 Value::build_text(""),
                 0,
-                format!("root={root_page}; {name}"),
+                format!("root={root_page}; iDb={db}"),
             ),
             Insn::Copy {
                 src_reg,
                 dst_reg,
-                amount,
+                extra_amount,
             } => (
                 "Copy",
                 *src_reg as i32,
                 *dst_reg as i32,
-                *amount as i32,
+                *extra_amount as i32,
                 Value::build_text(""),
                 0,
                 format!("r[{dst_reg}]=r[{src_reg}]"),
