@@ -1,6 +1,5 @@
 use std::{
     num::{NonZero, NonZeroUsize},
-    rc::Rc,
     sync::Arc,
 };
 
@@ -109,6 +108,7 @@ pub struct InsertFlags(pub u8);
 
 impl InsertFlags {
     pub const UPDATE: u8 = 0x01; // Flag indicating this is part of an UPDATE statement
+    pub const REQUIRE_SEEK: u8 = 0x02; // Flag indicating that a seek is required to insert the row
 
     pub fn new() -> Self {
         InsertFlags(0)
@@ -118,12 +118,8 @@ impl InsertFlags {
         (self.0 & flag) != 0
     }
 
-    pub fn update(mut self, is_update: bool) -> Self {
-        if is_update {
-            self.0 |= InsertFlags::UPDATE;
-        } else {
-            self.0 &= !InsertFlags::UPDATE;
-        }
+    pub fn require_seek(mut self) -> Self {
+        self.0 |= InsertFlags::REQUIRE_SEEK;
         self
     }
 }
@@ -333,6 +329,7 @@ pub enum Insn {
     OpenRead {
         cursor_id: CursorID,
         root_page: PageIdx,
+        db: usize,
     },
 
     /// Open a cursor for a virtual table.
@@ -419,7 +416,7 @@ pub enum Insn {
         /// GENERATED ALWAYS AS ... STATIC columns are only checked if P3 is zero.
         /// When P3 is non-zero, no type checking occurs for static generated columns.
         check_generated: bool, // P3
-        table_reference: Rc<BTreeTable>, // P4
+        table_reference: Arc<BTreeTable>, // P4
     },
 
     // Make a record and write it to destination register.
@@ -462,7 +459,8 @@ pub enum Insn {
 
     /// Start a transaction.
     Transaction {
-        write: bool,
+        db: usize,   // p1
+        write: bool, // p2
     },
 
     /// Set database auto-commit mode and potentially rollback.
@@ -794,13 +792,13 @@ pub enum Insn {
     OpenWrite {
         cursor_id: CursorID,
         root_page: RegisterOrLiteral<PageIdx>,
-        name: String,
+        db: usize,
     },
 
     Copy {
         src_reg: usize,
         dst_reg: usize,
-        amount: usize, // 0 amount means we include src_reg, dst_reg..=dst_reg+amount = src_reg..=src_reg+amount
+        extra_amount: usize, // 0 extra_amount means we include src_reg, dst_reg..=dst_reg+amount = src_reg..=src_reg+amount
     },
 
     /// Allocate a new b-tree.
@@ -1139,4 +1137,6 @@ pub enum Cookie {
     UserVersion = 6,
     /// The auto-vacuum mode setting.
     IncrementalVacuum = 7,
+    /// The application ID as set by the application_id pragma.
+    ApplicationId = 8,
 }
