@@ -27,13 +27,10 @@ pub mod sorter;
 use crate::{
     error::LimboError,
     function::{AggFunc, FuncCtx},
-    storage::{pager::PagerCacheflushStatus, sqlite3_ondisk::SmallVec},
+    storage::{pager, sqlite3_ondisk::SmallVec},
     translate::plan::TableReferences,
-    types::{RawSlice, TextRef},
-    vdbe::execute::OpIdxInsertState,
-    vdbe::execute::OpInsertState,
-    vdbe::execute::OpNewRowidState,
-    vdbe::execute::OpSeekState,
+    types::{IOResult, RawSlice, TextRef},
+    vdbe::execute::{OpIdxInsertState, OpInsertState, OpNewRowidState, OpSeekState},
     RefValue,
 };
 
@@ -159,13 +156,13 @@ pub enum StepResult {
 }
 
 /// If there is I/O, the instruction is restarted.
-/// Evaluate a Result<CursorResult<T>>, if IO return Ok(StepResult::IO).
+/// Evaluate a Result<IOResult<T>>, if IO return Ok(StepResult::IO).
 #[macro_export]
-macro_rules! return_if_io {
+macro_rules! return_step_if_io {
     ($expr:expr) => {
         match $expr? {
-            CursorResult::Ok(v) => v,
-            CursorResult::IO => return Ok(StepResult::IO),
+            IOResult::Ok(v) => v,
+            IOResult::IO => return Ok(StepResult::IO),
         }
     };
 }
@@ -258,7 +255,7 @@ pub struct ProgramState {
     op_new_rowid_state: OpNewRowidState,
     op_idx_insert_state: OpIdxInsertState,
     op_insert_state: OpInsertState,
-    op_seek_state: OpSeekState,
+    seek_state: OpSeekState,
 }
 
 impl ProgramState {
@@ -288,7 +285,7 @@ impl ProgramState {
             op_new_rowid_state: OpNewRowidState::Start,
             op_idx_insert_state: OpIdxInsertState::SeekIfUnique,
             op_insert_state: OpInsertState::Insert,
-            op_seek_state: OpSeekState::Start,
+            seek_state: OpSeekState::Start,
         }
     }
 
@@ -385,7 +382,11 @@ pub struct Program {
 }
 
 impl Program {
-    #[instrument(skip_all, level = Level::INFO)]
+    fn get_pager_from_database_index(&self, idx: &usize) -> Rc<Pager> {
+        self.connection.get_pager_from_database_index(idx)
+    }
+
+    #[instrument(skip_all, level = Level::DEBUG)]
     pub fn step(
         &self,
         state: &mut ProgramState,
@@ -408,25 +409,22 @@ impl Program {
             let _ = state.result_row.take();
             let (insn, insn_function) = &self.insns[state.pc as usize];
             trace_insn(self, state.pc as InsnReference, insn);
-            let res = insn_function(self, state, insn, &pager, mv_store.as_ref());
-            if res.is_err() {
-                let state = self.connection.transaction_state.get();
-                if let TransactionState::Write { schema_did_change } = state {
-                    pager.rollback(schema_did_change, &self.connection)?
+            match insn_function(self, state, insn, &pager, mv_store.as_ref()) {
+                Ok(InsnFunctionStepResult::Step) => {}
+                Ok(InsnFunctionStepResult::Done) => return Ok(StepResult::Done),
+                Ok(InsnFunctionStepResult::IO) => return Ok(StepResult::IO),
+                Ok(InsnFunctionStepResult::Row) => return Ok(StepResult::Row),
+                Ok(InsnFunctionStepResult::Interrupt) => return Ok(StepResult::Interrupt),
+                Ok(InsnFunctionStepResult::Busy) => return Ok(StepResult::Busy),
+                Err(err) => {
+                    handle_program_error(&pager, &self.connection, &err)?;
+                    return Err(err);
                 }
             }
-            match res? {
-                InsnFunctionStepResult::Step => {}
-                InsnFunctionStepResult::Done => return Ok(StepResult::Done),
-                InsnFunctionStepResult::IO => return Ok(StepResult::IO),
-                InsnFunctionStepResult::Row => return Ok(StepResult::Row),
-                InsnFunctionStepResult::Interrupt => return Ok(StepResult::Interrupt),
-                InsnFunctionStepResult::Busy => return Ok(StepResult::Busy),
-            }
         }
     }
 
-    #[instrument(skip_all, level = Level::INFO)]
+    #[instrument(skip_all, level = Level::DEBUG)]
     pub fn commit_txn(
         &self,
         pager: Rc<Pager>,
@@ -493,7 +491,7 @@ impl Program {
         }
     }
 
-    #[instrument(skip(self, pager, connection), level = Level::INFO)]
+    #[instrument(skip(self, pager, connection), level = Level::DEBUG)]
     fn step_end_write_txn(
         &self,
         pager: &Rc<Pager>,
@@ -509,20 +507,17 @@ impl Program {
             connection.wal_checkpoint_disabled.get(),
         )?;
         match cacheflush_status {
-            PagerCacheflushStatus::Done(status) => {
+            IOResult::Done(status) => {
                 if self.change_cnt_on {
                     self.connection.set_changes(self.n_change.get());
                 }
-                if matches!(
-                    status,
-                    crate::storage::pager::PagerCacheflushResult::Rollback
-                ) {
+                if matches!(status, pager::PagerCommitResult::Rollback) {
                     pager.rollback(schema_did_change, connection)?;
                 }
                 connection.transaction_state.replace(TransactionState::None);
                 *commit_state = CommitState::Ready;
             }
-            PagerCacheflushStatus::IO => {
+            IOResult::IO => {
                 tracing::trace!("Cacheflush IO");
                 *commit_state = CommitState::Committing;
                 return Ok(StepResult::IO);
@@ -577,7 +572,7 @@ pub fn registers_to_ref_values(registers: &[Register]) -> Vec<RefValue> {
         .collect()
 }
 
-#[instrument(skip(program), level = Level::INFO)]
+#[instrument(skip(program), level = Level::DEBUG)]
 fn trace_insn(program: &Program, addr: InsnReference, insn: &Insn) {
     if !tracing::enabled!(tracing::Level::TRACE) {
         return;
@@ -751,3 +746,29 @@ impl Row {
         self.count
     }
 }
+
+/// Handle a program error by rolling back the transaction
+pub fn handle_program_error(
+    pager: &Rc<Pager>,
+    connection: &Connection,
+    err: &LimboError,
+) -> Result<()> {
+    match err {
+        LimboError::TxError(_) => {}
+        _ => {
+            let state = connection.transaction_state.get();
+            if let TransactionState::Write { schema_did_change } = state {
+                if let Err(e) = pager.rollback(schema_did_change, connection) {
+                    tracing::error!("rollback failed: {e}");
+                }
+                if let Err(e) = pager.end_tx(false, schema_did_change, connection, false) {
+                    tracing::error!("end_tx failed: {e}");
+                }
+            } else if let Err(e) = pager.end_read_tx() {
+                tracing::error!("end_read_tx failed: {e}");
+            }
+            connection.transaction_state.replace(TransactionState::None);
+        }
+    }
+    Ok(())
+}
