@@ -1,6 +1,6 @@
 use crate::pragma::{PragmaVirtualTable, PragmaVirtualTableCursor};
 use crate::schema::Column;
-use crate::util::{columns_from_create_table_body, vtable_args};
+use crate::util::columns_from_create_table_body;
 use crate::{Connection, LimboError, SymbolTable, Value};
 use fallible_iterator::FallibleIterator;
 use std::cell::RefCell;
@@ -11,7 +11,7 @@ use turso_ext::{ConstraintInfo, IndexInfo, OrderByInfo, ResultCode, VTabKind, VT
 use turso_sqlite3_parser::{ast, lexer::sql::Parser};
 
 #[derive(Debug, Clone)]
-enum VirtualTableType {
+pub(crate) enum VirtualTableType {
     Pragma(PragmaVirtualTable),
     External(ExtVirtualTable),
 }
@@ -19,29 +19,40 @@ enum VirtualTableType {
 #[derive(Clone, Debug)]
 pub struct VirtualTable {
     pub(crate) name: String,
-    pub(crate) args: Option<Vec<ast::Expr>>,
     pub(crate) columns: Vec<Column>,
     pub(crate) kind: VTabKind,
     vtab_type: VirtualTableType,
 }
 
 impl VirtualTable {
-    pub(crate) fn function(
-        name: &str,
-        args: Option<Vec<ast::Expr>>,
-        syms: &SymbolTable,
-    ) -> crate::Result<Rc<VirtualTable>> {
+    pub(crate) fn readonly(self: &Arc<VirtualTable>) -> bool {
+        match &self.vtab_type {
+            VirtualTableType::Pragma(_) => true,
+            VirtualTableType::External(table) => table.readonly(),
+        }
+    }
+
+    pub(crate) fn builtin_functions() -> Vec<Arc<VirtualTable>> {
+        PragmaVirtualTable::functions()
+            .into_iter()
+            .map(|(tab, schema)| {
+                let vtab = VirtualTable {
+                    name: format!("pragma_{}", tab.pragma_name),
+                    columns: Self::resolve_columns(schema)
+                        .expect("built-in function schema resolution should not fail"),
+                    kind: VTabKind::TableValuedFunction,
+                    vtab_type: VirtualTableType::Pragma(tab),
+                };
+                Arc::new(vtab)
+            })
+            .collect()
+    }
+
+    pub(crate) fn function(name: &str, syms: &SymbolTable) -> crate::Result<Arc<VirtualTable>> {
         let module = syms.vtab_modules.get(name);
         let (vtab_type, schema) = if module.is_some() {
-            let ext_args = match args {
-                Some(ref args) => vtable_args(args),
-                None => vec![],
-            };
-            ExtVirtualTable::create(name, module, ext_args, VTabKind::TableValuedFunction)
+            ExtVirtualTable::create(name, module, Vec::new(), VTabKind::TableValuedFunction)
                 .map(|(vtab, columns)| (VirtualTableType::External(vtab), columns))?
-        } else if let Some(pragma_name) = name.strip_prefix("pragma_") {
-            PragmaVirtualTable::create(pragma_name)
-                .map(|(vtab, columns)| (VirtualTableType::Pragma(vtab), columns))?
         } else {
             return Err(LimboError::ParseError(format!(
                 "No such table-valued function: {name}"
@@ -50,12 +61,11 @@ impl VirtualTable {
 
         let vtab = VirtualTable {
             name: name.to_owned(),
-            args,
             columns: Self::resolve_columns(schema)?,
             kind: VTabKind::TableValuedFunction,
             vtab_type,
         };
-        Ok(Rc::new(vtab))
+        Ok(Arc::new(vtab))
     }
 
     pub fn table(
@@ -63,18 +73,17 @@ impl VirtualTable {
         module_name: &str,
         args: Vec<turso_ext::Value>,
         syms: &SymbolTable,
-    ) -> crate::Result<Rc<VirtualTable>> {
+    ) -> crate::Result<Arc<VirtualTable>> {
         let module = syms.vtab_modules.get(module_name);
         let (table, schema) =
             ExtVirtualTable::create(module_name, module, args, VTabKind::VirtualTable)?;
         let vtab = VirtualTable {
             name: tbl_name.unwrap_or(module_name).to_owned(),
-            args: None,
             columns: Self::resolve_columns(schema)?,
             kind: VTabKind::VirtualTable,
             vtab_type: VirtualTableType::External(table),
         };
-        Ok(Rc::new(vtab))
+        Ok(Arc::new(vtab))
     }
 
     fn resolve_columns(schema: String) -> crate::Result<Vec<Column>> {
@@ -121,12 +130,7 @@ impl VirtualTable {
         order_by: &[OrderByInfo],
     ) -> IndexInfo {
         match &self.vtab_type {
-            VirtualTableType::Pragma(_) => {
-                // SQLite tries to estimate cost and row count for pragma_ TVFs,
-                // but since Limbo doesn't have cost-based planning yet, this
-                // estimation is not currently implemented.
-                Default::default()
-            }
+            VirtualTableType::Pragma(table) => table.best_index(constraints),
             VirtualTableType::External(table) => table.best_index(constraints, order_by),
         }
     }
@@ -176,7 +180,7 @@ impl VirtualTableCursor {
 }
 
 #[derive(Clone, Debug)]
-struct ExtVirtualTable {
+pub(crate) struct ExtVirtualTable {
     implementation: Rc<VTabModuleImpl>,
     table_ptr: *const c_void,
     connection_ptr: RefCell<Option<*mut turso_ext::Conn>>,
@@ -198,6 +202,9 @@ impl Drop for ExtVirtualTable {
 }
 
 impl ExtVirtualTable {
+    pub(crate) fn readonly(&self) -> bool {
+        self.implementation.readonly
+    }
     fn best_index(&self, constraints: &[ConstraintInfo], order_by: &[OrderByInfo]) -> IndexInfo {
         unsafe {
             IndexInfo::from_ffi((self.implementation.best_idx)(
